; file: "mem.scm", 3/13/96

; Test program for Gambit-C's memory management system.

; Directions: Compile this program statically, run the binary and at
; the read-eval-print loop run the desired test.  For example:
;
; > gsc -link mem-test
; > cc mem-test.c mem-test_.c -lgambc -lm
; > a.out
; 
; : (test1)
; ...

;------------------------------------------------------------------------------

; Utilities:

; pseudo-random number generator

(define seed #f)

(define (reset-random)
  (set! seed 1923202963))

(define (random n)
  (let* ((hi (quotient seed 44488))
         (lo (remainder seed 44488))
         (test (- (* lo 48271) (* hi 3399))))
    (set! seed (if (> test 0) test (+ test 2147483647)))
    (modulo seed n)))

; repetition of an evaluation

(define (repeat n thunk)
  (let loop ((i n))
    (if (> i 0)
      (begin
        (thunk)
        (loop (- i 1))))))

(define gc-reports? #t)

(define (setup-gc-reports)
  (set-gc-report gc-reports?))

(define (trigger-gc)
  (##gc))

;------------------------------------------------------------------------------

; Test #1
;
; Allocate and deallocate medium size objects and large objects.  The GC
; reports show that movable and still objects are allocated and reclaimed.

(define (test1)

  (setup-gc-reports)

  (reset-random)

  (trigger-gc)

  (let* ((n 4)
         (v (make-vector n #f)))

    (repeat
      400
      (lambda ()
        (let ((i (random n))
              (size (if (= 0 (random 2)) 4000 5000)))
          (vector-set! v i (make-vector size))))))

  (trigger-gc))

;------------------------------------------------------------------------------

; Test #2
;
; Allocate and deallocate medium size objects and large objects.  The GC
; reports show that movable and still objects are allocated and reclaimed.

(define (test2)

  (setup-gc-reports)

  (reset-random)

  (trigger-gc)

  (set! test2-data '())

  ; allocate roughly 1 MB of data

  (repeat
    50
    (lambda ()
      (set! test2-data
        (cons (make-vector (if (= 0 (random 2)) 4000 5000))
              test2-data))))

  (idle)

  ; release roughly .5 MB of data

  (repeat
    25
    (lambda ()
      (set! test2-data (cdr test2-data))))

  (idle)

  ; allocate roughly 1 MB more

  (repeat
    50
    (lambda ()
      (set! test2-data
        (cons (make-vector (if (= 0 (random 2)) 4000 5000))
              test2-data))))

  (idle)

  ; release it all

  (set! test2-data #f)

  (trigger-gc))

(define test2-data #f)

(define (idle)
  (let loop ((i 2000))
    (if (> i 0)
      (begin
        (make-vector (if (= 0 (random 2)) 4000 5000))
        (loop (- i 1))))))

;------------------------------------------------------------------------------

; Test #3
;
; Show how to recover from heap overflows.
;
; Note: make sure you specify a heap size limit when you start the
; compiler or interpreter, otherwise it may take a long time before
; the virtual memory is exhausted.

(define (test3)

  (setup-gc-reports)

  (reset-random)

  (trigger-gc)

  (test3-helper 10)      ; this call should terminate normally

  (test3-helper 1000000) ; this one will cause a heap overflow

  (test3-helper 10)      ; this call should terminate normally
)

(define (test3-helper n)
  (let ((action
          (call-with-current-continuation
            (lambda (abort)
              (##catch-all
                (lambda (signal-name args)
                  (abort
                    (lambda ()
                      (trigger-gc) ; clean up heap and recover overflow reserve
                      (display "caught signal ")
                      (write signal-name)
                      (newline))))
                (lambda ()
                  (let ((data (create-big-set-of-data n)))
                    (lambda ()
                      (trigger-gc)
                      (write (map vector? data))
                      (newline)))))))))
    (action)))

(define (create-big-set-of-data n)
  (let loop ((lst '()) (i n))
    (if (> i 0)
     (loop
        (cons (case (random 6)
                ((0)  (make-string (random 50000)))
                ((1)  (make-vector (random 10000)))
                ((2)  (cons lst lst))
                ((3)  (gensym))
                ((4)  (lambda (x) lst))
                (else (* 1.23 (random 100))))
              lst)
        (- i 1))
     lst)))

;------------------------------------------------------------------------------

; Test #4
;
; Show how to recover from stack overflows.
;
; Note: make sure you specify a heap size limit when you start the
; compiler or interpreter, otherwise it may take a long time before
; the virtual memory is exhausted.

(define (test4)

  (setup-gc-reports)

  (reset-random)

  (trigger-gc)

  (test4-helper 10)      ; this call should terminate normally

  (test4-helper 1000000) ; this one will cause a heap overflow

  (test4-helper 10)      ; this call should terminate normally
)

(define (test4-helper n)
  (let ((action
          (call-with-current-continuation
            (lambda (abort)
              (##catch-all
                (lambda (signal-name args)
                  (abort
                    (lambda ()
                      (trigger-gc) ; clean up heap and recover overflow reserve
                      (display "caught signal ")
                      (write signal-name)
                      (newline))))
                (lambda ()
                  (let ((data (deeply-nested-computation n)))
                    (lambda ()
                      (trigger-gc)
                      (write (map even? data))
                      (newline)))))))))
    (action)))

(define (deeply-nested-computation n)
  (if (> n 0)
    (cons n (deeply-nested-computation (- n 1)))
    '()))

;------------------------------------------------------------------------------

; Test #5
;
; Allocation of Scheme objects from C.

(##c-declaration "

/* these declarations are from the header file mem.h */

extern ___WORD ___make_pair (/* car, cdr, kind */);
extern ___WORD ___make_vector (/* length, init, kind */);
extern ___WORD ___make_string (/* length, init, kind */);
extern ___WORD ___string_to_scheme (/* str, kind */);
extern ___WORD ___double_to_scheme (/* val, kind */);
extern void ___still_obj_refcount_inc (/* obj */);
extern void ___still_obj_refcount_dec (/* obj */);

static ___WORD obj1, obj2, obj3, obj4, obj5; /* some Scheme objects */

void foo ()
{
  obj1 = ___double_to_scheme (1.23, ___STILL);
  obj2 = ___make_vector (5, obj1, ___STILL);
  obj3 = ___string_to_scheme (\"hello world!\", ___STILL);
  obj4 = ___string_to_scheme (\"another string\", ___STILL);
  obj5 = ___make_pair (obj2, obj3, ___STILL);
  ___still_obj_refcount_dec (obj1); /* I have no direct need for obj1, etc */
  ___still_obj_refcount_dec (obj2);
  ___still_obj_refcount_dec (obj3);
}

___WORD bar ()
{
  return obj4;
}

___WORD baz ()
{
  ___still_obj_refcount_dec (obj5);
  return obj5;
}
")

(define foo (##c-procedure () void "foo"))
(define bar (##c-procedure () scheme-object "bar"))
(define baz (##c-procedure () scheme-object "baz"))

(define (test5)

  (setup-gc-reports)

  (trigger-gc)

  (foo) ; create objects from C

  (trigger-gc) ; they will show up as "not movable" objects in the GC report

  (let ((x (bar))) ; get obj4 from C

    (trigger-gc) ; no problem if a GC happens here

    (write x)
    (newline))

  (trigger-gc)

  (let ((y (baz)))

    (trigger-gc)

    (write y)
    (newline))

  (trigger-gc)) ; at this point, only obj4 is not reclaimed (because refcount != 0)

;------------------------------------------------------------------------------

; Test #6
;
; Finalization test.

(define (test6)

  (setup-gc-reports)

  (reset-random)

  (trigger-gc)

  (let* ((n 100)
         (v (make-vector n #f))
         (c 0)
         (expect 0)
         (result 0))

    (define (new)
      (let ((id c))
        (set! expect (+ expect id))
        (set! c (+ c 1))
        (let* ((action
                 (lambda (x) (set! result (+ result id))))
               (obj
                 (case (random 6)
                   ((0)  (make-string (random 10)))
                   ((1)  (make-vector (random 10) id))
                   ((2)  (cons id id))
                   ((3)  (gensym))
                   ((4)  (lambda (x) id))
                   (else (* 1.23 (random 100))))))
          (make-will obj action)
          obj)))

    (repeat
      20000
      (lambda ()
        (let ((i (random n)))
          (vector-set! v i (new)))))

    (set! v #f)

    (write '(final-gc))
    (newline)

    (trigger-gc)

    (write expect)
    (newline)

    (write result)
    (newline)))

;------------------------------------------------------------------------------

(define (test-all)
  (set! gc-reports? #f)
  (test1)
  (test2)
  (test3)
  (test4)
  (test5)
  (test6))

(##repl) ; start a read-eval-print loop
