;==============================================================================

; file: "_gvm.scm"

; Copyright (C) 1994, 1995, 1996 by Marc Feeley, All Rights Reserved.

(include "fixnum.scm")

;------------------------------------------------------------------------------
;
; Gambit virtual machine abstraction module:
; -----------------------------------------

; (See file 'doc/gvm' for details on the virtual machine)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Virtual machine operands:
; ------------------------
;
; Operands are represented with small integers.  Operands can thus be tested
; for equality using 'eqv?'.  The encoding is as follows:
;
; OPERAND      ENCODING         
;
; reg(n)       0     + n
; stk(n)       10000 + n
; lbl(n)       20000 + n
; glo(name)    30000 + index in operand table
; clo(opnd,n)  40000 + index in operand table
; obj(x)       50000 + index in operand table

; Utilities:
; ---------

(define *opnd-table* '())
(define *opnd-table-alloc* '())

(define opnd-table-size 10000)

(define (enter-opnd arg1 arg2)
  (let loop ((i 0))
    (if (< i *opnd-table-alloc*)
      (let ((x (vector-ref *opnd-table* i)))
        (if (and (eqv? (car x) arg1) (eqv? (cdr x) arg2))
          i
          (loop (+ i 1))))
      (if (< *opnd-table-alloc* opnd-table-size)
        (begin
          (set! *opnd-table-alloc* (+ *opnd-table-alloc* 1))
          (vector-set! *opnd-table* i (cons arg1 arg2))
          i)
        (compiler-limitation-error
          "program is too long [virtual machine operand table overflow]")))))

(define (contains-opnd? opnd1 opnd2) ; does opnd2 contain opnd1?
  (cond ((eqv? opnd1 opnd2)
         #t)
        ((clo? opnd2)
         (contains-opnd? opnd1 (clo-base opnd2)))
        (else
         #f)))

(define (any-contains-opnd? opnd opnds)
  (if (null? opnds)
    #f
    (or (contains-opnd? opnd (car opnds))
        (any-contains-opnd? opnd (cdr opnds)))))

; Locations:
; ---------

; -- location is a register (first is number 0)
(define (make-reg num) num)
(define (reg? x) (< x 10000))
(define (reg-num x) (modulo x 10000))

; -- location is in the stack (first slot in procedure's frame is number 1)
(define (make-stk num) (+ num 10000))
(define (stk? x) (= (quotient x 10000) 1))
(define (stk-num x) (modulo x 10000))

; -- location is a global variable
(define (make-glo name) (+ (enter-opnd name #t) 30000))
(define (glo? x) (= (quotient x 10000) 3))
(define (glo-name x) (car (vector-ref *opnd-table* (modulo x 10000))))

; -- location is a closed variable (base is ptr to closure env, index >= 1)
(define (make-clo base index) (+ (enter-opnd base index) 40000))
(define (clo? x) (= (quotient x 10000) 4))
(define (clo-base x) (car (vector-ref *opnd-table* (modulo x 10000))))
(define (clo-index x) (cdr (vector-ref *opnd-table* (modulo x 10000))))

; Values:
; ------

; -- value is the address of a local label
(define (make-lbl num) (+ num 20000))
(define (lbl? x) (= (quotient x 10000) 2))
(define (lbl-num x) (modulo x 10000))
(define label-limit 9999) ; largest label

; -- value is a scheme object
(define (make-obj val) (+ (enter-opnd val #f) 50000))                    
(define (obj? x) (= (quotient x 10000) 5))         
(define (obj-val x) (car (vector-ref *opnd-table* (modulo x 10000))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Processor context descriptions:
; ------------------------------

(define (make-pcontext fs map)
  (vector fs map))

(define (pcontext-fs  x) (vector-ref x 0))
(define (pcontext-map x) (vector-ref x 1))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Frame description:
; -----------------

(define (make-frame size slots regs closed live)
  (vector size slots regs closed live))

(define (frame-size x)   (vector-ref x 0))
(define (frame-slots x)  (vector-ref x 1))
(define (frame-regs x)   (vector-ref x 2))
(define (frame-closed x) (vector-ref x 3))
(define (frame-live x)   (vector-ref x 4))

(define (frame-eq? x y)
  (= (frame-size x) (frame-size y)))

(define (frame-truncate frame nb-slots)
  (let ((fs (frame-size frame)))
    (make-frame nb-slots
                (nth-after (frame-slots frame) (- fs nb-slots))
                (frame-regs frame)
                (frame-closed frame)
                (frame-live frame))))

(define (frame-live? var frame)
  (let ((live (frame-live frame)))
    (if (eq? var closure-env-var)
      (let ((closed (frame-closed frame)))
        (if (or (set-member? var live)
                (not (set-empty? (set-intersection live (list->set closed)))))
          closed
          #f))
      (if (set-member? var live)
        var
        #f))))

(define (frame-first-empty-slot frame)
  (let loop ((i 1) (s (reverse (frame-slots frame))))
    (if (pair? s)
      (if (frame-live? (car s) frame)
        (loop (+ i 1) (cdr s))
        i)
      i)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Procedure objects:
; -----------------

(define (make-proc-obj
          name
          primitive?
          code
          call-pat
          side-effects?
          strict-pat
          type
          standard)
  (let ((proc-obj
          (vector
            proc-obj-tag
            name
            primitive?
            code
            call-pat
            #f ; test
            #f ; inlinable
            #f ; specialize
            side-effects?
            strict-pat
            type
            standard)))
    (proc-obj-specialize-set! proc-obj (lambda (decls args) proc-obj))
    proc-obj))

(define proc-obj-tag (list 'PROC-OBJ))

(define (proc-obj? x)
  (and (vector? x)
       (> (vector-length x) 0)
       (eq? (vector-ref x 0) proc-obj-tag)))

(define (proc-obj-name obj)                (vector-ref obj 1))
(define (proc-obj-primitive? obj)          (vector-ref obj 2))
(define (proc-obj-code obj)                (vector-ref obj 3))
(define (proc-obj-call-pat obj)            (vector-ref obj 4))
(define (proc-obj-test obj)                (vector-ref obj 5))
(define (proc-obj-inlinable obj)           (vector-ref obj 6))
(define (proc-obj-specialize obj)          (vector-ref obj 7))
(define (proc-obj-side-effects? obj)       (vector-ref obj 8))
(define (proc-obj-strict-pat obj)          (vector-ref obj 9))
(define (proc-obj-type obj)                (vector-ref obj 10))
(define (proc-obj-standard obj)            (vector-ref obj 11))

(define (proc-obj-code-set! obj x)         (vector-set! obj 3 x))
(define (proc-obj-test-set! obj x)         (vector-set! obj 5 x))
(define (proc-obj-inlinable-set! obj x)    (vector-set! obj 6 x))
(define (proc-obj-specialize-set! obj x)   (vector-set! obj 7 x))

(define (make-pattern min-args nb-parms rest?)
  (let loop ((x (if rest? (- nb-parms 1) (list nb-parms)))
             (y (if rest? (- nb-parms 1) nb-parms)))
    (let ((z (- y 1)))
      (if (< z min-args) x (loop (cons z x) z)))))

(define (pattern-member? n pat) ; tests if 'n' is a member of pattern 'pat'
  (cond ((pair? pat)
         (if (= (car pat) n) #t (pattern-member? n (cdr pat))))
        ((null? pat)
         #f)
        (else
         (<= pat n))))

(define (type-name type)
  (if (pair? type) (car type) type))

(define (type-pot-fut? type)
  (pair? type))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Basic block set manipulation:
; ----------------------------

; Virtual instructions have a linear structure.  However, this is not how
; they are put together to form a piece of code.  Rather, virtual instructions
; are grouped into 'basic blocks' which are 'linked' together.  A basic block
; is a LABEL instruction followed by a sequence of non-branching instructions
; (i.e. APPLY, COPY or CLOSE) terminated by a single branch instruction (i.e.
; IFJUMP or JUMP).  Links between basic blocks are denoted using label
; references.  When a basic block ends with an IFJUMP instruction, the block is
; linked to the two basic blocks corresponding to the two possible control
; paths out of the IFJUMP instruction.  When a basic block ends with a JUMP
; instruction, there is either zero or one link.
;
; Basic blocks naturally group together to form 'basic block sets'.  A basic
; block set describes all the code of a procedure.

(define (make-bbs)
  (vector bbs-tag
          (make-counter 1 label-limit bbs-limit-err) ; 1 - local label counter
          (queue-empty)                              ; 2 - basic block queue
          '()))                                      ; 3 - entry label number

(define (bbs-limit-err)
  (compiler-limitation-error "procedure is too long [too many labels]"))

(define bbs-tag (list 'BBS))

(define (bbs? x)
  (and (vector? x)
       (> (vector-length x) 0)
       (eq? (vector-ref x 0) bbs-tag)))

(define (bbs-lbl-counter bbs)                (vector-ref bbs 1))
(define (bbs-lbl-counter-set! bbs cntr)      (vector-set! bbs 1 cntr))
(define (bbs-bb-queue bbs)                   (vector-ref bbs 2))
(define (bbs-bb-queue-set! bbs bbq)          (vector-set! bbs 2 bbq))
(define (bbs-entry-lbl-num bbs)              (vector-ref bbs 3))
(define (bbs-entry-lbl-num-set! bbs lbl-num) (vector-set! bbs 3 lbl-num))

(define (bbs-new-lbl! bbs)
  ((bbs-lbl-counter bbs)))

(define (lbl-num->bb lbl-num bbs)
  (let loop ((bb-list (queue->list (bbs-bb-queue bbs))))
    (if (= (bb-lbl-num (car bb-list)) lbl-num)
      (car bb-list)
      (loop (cdr bb-list)))))

; Basic block manipulation procedures:

(define (make-bb label-instr bbs)
  (let ((bb (vector
              label-instr   ; 0 - LABEL instr
              (queue-empty) ; 1 - sequence of non-branching instrs
              '()           ; 2 - branch instruction
              '()           ; 3 - basic blocks referenced by this block
              '())))        ; 4 - basic blocks which jump to this block
                            ;     (both filled in by 'bbs-purify!')
    (queue-put! (bbs-bb-queue bbs) bb)
    bb))

(define (bb-lbl-num bb)                  (LABEL-lbl-num (vector-ref bb 0)))
(define (bb-label-type bb)               (LABEL-type (vector-ref bb 0)))
(define (bb-label-instr bb)              (vector-ref bb 0))
(define (bb-label-instr-set! bb l)       (vector-set! bb 0 l))
(define (bb-non-branch-instrs bb)        (queue->list (vector-ref bb 1)))
(define (bb-non-branch-instrs-set! bb l) (vector-set! bb 1 (list->queue l)))
(define (bb-branch-instr bb)             (vector-ref bb 2))
(define (bb-branch-instr-set! bb b)      (vector-set! bb 2 b))
(define (bb-references bb)               (vector-ref bb 3))
(define (bb-references-set! bb l)        (vector-set! bb 3 l))
(define (bb-precedents bb)               (vector-ref bb 4))
(define (bb-precedents-set! bb l)        (vector-set! bb 4 l))

(define (bb-entry-frame-size bb)
  (frame-size (gvm-instr-frame (bb-label-instr bb))))

(define (bb-exit-frame-size bb)
  (frame-size (gvm-instr-frame (bb-branch-instr bb))))

(define (bb-slots-gained bb)
  (- (bb-exit-frame-size bb) (bb-entry-frame-size bb)))

(define (bb-put-non-branch! bb gvm-instr)
  (queue-put! (vector-ref bb 1) gvm-instr))

(define (bb-put-branch! bb gvm-instr)
  (vector-set! bb 2 gvm-instr))

(define (bb-add-reference! bb ref)
  (if (not (memq ref (vector-ref bb 3)))
    (vector-set! bb 3 (cons ref (vector-ref bb 3)))))

(define (bb-add-precedent! bb prec)
  (if (not (memq prec (vector-ref bb 4)))
    (vector-set! bb 4 (cons prec (vector-ref bb 4)))))

(define (bb-last-non-branch-instr bb)
  (let ((non-branch-instrs (bb-non-branch-instrs bb)))
    (if (null? non-branch-instrs)
      (bb-label-instr bb)
      (let loop ((l non-branch-instrs))
        (if (pair? (cdr l))
          (loop (cdr l))
          (car l))))))

; Virtual machine instruction representation:

(define (gvm-instr-type gvm-instr)    (vector-ref gvm-instr 0))
(define (gvm-instr-frame gvm-instr)   (vector-ref gvm-instr 1))
(define (gvm-instr-comment gvm-instr) (vector-ref gvm-instr 2))

(define (make-LABEL-SIMPLE lbl-num frame comment)
  (vector 'LABEL frame comment lbl-num 'SIMPLE))

(define (make-LABEL-ENTRY lbl-num nb-parms min rest? closed? frame comment)
  (vector 'LABEL frame comment lbl-num 'ENTRY nb-parms min rest? closed?))

(define (make-LABEL-RETURN lbl-num frame comment)
  (vector 'LABEL frame comment lbl-num 'RETURN))

(define (make-LABEL-TASK-ENTRY lbl-num frame comment)
  (vector 'LABEL frame comment lbl-num 'TASK-ENTRY))

(define (make-LABEL-TASK-RETURN lbl-num frame comment)
  (vector 'LABEL frame comment lbl-num 'TASK-RETURN))

(define (LABEL-lbl-num gvm-instr)        (vector-ref gvm-instr 3))
(define (LABEL-lbl-num-set! gvm-instr n) (vector-set! gvm-instr 3 n))
(define (LABEL-type gvm-instr)           (vector-ref gvm-instr 4))

(define (LABEL-ENTRY-nb-parms gvm-instr) (vector-ref gvm-instr 5))
(define (LABEL-ENTRY-min gvm-instr)      (vector-ref gvm-instr 6))
(define (LABEL-ENTRY-rest? gvm-instr)    (vector-ref gvm-instr 7))
(define (LABEL-ENTRY-closed? gvm-instr)  (vector-ref gvm-instr 8))

(define (make-APPLY prim opnds loc frame comment)
  (vector 'APPLY frame comment prim opnds loc))
(define (APPLY-prim gvm-instr)  (vector-ref gvm-instr 3))
(define (APPLY-opnds gvm-instr) (vector-ref gvm-instr 4))
(define (APPLY-loc gvm-instr)   (vector-ref gvm-instr 5))

(define (make-COPY opnd loc frame comment)
  (vector 'COPY frame comment opnd loc))

(define (COPY-opnd gvm-instr) (vector-ref gvm-instr 3))
(define (COPY-loc gvm-instr)  (vector-ref gvm-instr 4))

(define (make-CLOSE parms frame comment)
  (vector 'CLOSE frame comment parms))
(define (CLOSE-parms gvm-instr) (vector-ref gvm-instr 3))

(define (make-closure-parms loc lbl opnds)
  (vector loc lbl opnds))
(define (closure-parms-loc x)   (vector-ref x 0))
(define (closure-parms-lbl x)   (vector-ref x 1))
(define (closure-parms-opnds x) (vector-ref x 2))

(define (make-IFJUMP test opnds true false poll? frame comment)
  (vector 'IFJUMP frame comment test opnds true false poll?))
(define (IFJUMP-test gvm-instr)  (vector-ref gvm-instr 3))
(define (IFJUMP-opnds gvm-instr) (vector-ref gvm-instr 4))
(define (IFJUMP-true gvm-instr)  (vector-ref gvm-instr 5))
(define (IFJUMP-false gvm-instr) (vector-ref gvm-instr 6))
(define (IFJUMP-poll? gvm-instr) (vector-ref gvm-instr 7))

(define (make-JUMP opnd nb-args poll? safe? frame comment)
  (vector 'JUMP frame comment opnd nb-args poll? safe?))
(define (JUMP-opnd gvm-instr)    (vector-ref gvm-instr 3))
(define (JUMP-nb-args gvm-instr) (vector-ref gvm-instr 4))
(define (JUMP-poll? gvm-instr)   (vector-ref gvm-instr 5))
(define (JUMP-safe? gvm-instr)   (vector-ref gvm-instr 6))
(define (first-class-JUMP? gvm-instr) (JUMP-nb-args gvm-instr))

(define (make-comment)
  (cons 'COMMENT '()))

(define (comment-put! comment name val)
  (set-cdr! comment (cons (cons name val) (cdr comment))))

(define (comment-get comment name)
  (and comment
       (let ((x (assq name (cdr comment))))
         (if x (cdr x) #f))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; 'Purification' of basic block sets:
; ----------------------------------

; This step removes unreachable basic blocks (i.e. dead code), duplicate
; basic blocks (i.e. common code), useless jumps and jump cascades from
; a basic block set.  It also orders the basic blocks so that the destination
; of a branch is put (if possible) right after the branch instruction.  The
; 'references' and 'precedents' fields of each basic block are also filled in
; through the process.  The first basic block of a 'purified' basic block set
; is always the entry point.

(define (bbs-purify! bbs)
  (let loop () ; iterate until code does not change
    (bbs-remove-jump-cascades! bbs)
    (bbs-remove-dead-code! bbs)
    (let* ((changed1? (bbs-remove-common-code! bbs))
           (changed2? (bbs-remove-useless-jumps! bbs)))
      (if (or changed1? changed2?) (loop) (bbs-order! bbs)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Step 1, Jump cascade removal:

(define (bbs-remove-jump-cascades! bbs)

  (define (empty-bb? bb)
    (and (eq? (bb-label-type bb) 'SIMPLE)     ; simple label and
         (null? (bb-non-branch-instrs bb))))  ; no non-branching instrs

  (define (jump-to-non-entry-lbl? branch)
    (and (eq? (gvm-instr-type branch) 'JUMP)
         (not (first-class-JUMP? branch)) ; not a jump to an entry label
         (jump-lbl? branch)))

  (define (jump-cascade-to lbl-num fs poll? seen thunk)
    (if (memq lbl-num seen) ; infinite loop?
      (thunk lbl-num fs poll?)
      (let ((bb (lbl-num->bb lbl-num bbs)))
        (if (and (empty-bb? bb) (<= (bb-slots-gained bb) 0))
          (let ((jump-lbl-num
                 (jump-to-non-entry-lbl? (bb-branch-instr bb))))
            (if jump-lbl-num
              (jump-cascade-to
                jump-lbl-num
                (+ fs (bb-slots-gained bb))
                (or poll? (JUMP-poll? (bb-branch-instr bb)))
                (cons lbl-num seen)
                thunk)
              (thunk lbl-num fs poll?)))
          (thunk lbl-num fs poll?)))))

  (define (equiv-lbl lbl-num seen)
    (if (memq lbl-num seen) ; infinite loop?
      lbl-num
      (let ((bb (lbl-num->bb lbl-num bbs)))
        (if (empty-bb? bb)
          (let ((jump-lbl-num
                 (jump-to-non-entry-lbl? (bb-branch-instr bb))))
            (if (and jump-lbl-num
                     (not (JUMP-poll? (bb-branch-instr bb)))
                     (= (bb-slots-gained bb) 0))
              (equiv-lbl jump-lbl-num (cons lbl-num seen))
              lbl-num))
          lbl-num))))

  (define (remove-cascade! bb)
    (let ((branch (bb-branch-instr bb)))

      (case (gvm-instr-type branch)

        ((IFJUMP)
         (bb-put-branch! bb  ; branch is an IFJUMP
           (make-IFJUMP (IFJUMP-test branch)
                        (IFJUMP-opnds branch)
                        (equiv-lbl (IFJUMP-true branch) '())
                        (equiv-lbl (IFJUMP-false branch) '())
                        (IFJUMP-poll? branch)
                        (gvm-instr-frame branch)
                        (gvm-instr-comment branch))))

        ((JUMP)  ; branch is a JUMP
         (if (not (first-class-JUMP? branch)) ; but not to an entry label
           (let ((dest-lbl-num (jump-lbl? branch)))
             (if dest-lbl-num

               (jump-cascade-to
                 dest-lbl-num
                 (frame-size (gvm-instr-frame branch))
                 (JUMP-poll? branch)
                 '()
                 (lambda (lbl-num fs poll?)
                   (let* ((dest-bb (lbl-num->bb lbl-num bbs))
                          (last-branch (bb-branch-instr dest-bb)))
                     (if (and (empty-bb? dest-bb)
                              (or (not poll?)
                                  put-poll-on-IFJUMP?
                                  (not (eq? (gvm-instr-type last-branch)
                                            'IFJUMP))))

                       (let* ((new-fs (+ fs (bb-slots-gained dest-bb)))
                              (new-frame (frame-truncate
                                           (gvm-instr-frame branch)
                                           new-fs)))

                         (define (adjust-opnd opnd)
                           (cond ((stk? opnd)
                                  (make-stk
                                    (+ (- fs (bb-entry-frame-size dest-bb))
                                       (stk-num opnd))))
                                 ((clo? opnd)
                                  (make-clo (adjust-opnd (clo-base opnd))
                                            (clo-index opnd)))
                                 (else
                                  opnd)))

                         (case (gvm-instr-type last-branch)
                           ((IFJUMP)
                            (bb-put-branch! bb
                              (make-IFJUMP (IFJUMP-test last-branch)
                                           (map adjust-opnd
                                                (IFJUMP-opnds last-branch))
                                           (equiv-lbl
                                             (IFJUMP-true last-branch)
                                             '())
                                           (equiv-lbl
                                             (IFJUMP-false last-branch)
                                             '())
                                           (or poll?
                                               (IFJUMP-poll? last-branch))
                                           new-frame
                                           (gvm-instr-comment last-branch))))
                           ((JUMP)
                            (bb-put-branch! bb
                              (make-JUMP (adjust-opnd (JUMP-opnd last-branch))
                                         (JUMP-nb-args last-branch)
                                         (or poll?
                                             (JUMP-poll? last-branch))
                                         (JUMP-safe? last-branch)
                                         new-frame
                                         (gvm-instr-comment last-branch))))
                           (else
                            (compiler-internal-error
                              "bbs-remove-jump-cascades!, unknown branch type"))))

                       (bb-put-branch! bb
                         (make-JUMP (make-lbl lbl-num)
                                    (JUMP-nb-args branch)
                                    (or poll?
                                        (JUMP-poll? branch))
                                    (JUMP-safe? branch)
                                    (frame-truncate
                                      (gvm-instr-frame branch)
                                      fs)
                                    (gvm-instr-comment branch)))))))))))

        (else
         (compiler-internal-error
           "bbs-remove-jump-cascades!, unknown branch type")))))

  (for-each remove-cascade!
            (queue->list (bbs-bb-queue bbs))))

(define (jump-lbl? branch)
  (let ((opnd (JUMP-opnd branch)))
    (if (lbl? opnd) (lbl-num opnd) #f)))

(define put-poll-on-IFJUMP? #f)
(set! put-poll-on-IFJUMP? #t)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Step 2, Dead code removal:

(define (bbs-remove-dead-code! bbs)

  (let ((new-bb-queue (queue-empty))
        (scan-queue (queue-empty)))

    (define (reachable ref bb)
      (if bb (bb-add-reference! bb ref))
      (if (not (memq ref (queue->list new-bb-queue)))
        (begin
          (bb-references-set! ref '())
          (bb-precedents-set! ref '())
          (queue-put! new-bb-queue ref)
          (queue-put! scan-queue ref))))

    (define (direct-jump to-bb from-bb)
      (reachable to-bb from-bb)
      (bb-add-precedent! to-bb from-bb))

    (define (scan-instr gvm-instr bb)

      (define (scan-opnd gvm-opnd)
        (cond ((lbl? gvm-opnd)
               (reachable (lbl-num->bb (lbl-num gvm-opnd) bbs) bb))
              ((clo? gvm-opnd)
               (scan-opnd (clo-base gvm-opnd)))))

      (case (gvm-instr-type gvm-instr)

        ((LABEL)
         '())

        ((APPLY)
         (for-each scan-opnd (APPLY-opnds gvm-instr))
         (if (APPLY-loc gvm-instr)
           (scan-opnd (APPLY-loc gvm-instr))))

        ((COPY)
         (scan-opnd (COPY-opnd gvm-instr))
         (scan-opnd (COPY-loc gvm-instr)))

        ((CLOSE)
         (for-each (lambda (parm)
                     (reachable (lbl-num->bb (closure-parms-lbl parm) bbs) bb)
                     (scan-opnd (closure-parms-loc parm))
                     (for-each scan-opnd (closure-parms-opnds parm)))
                   (CLOSE-parms gvm-instr)))

        ((IFJUMP)
         (for-each scan-opnd (IFJUMP-opnds gvm-instr))
         (direct-jump (lbl-num->bb (IFJUMP-true gvm-instr) bbs) bb)
         (direct-jump (lbl-num->bb (IFJUMP-false gvm-instr) bbs) bb))

        ((JUMP)
         (let ((opnd (JUMP-opnd gvm-instr)))
           (if (lbl? opnd)
             (direct-jump (lbl-num->bb (lbl-num opnd) bbs) bb)
             (scan-opnd (JUMP-opnd gvm-instr)))))

        (else
         (compiler-internal-error
           "bbs-remove-dead-code!, unknown GVM instruction type"))))

    (reachable (lbl-num->bb (bbs-entry-lbl-num bbs) bbs) #f)

    (let loop ()
      (if (not (queue-empty? scan-queue))
        (let ((bb (queue-get! scan-queue)))
          (begin
            (scan-instr (bb-label-instr bb) bb)
            (for-each (lambda (gvm-instr) (scan-instr gvm-instr bb))
                      (bb-non-branch-instrs bb))
            (scan-instr (bb-branch-instr bb) bb)
            (loop)))))

    (bbs-bb-queue-set! bbs new-bb-queue)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Step 3, Useless jump removal:

(define (bbs-remove-useless-jumps! bbs)
  (let ((changed? #f))

    (define (remove-useless-jump bb)
      (let ((branch (bb-branch-instr bb)))

        ; is it a non-polling JUMP to a label?

        (if (and (eq? (gvm-instr-type branch) 'JUMP)
                 (not (first-class-JUMP? branch))
                 (not (JUMP-poll? branch))
                 (jump-lbl? branch))
          (let* ((dest-bb (lbl-num->bb (jump-lbl? branch) bbs))
                 (frame1 (gvm-instr-frame (bb-last-non-branch-instr bb)))
                 (frame2 (gvm-instr-frame (bb-label-instr dest-bb))))

            ; is it a SIMPLE label with the same frame as the last non-branch instruction?

            (if (and (eq? (bb-label-type dest-bb) 'SIMPLE)
                     (frame-eq? frame1 frame2)
                     (= (length (bb-precedents dest-bb)) 1))

              (begin
                (set! changed? #t)
                (bb-non-branch-instrs-set! bb
                  (append (bb-non-branch-instrs bb)
                          (bb-non-branch-instrs dest-bb)
                          '()))
                (bb-branch-instr-set! bb
                  (bb-branch-instr dest-bb))
                (remove-useless-jump bb)))))))

    (for-each remove-useless-jump
              (queue->list (bbs-bb-queue bbs)))

    changed?))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Step 4, Common code removal:

(define (bbs-remove-common-code! bbs)
  (let* ((bb-list (queue->list (bbs-bb-queue bbs)))
         (n (length bb-list)))
    (if (> n 500) ; if code is too large, don't optimize

      #f

      (let* ((hash-table-length (if (< n 50) 43 403))
             (hash-table (make-vector hash-table-length '()))
             (prim-table '())
             (block-map '())
             (changed? #f))

      (define (hash-prim prim)
        (let ((n (length prim-table))
              (i (pos-in-list prim prim-table)))
          (if i
            (- n i)
            (begin
              (set! prim-table (cons prim prim-table))
              (+ n 1)))))

      (define (hash-opnds l) ; this assumes that operands are encoded with nbs
        (let loop ((l l) (n 0))
          (if (pair? l)
            (loop (cdr l)
                  (let ((x (car l)))
                    (if (lbl? x)
                      n
                      (modulo (+ (* n 10000) x) hash-table-length))))
            n)))

      (define (hash-bb bb) ; compute hash address for a basic block
        (let ((branch (bb-branch-instr bb)))
          (modulo
            (case (gvm-instr-type branch)
              ((IFJUMP)
               (+ (hash-opnds (IFJUMP-opnds branch))
                  (* 10 (hash-prim (IFJUMP-test branch)))
                  (* 100 (frame-size (gvm-instr-frame branch)))))
              ((JUMP)
               (+ (hash-opnds (list (JUMP-opnd branch)))
                  (* 10 (or (JUMP-nb-args branch) -1))
                  (* 100 (frame-size (gvm-instr-frame branch)))))
              (else
               0))
            hash-table-length)))

      (define (replacement-lbl-num lbl)
        (let ((x (assv lbl block-map)))
          (if x (cdr x) lbl)))

      (define (fix-map! bb1 bb2) ; bb1 should be replaced by bb2 in the block-map
        (let loop ((l block-map))
          (if (pair? l)
            (let ((x (car l)))
              (if (= bb1 (cdr x)) (set-cdr! x bb2))
              (loop (cdr l))))))

      (define (enter-bb! bb) ; enter a basic block in the hash table
        (let ((h (hash-bb bb)))
          (vector-set! hash-table h
            (add-bb bb (vector-ref hash-table h)))))

      (define (add-bb bb l) ; add basic block 'bb' to list of basic blocks
        (if (pair? l)
          (let ((bb* (car l))) ; pick next basic block in list

            (set! block-map ; for now, assume that 'bb' = 'bb*'
              (cons (cons (bb-lbl-num bb) (bb-lbl-num bb*))
                    block-map))

            (if (eqv-bb? bb bb*) ; are they the same?

              (begin
                (fix-map! (bb-lbl-num bb) (bb-lbl-num bb*)) ; record the equivalence
            (set! changed? #t)
                l)

              (begin
                (set! block-map (cdr block-map)) ; they are not the same!
                (if (eqv-gvm-instr? (bb-branch-instr bb) (bb-branch-instr bb*))

                  (extract-common-tail bb bb* ; check if tail is the same
                    (lambda (head head* tail)
                      (if (null? tail) ; common tail long enough?

                        (cons bb* (add-bb bb (cdr l))) ; no, so try rest of list

                        (let* ((lbl (bbs-new-lbl! bbs)) ; create bb for common tail
                               (branch (bb-branch-instr bb))
                               (fs** (need-gvm-instrs tail branch))
                               (frame (frame-truncate
                                        (gvm-instr-frame
                                          (if (null? head)
                                            (bb-label-instr bb)
                                            (car head)))
                                        fs**))
                               (bb** (make-bb (make-LABEL-SIMPLE lbl frame #f) bbs)))
                          (bb-non-branch-instrs-set! bb** tail)
                          (bb-branch-instr-set! bb** branch)
                          (bb-non-branch-instrs-set! bb* (reverse head*))
                          (bb-branch-instr-set! bb*
                            (make-JUMP (make-lbl lbl) #f #f #f frame #f))
                          (bb-non-branch-instrs-set! bb (reverse head))
                          (bb-branch-instr-set! bb
                            (make-JUMP (make-lbl lbl) #f #f #f frame #f))
                          (set! changed? #t)
                          (cons bb (cons bb* (add-bb bb** (cdr l))))))))

                    (cons bb* (add-bb bb (cdr l)))))))

            (list bb)))

      (define (extract-common-tail bb1 bb2 cont)
        (let loop ((l1 (reverse (bb-non-branch-instrs bb1)))
                   (l2 (reverse (bb-non-branch-instrs bb2)))
                   (tail '()))
          (if (and (pair? l1) (pair? l2))
            (let ((i1 (car l1))
                  (i2 (car l2)))
              (if (eqv-gvm-instr? i1 i2)
                (loop (cdr l1) (cdr l2) (cons i1 tail))
                (cont l1 l2 tail)))
            (cont l1 l2 tail))))

      (define (eqv-bb? bb1 bb2)
        (let ((bb1-non-branch (bb-non-branch-instrs bb1))
              (bb2-non-branch (bb-non-branch-instrs bb2)))
          (and (= (length bb1-non-branch) (length bb2-non-branch))
               (eqv-gvm-instr? (bb-label-instr bb1) (bb-label-instr bb2))
               (eqv-gvm-instr? (bb-branch-instr bb1) (bb-branch-instr bb2))
               (eqv-list? eqv-gvm-instr? bb1-non-branch bb2-non-branch))))

      (define (eqv-list? pred? l1 l2)
        (if (pair? l1)
          (and (pair? l2)
               (pred? (car l1) (car l2))
               (eqv-list? pred? (cdr l1) (cdr l2)))
          (not (pair? l2))))

      (define (eqv-lbl-num? lbl1 lbl2)
        (= (replacement-lbl-num lbl1)
           (replacement-lbl-num lbl2)))

      (define (eqv-gvm-opnd? opnd1 opnd2)
        (if (not opnd1)
          (not opnd2)
          (and opnd2
               (cond ((lbl? opnd1)
                      (and (lbl? opnd2)
                           (eqv-lbl-num? (lbl-num opnd1) (lbl-num opnd2))))
                     ((clo? opnd1)
                      (and (clo? opnd2)
                           (= (clo-index opnd1) (clo-index opnd2))
                           (eqv-gvm-opnd? (clo-base opnd1)
                                          (clo-base opnd2))))
                     (else
                      (eqv? opnd1 opnd2))))))
    
      (define (eqv-gvm-instr? instr1 instr2)

        (define (eqv-closure-parms? p1 p2)
          (and (eqv-gvm-opnd? (closure-parms-loc p1)
                              (closure-parms-loc p2))
               (eqv-lbl-num? (closure-parms-lbl p1)
                             (closure-parms-lbl p2))
               (eqv-list? eqv-gvm-opnd?
                          (closure-parms-opnds p1)
                          (closure-parms-opnds p2))))

        (let ((type1 (gvm-instr-type instr1))
              (type2 (gvm-instr-type instr2)))
          (and (eq? type1 type2)
               (frame-eq? (gvm-instr-frame instr1) (gvm-instr-frame instr2))
               (case type1

                 ((LABEL)
                  (let ((ltype1 (LABEL-type instr1))
                        (ltype2 (LABEL-type instr2)))
                    (and (eq? ltype1 ltype2)
                         (case ltype1
                           ((SIMPLE RETURN TASK-ENTRY TASK-RETURN)
                            #t)
                           ((ENTRY)
                            (and (= (LABEL-ENTRY-min instr1)
                                    (LABEL-ENTRY-min instr2))
                                 (= (LABEL-ENTRY-nb-parms instr1)
                                    (LABEL-ENTRY-nb-parms instr2))
                                 (eq? (LABEL-ENTRY-rest? instr1)
                                      (LABEL-ENTRY-rest? instr2))
                                 (eq? (LABEL-ENTRY-closed? instr1)
                                      (LABEL-ENTRY-closed? instr2))))
                           (else
                            (compiler-internal-error
                              "eqv-gvm-instr?, unknown label type"))))))

                 ((APPLY)
                  (and (eq? (APPLY-prim instr1) (APPLY-prim instr2))
                       (eqv-list? eqv-gvm-opnd?
                                  (APPLY-opnds instr1)
                                  (APPLY-opnds instr2))
                       (eqv-gvm-opnd? (APPLY-loc instr1)
                                      (APPLY-loc instr2))))

                 ((COPY)
                  (and (eqv-gvm-opnd? (COPY-opnd instr1)
                                      (COPY-opnd instr2))
                       (eqv-gvm-opnd? (COPY-loc instr1)
                                      (COPY-loc instr2))))

                 ((CLOSE)
                  (eqv-list? eqv-closure-parms?
                             (CLOSE-parms instr1)
                             (CLOSE-parms instr2)))

                 ((IFJUMP)
                  (and (eq? (IFJUMP-test instr1)
                            (IFJUMP-test instr2))
                       (eqv-list? eqv-gvm-opnd?
                                  (IFJUMP-opnds instr1)
                                  (IFJUMP-opnds instr2))
                       (eqv-lbl-num? (IFJUMP-true instr1)
                                     (IFJUMP-true instr2))
                       (eqv-lbl-num? (IFJUMP-false instr1)
                                     (IFJUMP-false instr2))
                       (eq? (IFJUMP-poll? instr1)
                            (IFJUMP-poll? instr2))))

                 ((JUMP)
                  (and (eqv-gvm-opnd? (JUMP-opnd instr1)
                                      (JUMP-opnd instr2))
                       (eqv? (JUMP-nb-args instr1)
                             (JUMP-nb-args instr2))
                       (eq? (JUMP-poll? instr1)
                            (JUMP-poll? instr2))
                       (eq? (JUMP-safe? instr1)
                            (JUMP-safe? instr2))))

                 (else
                  (compiler-internal-error
                    "eqv-gvm-instr?, unknown 'gvm-instr':" instr1))))))

      (define (update-bb! bb)
        (replace-label-references! bb replacement-lbl-num))

      ; Fill hash table, remove equivalent basic blocks and common tails

      (for-each enter-bb! bb-list)

      ; Reconstruct bbs

      (bbs-entry-lbl-num-set! bbs
        (replacement-lbl-num (bbs-entry-lbl-num bbs)))

      (let loop ((i 0) (result '()))
        (if (< i hash-table-length)
          (let ((bb-kept (vector-ref hash-table i)))
            (for-each update-bb! bb-kept)
            (loop (+ i 1) (append bb-kept result)))
          (bbs-bb-queue-set! bbs (list->queue result))))

      changed?))))

(define (replace-label-references! bb replacement-lbl-num)

  (define (update-gvm-opnd opnd)
    (if opnd
      (cond ((lbl? opnd)
             (make-lbl (replacement-lbl-num (lbl-num opnd))))
            ((clo? opnd)
             (make-clo (update-gvm-opnd (clo-base opnd)) (clo-index opnd)))
            (else
             opnd))
      opnd))

  (define (update-gvm-instr instr)

    (define (update-closure-parms p)
      (make-closure-parms
        (update-gvm-opnd (closure-parms-loc p))
        (replacement-lbl-num (closure-parms-lbl p))
        (map update-gvm-opnd (closure-parms-opnds p))))

    (case (gvm-instr-type instr)

      ((APPLY)
       (make-APPLY (APPLY-prim instr)
                   (map update-gvm-opnd (APPLY-opnds instr))
                   (update-gvm-opnd (APPLY-loc instr))
                   (gvm-instr-frame instr)
                   (gvm-instr-comment instr)))

      ((COPY)
       (make-COPY (update-gvm-opnd (COPY-opnd instr))
                  (update-gvm-opnd (COPY-loc instr))
                  (gvm-instr-frame instr)
                  (gvm-instr-comment instr)))

      ((CLOSE)
       (make-CLOSE
         (map update-closure-parms (CLOSE-parms instr))
         (gvm-instr-frame instr)
         (gvm-instr-comment instr)))

      ((IFJUMP)
       (make-IFJUMP (IFJUMP-test instr)
                    (map update-gvm-opnd (IFJUMP-opnds instr))
                    (replacement-lbl-num (IFJUMP-true instr))
                    (replacement-lbl-num (IFJUMP-false instr))
                    (IFJUMP-poll? instr)
                    (gvm-instr-frame instr)
                    (gvm-instr-comment instr)))

      ((JUMP)
       (make-JUMP (update-gvm-opnd (JUMP-opnd instr))
                  (JUMP-nb-args instr)
                  (JUMP-poll? instr)
                  (JUMP-safe? instr)
                  (gvm-instr-frame instr)
                  (gvm-instr-comment instr)))

      (else
       (compiler-internal-error
         "update-gvm-instr, unknown 'instr':" instr))))

  (bb-non-branch-instrs-set! bb
    (map update-gvm-instr (bb-non-branch-instrs bb)))
  (bb-branch-instr-set! bb
    (update-gvm-instr (bb-branch-instr bb))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Step 5, Basic block set ordering:

(define (bbs-order! bbs)

  (let ((new-bb-queue (queue-empty))
        (left-to-schedule (queue->list (bbs-bb-queue bbs))))

    (define (remove x l)
      (if (eq? (car l) x)
        (cdr l)
        (cons (car l) (remove x (cdr l)))))

    ; update list of basic blocks not yet scheduled

    (define (remove-bb! bb)
      (set! left-to-schedule (remove bb left-to-schedule))
      bb)

    ; return a basic block which ends with a branch to 'bb' (and that is
    ; still in 'left-to-schedule') or #f if there aren't any

    (define (prec-bb bb)
      (let loop ((l (bb-precedents bb)) (best #f) (best-fs #f))
        (if (null? l)
          best
          (let* ((x (car l))
                 (x-fs (bb-exit-frame-size x)))
            (if (and (memq x left-to-schedule)
                     (or (not best) (< x-fs best-fs)))
              (loop (cdr l) x x-fs)
              (loop (cdr l) best best-fs))))))

    ; return the basic block which 'bb' jumps to (and that is still in
    ; 'left-to-schedule') or #f if there aren't any

    (define (succ-bb bb)

      (define (branches-to-lbl? bb)
        (let ((branch (bb-branch-instr bb)))
          (case (gvm-instr-type branch)
            ((IFJUMP) #t)
            ((JUMP) (lbl? (JUMP-opnd branch)))
            (else
             (compiler-internal-error
              "bbs-order!, unknown branch type")))))

      (define (best-succ bb1 bb2)   ; heuristic that determines which
        (if (branches-to-lbl? bb1)  ; bb is most frequently executed
           bb1
           (if (branches-to-lbl? bb2)
             bb2
             (if (< (bb-exit-frame-size bb1)
                    (bb-exit-frame-size bb2))
               bb2
               bb1))))

      (let ((branch (bb-branch-instr bb)))
        (case (gvm-instr-type branch)
          ((IFJUMP)
           (let* ((true-bb (lbl-num->bb (IFJUMP-true branch) bbs))
                  (true-bb* (and (memq true-bb left-to-schedule)
                                 true-bb))
                  (false-bb (lbl-num->bb (IFJUMP-false branch) bbs))
                  (false-bb* (and (memq false-bb left-to-schedule)
                                  false-bb)))
             (if (and true-bb* false-bb*)
               (best-succ true-bb* false-bb*)
               (or true-bb* false-bb*))))
          ((JUMP)
           (let ((opnd (JUMP-opnd branch)))
             (and (lbl? opnd)
                  (let ((bb (lbl-num->bb (lbl-num opnd) bbs)))
                    (and (memq bb left-to-schedule) bb)))))
          (else
           (compiler-internal-error
             "bbs-order!, unknown branch type")))))

    ; schedule a given basic block 'bb' with it's predecessors and
    ; successors.

    (define (schedule-from bb)
      (queue-put! new-bb-queue bb)
      (let ((x (succ-bb bb)))
        (if x
          (begin
            (schedule-around (remove-bb! x))
            (let ((y (succ-bb bb)))
              (if y
                (schedule-around (remove-bb! y)))))))
      (schedule-refs bb))

    (define (schedule-around bb)
      (let ((x (prec-bb bb)))
        (if x
          (let ((bb-list (schedule-back (remove-bb! x) '())))
            (queue-put! new-bb-queue x)
            (schedule-forw bb)
            (for-each schedule-refs bb-list))
          (schedule-from bb))))

    (define (schedule-back bb bb-list)
      (let ((bb-list* (cons bb bb-list))
            (x (prec-bb bb)))
        (if x
          (let ((bb-list (schedule-back (remove-bb! x) bb-list*)))
            (queue-put! new-bb-queue x)
            bb-list)
          bb-list*)))

    (define (schedule-forw bb)
      (queue-put! new-bb-queue bb)
      (let ((x (succ-bb bb)))
        (if x
          (begin
            (schedule-forw (remove-bb! x))
            (let ((y (succ-bb bb)))
              (if y
                (schedule-around (remove-bb! y)))))))
      (schedule-refs bb))

    (define (schedule-refs bb)
      (for-each
        (lambda (x)
          (if (memq x left-to-schedule) (schedule-around (remove-bb! x))))
        (bb-references bb)))

    (schedule-from (remove-bb! (lbl-num->bb (bbs-entry-lbl-num bbs) bbs)))

    (bbs-bb-queue-set! bbs new-bb-queue)

    (let ((bb-list (queue->list new-bb-queue)))
      (let loop ((l bb-list) (i 1) (lbl-map '()))
        (if (pair? l)

          (let* ((label-instr (bb-label-instr (car l)))
                 (old-lbl-num (LABEL-lbl-num label-instr)))
            (LABEL-lbl-num-set! label-instr i)
            (loop (cdr l) (+ i 1) (cons (cons old-lbl-num i) lbl-map)))

          (let ()

            (define (replacement-lbl-num x)
              (cdr (assv x lbl-map)))

            (define (update-bb! bb)
              (replace-label-references! bb replacement-lbl-num))

            (for-each update-bb! bb-list)

            (bbs-lbl-counter-set!
              bbs
              (make-counter (* (+ 1 (quotient (bbs-new-lbl! bbs) 1000)) 1000)
                            label-limit
                            bbs-limit-err))))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Sequentialization of a basic block set:
; --------------------------------------

; The procedure 'bbs->code-list' transforms a 'purified' basic block set
; into a sequence of virtual machine instructions.  Each element of the
; resulting list is a 'code' object that contains a GVM instruction,
; a pointer to the basic block it came from and a `slots needed' index
; that specifies the minimum number of slots that have to be kept (relative
; to the start of the frame) after the instruction is executed.
; The first element of the code list is the entry label for the piece of code.

(define (make-code bb gvm-instr sn)     (vector bb gvm-instr sn))
(define (code-bb code)                  (vector-ref code 0))
(define (code-gvm-instr code)           (vector-ref code 1))
(define (code-slots-needed code)        (vector-ref code 2))
(define (code-slots-needed-set! code n) (vector-set! code 2 n))

(define (bbs->code-list bbs)
  (let ((code-list (linearize bbs)))
    (setup-slots-needed! code-list)
    code-list))

(define (linearize bbs) ; turn bbs into list and remove LABELs & JUMPs
  (let ((code-queue (queue-empty)))

    (define (put-bb bb)

      (define (put-instr gvm-instr)
        (queue-put! code-queue (make-code bb gvm-instr #f)))

      (put-instr (bb-label-instr bb))
      (for-each put-instr (bb-non-branch-instrs bb))
      (put-instr (bb-branch-instr bb)))

    (for-each put-bb (queue->list (bbs-bb-queue bbs)))
    (queue->list code-queue)))

(define (setup-slots-needed! code-list) ; setup `slots-needed' field
  (if (null? code-list)
    #f
    (let* ((code (car code-list))
           (gvm-instr (code-gvm-instr code))
           (sn-rest (setup-slots-needed! (cdr code-list))))

      (case (gvm-instr-type gvm-instr)

        ((LABEL)
         (if (> sn-rest (frame-size (gvm-instr-frame gvm-instr)))
           (compiler-internal-error
             "setup-slots-needed!, incoherent slots needed for LABEL"))
         (code-slots-needed-set! code sn-rest)
         #f)

        ((IFJUMP JUMP)
         (let ((sn (frame-size (gvm-instr-frame gvm-instr))))
           (code-slots-needed-set! code sn)
           (need-gvm-instr gvm-instr sn)))
         
        (else 
         (code-slots-needed-set! code sn-rest)
         (need-gvm-instr gvm-instr sn-rest))))))

(define (need-gvm-instrs non-branch branch)
  (if (pair? non-branch)
    (need-gvm-instr (car non-branch)
                    (need-gvm-instrs (cdr non-branch) branch))
    (need-gvm-instr branch (frame-size (gvm-instr-frame branch)))))

(define (need-gvm-instr gvm-instr sn-rest)
  (case (gvm-instr-type gvm-instr)

    ((LABEL)
     sn-rest)

    ((APPLY)
     (let ((loc (APPLY-loc gvm-instr)))
       (need-gvm-opnds (APPLY-opnds gvm-instr)
         (need-gvm-loc-opnd loc
           (need-gvm-loc loc sn-rest)))))

    ((COPY)
     (let ((loc (COPY-loc gvm-instr)))
       (need-gvm-opnd (COPY-opnd gvm-instr)
         (need-gvm-loc-opnd loc
           (need-gvm-loc loc sn-rest)))))

    ((CLOSE)
     (let ((parms (CLOSE-parms gvm-instr)))

       (define (need-parms-opnds p)
         (if (null? p)
           sn-rest
           (need-gvm-opnds (closure-parms-opnds (car p))
             (need-parms-opnds (cdr p)))))

       (define (need-parms-loc p)
         (if (null? p)
           (need-parms-opnds parms)
           (let ((loc (closure-parms-loc (car p))))
             (need-gvm-loc-opnd loc
               (need-gvm-loc loc (need-parms-loc (cdr p)))))))

       (need-parms-loc parms)))

    ((IFJUMP)
     (need-gvm-opnds (IFJUMP-opnds gvm-instr) sn-rest))

    ((JUMP)
     (need-gvm-opnd (JUMP-opnd gvm-instr) sn-rest))
         
    (else 
     (compiler-internal-error
       "need-gvm-instr, unknown 'gvm-instr':" gvm-instr))))

(define (need-gvm-loc loc sn-rest)
  (if (and loc (stk? loc) (>= (stk-num loc) sn-rest))
    (- (stk-num loc) 1)
    sn-rest))

(define (need-gvm-loc-opnd gvm-loc slots-needed)
  (if (and gvm-loc (clo? gvm-loc))
    (need-gvm-opnd (clo-base gvm-loc) slots-needed)
    slots-needed))

(define (need-gvm-opnd gvm-opnd slots-needed)
  (cond ((stk? gvm-opnd)
         (max (stk-num gvm-opnd) slots-needed))
        ((clo? gvm-opnd)
         (need-gvm-opnd (clo-base gvm-opnd) slots-needed))
        (else
         slots-needed)))

(define (need-gvm-opnds gvm-opnds slots-needed)
  (if (null? gvm-opnds)
    slots-needed
    (need-gvm-opnd (car gvm-opnds)
                   (need-gvm-opnds (cdr gvm-opnds) slots-needed))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Basic block writing:
; -------------------

(define (write-bb bb port)
  (write-gvm-instr (bb-label-instr bb) port)
  (display " [precedents=" port)
  (write (map bb-lbl-num (bb-precedents bb)) port)
  (display "]" port)
  (newline port)

  (for-each (lambda (x) (write-gvm-instr x port) (newline port))
            (bb-non-branch-instrs bb))

  (write-gvm-instr (bb-branch-instr bb) port))

(define (write-bbs bbs port)
  (for-each (lambda (bb)
              (if (= (bb-lbl-num bb) (bbs-entry-lbl-num bbs))
                (begin (display "**** Entry block:" port) (newline port)))
              (write-bb bb port)
              (newline port))
            (queue->list (bbs-bb-queue bbs))))

(define (virtual.dump procs port)

  (let ((proc-seen (queue-empty))
        (proc-left (queue-empty)))

    (define (scan-opnd gvm-opnd)
      (cond ((obj? gvm-opnd)
             (let ((val (obj-val gvm-opnd)))
               (if (and (proc-obj? val)
                        (proc-obj-code val)
                        (not (memq val (queue->list proc-seen))))
                 (begin
                   (queue-put! proc-seen val)
                   (queue-put! proc-left val)))))
            ((clo? gvm-opnd)
             (scan-opnd (clo-base gvm-opnd)))))

    (define (dump-proc p)

      (define (scan-code code)
        (let ((gvm-instr (code-gvm-instr code)))

          (write-gvm-instr gvm-instr port)
          (newline port)
          (case (gvm-instr-type gvm-instr)

            ((APPLY)
             (for-each scan-opnd (APPLY-opnds gvm-instr))
             (if (APPLY-loc gvm-instr)
               (scan-opnd (APPLY-loc gvm-instr))))

            ((COPY)
             (scan-opnd (COPY-opnd gvm-instr))
             (scan-opnd (COPY-loc gvm-instr)))

            ((CLOSE)
             (for-each (lambda (parms)
                         (scan-opnd (closure-parms-loc parms))
                         (for-each scan-opnd (closure-parms-opnds parms)))
                       (CLOSE-parms gvm-instr)))

            ((IFJUMP)
             (for-each scan-opnd (IFJUMP-opnds gvm-instr)))

            ((JUMP)
             (scan-opnd (JUMP-opnd gvm-instr)))

            (else
             '()))))

      (if (proc-obj-primitive? p)
        (display "**** #[primitive " port)
        (display "**** #[procedure " port))
      (display (proc-obj-name p) port)
      (display "] =" port)
      (newline port)

      (let ((x (proc-obj-code p)))
        (if (bbs? x)

          (let loop ((l (bbs->code-list x))
                     (prev-filename "")
                     (prev-line 0))
            (if (pair? l)
              (let* ((code (car l))
                     (instr (code-gvm-instr code))
                     (src (comment-get (gvm-instr-comment instr) 'SOURCE))
                     (loc (and src (source-locat src)))
                     (filename
                       (if (and loc (eq? (vector-ref loc 0) 'FILE))
                         (vector-ref loc 1)
                         prev-filename))
                     (line
                       (if (and loc (eq? (vector-ref loc 0) 'FILE))
                         (vector-ref loc 3)
                         prev-line)))
                (if (or (not (string=? filename prev-filename))
                        (not (= line prev-line)))
                  (begin
                    (display "#line " port)
                    (display line port)
                    (if (not (string=? filename prev-filename))
                      (begin
                        (display " " port)
                        (write filename port)))
                    (newline port)))

                (scan-code code)
                (loop (cdr l) filename line))
              (newline port)))

          (begin
            (display "C procedure of arity " port)
            (display (c-proc-arity x) port)
            (display "and body:" port)
            (newline port)
            (display (c-proc-body x) port)
            (newline port)))))

    (for-each (lambda (proc) (scan-opnd (make-obj proc))) procs)

    (let loop ()
      (if (not (queue-empty? proc-left))
        (begin
          (dump-proc (queue-get! proc-left))
          (loop))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Virtual instruction writing:
; ---------------------------

(define (write-gvm-instr gvm-instr port)

  (define (write-closure-parms parms)
    (display " " port)
    (let ((len (+ 1 (write-gvm-opnd (closure-parms-loc parms) port))))
      (display " = (" port)
      (let ((len (+ len (+ 4 (write-gvm-lbl (closure-parms-lbl parms) port)))))
        (+ len
           (write-terminated-opnd-list (closure-parms-opnds parms) port)))))

  (define (write-terminated-opnd-list l port)
    (let loop ((l l) (len 0))
      (if (pair? l)
        (let ((opnd (car l)))
          (display " " port)
          (loop (cdr l) (+ len (+ 1 (write-gvm-opnd opnd port)))))
        (begin
          (display ")" port)
          (+ len 1)))))

  (define (write-param-pattern gvm-instr port)
    (let ((len (if (not (= (LABEL-ENTRY-min gvm-instr)
                           (LABEL-ENTRY-nb-parms gvm-instr)))
                   (let ((len (write-returning-len
                                (LABEL-ENTRY-min gvm-instr)
                                port)))
                     (display "-" port)
                     (+ len 1))
                   0)))
      (let ((len (+ len
                    (write-returning-len
                      (LABEL-ENTRY-nb-parms gvm-instr)
                       port))))
        (if (LABEL-ENTRY-rest? gvm-instr)
          (begin (display "+" port) (+ len 1))
          len))))

  (define (write-prim-applic prim opnds port)
    (display "(" port)
    (let ((len (+ 1 (display-returning-len (proc-obj-name prim) port))))
      (+ len (write-terminated-opnd-list opnds port))))

  (define (write-instr gvm-instr)
    (case (gvm-instr-type gvm-instr)

      ((LABEL)
       (let ((len (write-gvm-lbl (LABEL-lbl-num gvm-instr) port)))
         (display " " port)
         (let ((len (+ len
                       (+ 1 (write-returning-len
                              (frame-size (gvm-instr-frame gvm-instr))
                              port)))))
           (case (LABEL-type gvm-instr)
             ((SIMPLE)
              len)
             ((ENTRY)
              (if (LABEL-ENTRY-closed? gvm-instr)
                (begin
                  (display " closure-entry-point " port)
                  (+ len (+ 21 (write-param-pattern gvm-instr port))))
                (begin
                  (display " entry-point " port)
                  (+ len (+ 13 (write-param-pattern gvm-instr port))))))
             ((RETURN)
              (display " return-point" port)
              (+ len 13))
             ((TASK-ENTRY)
              (display " task-entry-point" port)
              (+ len 17))
             ((TASK-RETURN)
              (display " task-return-point" port)
              (+ len 18))
             (else
              (compiler-internal-error
                "write-gvm-instr, unknown label type"))))))

      ((APPLY)
       (display "  " port)
       (let ((len (+ 2
                     (if (APPLY-loc gvm-instr)
                       (let ((len (write-gvm-opnd (APPLY-loc gvm-instr) port)))
                         (display " = " port)
                         (+ len 3))
                       0))))
         (+ len
            (write-prim-applic (APPLY-prim gvm-instr)
                               (APPLY-opnds gvm-instr)
                               port))))

      ((COPY)
       (display "  " port)
       (let ((len (+ 2 (write-gvm-opnd (COPY-loc gvm-instr) port))))
         (display " = " port)
         (+ len (+ 3 (write-gvm-opnd (COPY-opnd gvm-instr) port)))))

      ((CLOSE)
       (display "  close" port)
       (let ((len (+ 7 (write-closure-parms (car (CLOSE-parms gvm-instr))))))
         (let loop ((l (cdr (CLOSE-parms gvm-instr))) (len len))
           (if (pair? l)
             (let ((x (car l)))
               (display "," port)
               (loop (cdr l) (+ len (+ 1 (write-closure-parms x)))))
             len))))

      ((IFJUMP)
       (display "  if " port)
       (let ((len (+ 5
                     (write-prim-applic (IFJUMP-test gvm-instr)
                                        (IFJUMP-opnds gvm-instr)
                                        port))))
         (let ((len (+ len
                       (if (IFJUMP-poll? gvm-instr)
                         (begin (display " jump* " port) 7)
                         (begin (display " jump " port) 6)))))
           (let ((len (+ len
                         (write-returning-len
                           (frame-size (gvm-instr-frame gvm-instr))
                           port))))
             (display " " port)
             (let ((len (+ len
                           (+ 1 (write-gvm-lbl
                                  (IFJUMP-true gvm-instr)
                                  port)))))
               (display " else " port)
               (+ len (+ 6 (write-gvm-lbl
                             (IFJUMP-false gvm-instr)
                             port))))))))

      ((JUMP)
       (display "  " port)
       (let ((len (+ 2
                     (if (JUMP-poll? gvm-instr)
                       (begin (display "jump*" port) 5)
                       (begin (display "jump" port) 4)))))
         (let ((len (+ len
                       (if (JUMP-safe? gvm-instr)
                         (begin (display "$ " port) 2)
                         (begin (display " " port) 1)))))
           (let ((len (+ len
                         (write-returning-len
                           (frame-size (gvm-instr-frame gvm-instr))
                           port))))
             (display " " port)
             (let ((len (+ len
                           (+ 1 (write-gvm-opnd (JUMP-opnd gvm-instr) port)))))
               (+ len
                  (if (JUMP-nb-args gvm-instr)
                    (begin
                      (display " " port)
                      (+ 1 (write-returning-len
                             (JUMP-nb-args gvm-instr)
                             port)))
                    0)))))))

      (else
       (compiler-internal-error
         "write-gvm-instr, unknown 'gvm-instr':"
         gvm-instr))))

  (define (spaces n)
    (if (> n 0)
      (if (> n 7)
        (begin (display "        " port) (spaces (- n 8)))
        (begin (display " " port) (spaces (- n 1))))))

  (let ((len (write-instr gvm-instr)))
    (spaces (- 40 len))
    (display " " port)
    (write-frame (gvm-instr-frame gvm-instr) port))

  (let ((x (gvm-instr-comment gvm-instr)))
    (if x
      (let ((y (comment-get x 'TEXT)))
        (if y
          (begin
            (display " ; " port)
            (display y port)))))))

(define (write-frame frame port)

  (define (write-var var opnd sep)
    (display sep port)
    (write-gvm-opnd opnd port)
    (if var
      (begin
        (display "=" port)
        (cond ((eq? var closure-env-var)
               (write (map (lambda (var) (var-name var)) (frame-closed frame))
                      port))
              ((eq? var ret-var)
               (display "#" port))
              ((temp-var? var)
               (display "." port))
              (else
               (write (var-name var) port))))))

  (define (live? var)
    (let ((live (frame-live frame)))
      (or (set-member? var live)
          (and (eq? var closure-env-var)
               (not (set-empty? (set-intersection
                                  live
                                  (list->set (frame-closed frame)))))))))

  (let loop1 ((i 1) (l (reverse (frame-slots frame))) (sep "; "))
    (if (pair? l)
      (let ((var (car l)))
        (write-var (if (live? var) var #f) (make-stk i) sep)
        (loop1 (+ i 1) (cdr l) " "))
      (let loop2 ((i 0) (l (frame-regs frame)) (sep sep))
        (if (pair? l)
          (let ((var (car l)))
            (if (live? var)
              (begin
                (write-var var (make-reg i) sep)
                (loop2 (+ i 1) (cdr l) " "))
              (loop2 (+ i 1) (cdr l) sep))))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Operand writing:
; ---------------

(define (write-gvm-opnd gvm-opnd port)

  (define (write-opnd)
    (cond ((reg? gvm-opnd)
           (display "+" port)
           (+ 1 (write-returning-len (reg-num gvm-opnd) port)))
          ((stk? gvm-opnd)
           (display "-" port)
           (+ 1 (write-returning-len (stk-num gvm-opnd) port)))
          ((glo? gvm-opnd)
           (write-returning-len (glo-name gvm-opnd) port))
          ((clo? gvm-opnd)
           (let ((len (write-gvm-opnd (clo-base gvm-opnd) port)))
             (display "(" port)
             (let ((len (+ len
                           (+ 1 (write-returning-len
                                  (clo-index gvm-opnd)
                                  port)))))
               (display ")" port)
               (+ len 1))))
          ((lbl? gvm-opnd)
           (write-gvm-lbl (lbl-num gvm-opnd) port))
          ((obj? gvm-opnd)
           (display "'" port)
           (+ (write-gvm-obj (obj-val gvm-opnd) port) 1))
          (else
           (compiler-internal-error
             "write-gvm-opnd, unknown 'gvm-opnd':"
             gvm-opnd))))

  (write-opnd))

(define (write-gvm-lbl lbl port)
  (display "#" port)
  (+ (write-returning-len lbl port) 1))

(define (write-gvm-obj val port)
  (cond ((false-object? val)
         (display "#f" port)
         2)
        ((undef-object? val)
         (display "#[undefined]" port)
         12)
        ((proc-obj? val)
         (if (proc-obj-primitive? val)
           (display "#[primitive " port)
           (display "#[procedure " port))
         (let ((len (display-returning-len (proc-obj-name val) port)))
           (display "]" port)
           (+ len 13)))
        (else
         (write-returning-len val port))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (virtual.begin!) ; initialize module
  (set! *opnd-table* (make-vector opnd-table-size))
  (set! *opnd-table-alloc* 0)
  '())

(define (virtual.end!) ; finalize module
  (set! *opnd-table* '())
  '())

;==============================================================================
