;==============================================================================

; file: "_ptree2.scm"

; Copyright (C) 1994, 1995, 1996 by Marc Feeley, All Rights Reserved.

(include "fixnum.scm")

;------------------------------------------------------------------------------
;
; Parse tree manipulation module: (part 2)
; ------------------------------

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (normalize-program lst1)
  (let ((lst2 (beta-reduce (assignment-convert lst1))))
    (for-each lambda-lift! lst2)
    lst2))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; (delete-ptree ptree) removes parse tree 'ptree' from program and updates
; references and assignments to variables.

(define (delete-ptree ptree)

  (cond ((ref? ptree)
         (let ((var (ref-var ptree)))
           (var-refs-set! var (set-remove (var-refs var) ptree))))

        ((set? ptree)
         (let ((var (set-var ptree)))
           (var-sets-set! var (set-remove (var-sets var) ptree))))

        ((def? ptree)
         (let ((var (def-var ptree)))
           (var-sets-set! var (set-remove (var-sets var) ptree)))))

  (for-each delete-ptree (node-children ptree)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Assignment conversion:
; ---------------------

; (assignment-convert lst) takes a list of pairs of parse-trees and
; environments and returns a similar list where each parse-tree has
; been replaced with an equivalent parse-tree containing no assignments
; to non-global variables.  In the converted parse-tree, 'cells' are used
; to implement mutable variables and calls to the procedures:
;
;   ##MAKE-CELL, ##CELL-REF, ##CELL-SET!
;
; are added to create and access the cells.

(define (assignment-convert lst)
  (map (lambda (x)
         (let ((ptree (car x)) (env (cdr x)))
           (cons (ac ptree (env-declare env (list SAFE-sym #f)) '())
                 env)))
       lst))

(define (ac ptree env mut)

  (cond ((cst? ptree)
         ptree)

        ((ref? ptree)
         (let ((var (ref-var ptree)))
           (if (global? var)
             ptree
             (let ((x (assq var mut)))
               (if x
                 (let ((source (node-source ptree)))
                   (var-refs-set! var (set-remove (var-refs var) ptree))
                   (new-call source (node-decl ptree)
                     (new-ref-extended-bindings source **CELL-REF-sym env)
                     (list (new-ref source (node-decl ptree) (cdr x)))))
                 ptree)))))

        ((set? ptree)
         (let ((var (set-var ptree))
               (source (node-source ptree))
               (val (ac (set-val ptree) env mut)))
           (if (global? var)
             (new-set source (node-decl ptree)
               var
               val)
             (new-call source (node-decl ptree)
               (new-ref-extended-bindings source **CELL-SET!-sym env)
               (list (new-ref source (node-decl ptree) (cdr (assq var mut)))
                     val)))))

        ((def? ptree) ; guaranteed to be a toplevel definition
         (let ((var (def-var ptree)))
           (var-sets-set! var (set-remove (var-sets var) ptree))
           (new-def (node-source ptree) (node-decl ptree)
             var
             (ac (def-val ptree) env mut))))

        ((tst? ptree)
         (new-tst (node-source ptree) (node-decl ptree)
           (ac (tst-pre ptree) env mut)
           (ac (tst-con ptree) env mut)
           (ac (tst-alt ptree) env mut)))

        ((conj? ptree)
         (new-conj (node-source ptree) (node-decl ptree)
           (ac (conj-pre ptree) env mut)
           (ac (conj-alt ptree) env mut)))

        ((disj? ptree)
         (new-disj (node-source ptree) (node-decl ptree)
           (ac (disj-pre ptree) env mut)
           (ac (disj-alt ptree) env mut)))

        ((prc? ptree)
         (ac-proc ptree env mut))

        ((app? ptree)
         (let ((oper (app-oper ptree))
               (args (app-args ptree)))
           (if (and (prc? oper) ; applying a lambda-expr is like a 'let'
                    (not (prc-rest oper))
                    (= (length (prc-parms oper)) (length args)))
             (ac-let ptree env mut)
             (new-call (node-source ptree) (node-decl ptree)
               (ac oper env mut)
               (map (lambda (x) (ac x env mut)) args)))))

        ((fut? ptree)
         (new-fut (node-source ptree) (node-decl ptree)
           (ac (fut-val ptree) env mut)))

        (else
         (compiler-internal-error "ac, unknown parse tree node type"))))

(define (ac-proc ptree env mut)
  (let* ((mut-parms (ac-mutables (prc-parms ptree)))
         (cloned-mut-parms (clone-vars mut-parms)))

    (for-each (lambda (var) (var-sets-set! var (set-empty)))
              mut-parms)

    (new-prc (node-source ptree) (node-decl ptree)
      (prc-name ptree)
      (prc-min ptree)
      (prc-rest ptree)
      (prc-parms ptree)
      (new-let ptree
               ptree
               cloned-mut-parms
               (map (lambda (var)
                      (new-call (var-source var) (node-decl ptree)
                        (new-ref-extended-bindings (var-source var) **MAKE-CELL-sym env)
                        (list (new-ref (var-source var) (node-decl ptree) var))))
                    mut-parms)
               (ac (prc-body ptree)
                   env
                   (append (pair-up mut-parms cloned-mut-parms) mut))))))

(define (ac-let ptree env mut)
  (let* ((proc (app-oper ptree))
         (vals (app-args ptree))
         (vars (prc-parms proc))
         (vals-fv (apply set-union (map free-variables vals)))
         (mut-parms (ac-mutables vars))
         (cloned-mut-parms (clone-vars mut-parms))
         (mut (append (pair-up mut-parms cloned-mut-parms) mut)))

    (for-each (lambda (var) (var-sets-set! var (set-empty)))
              mut-parms)

    (let loop ((l1 vars)
               (l2 vals)
               (new-vars '())
               (new-vals '())
               (new-body (ac (prc-body proc) env mut)))
      (if (null? l1)

        (new-let ptree proc new-vars new-vals new-body)

        (let ((var (car l1))
              (val (car l2)))

          (if (memq var mut-parms)

            (let ((src (node-source val))
                  (decl (node-decl val))
                  (var* (cdr (assq var mut))))

              (if (set-member? var vals-fv)

                (loop (cdr l1)
                      (cdr l2)
                      (cons var* new-vars)
                      (cons (new-call src decl
                              (new-ref-extended-bindings src **MAKE-CELL-sym env)
                              (list (new-cst src decl undef-object)))
                            new-vals)
                      (new-seq src decl
                        (new-call src decl
                          (new-ref-extended-bindings src **CELL-SET!-sym env)
                          (list (new-ref src decl var*)
                                (ac val env mut)))
                        new-body))

                (loop (cdr l1)
                      (cdr l2)
                      (cons var* new-vars)
                      (cons (new-call src decl
                              (new-ref-extended-bindings src **MAKE-CELL-sym env)
                              (list (ac val env mut)))
                            new-vals)
                      new-body)))

            (loop (cdr l1)
                  (cdr l2)
                  (cons var new-vars)
                  (cons (ac val env mut) new-vals)
                  new-body)))))))

(define (ac-mutables lst)
  (set->list (set-keep mutable? (list->set lst))))

(define (clone-vars vars)
  (map var-copy vars))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; (clone-ptree ptree) returns a fresh copy of 'ptree'.  All the
; bound variables (i.e. not free variables) in ptree are also copied.
; It is assumed that 'ptree' has already been assignment-converted.

(define (clone-ptree ptree)
  (cp ptree '()))

(define (cp ptree substs)

  (define (rename-var var)
    (let ((x (assq var substs)))
      (if x (cdr x) var)))

  (cond ((cst? ptree)
         (new-cst (node-source ptree) (node-decl ptree)
           (cst-val ptree)))

        ((ref? ptree)
         (let ((var (rename-var (ref-var ptree))))
           (new-ref (node-source ptree) (node-decl ptree)
             var)))

        ((set? ptree)
         (let ((var (rename-var (set-var ptree))))
           (new-set (node-source ptree) (node-decl ptree)
             var
             (cp (set-val ptree) substs))))

        ((def? ptree) ; guaranteed to be a toplevel definition
         (new-def (node-source ptree) (node-decl ptree)
           (def-var ptree)
           (cp (def-val ptree) substs)))

        ((tst? ptree)
         (new-tst (node-source ptree) (node-decl ptree)
           (cp (tst-pre ptree) substs)
           (cp (tst-con ptree) substs)
           (cp (tst-alt ptree) substs)))

        ((conj? ptree)
         (new-conj (node-source ptree) (node-decl ptree)
           (cp (conj-pre ptree) substs)
           (cp (conj-alt ptree) substs)))

        ((disj? ptree)
         (new-disj (node-source ptree) (node-decl ptree)
           (cp (disj-pre ptree) substs)
           (cp (disj-alt ptree) substs)))

        ((prc? ptree)
         (let* ((parms (prc-parms ptree))
                (vars (clone-vars parms)))
           (new-prc (node-source ptree) (node-decl ptree)
             (prc-name ptree)
             (prc-min ptree)
             (prc-rest ptree)
             vars
             (cp (prc-body ptree)
                 (append (pair-up parms vars) substs)))))

        ((app? ptree)
         (let ((oper (app-oper ptree))
               (args (app-args ptree)))
           (if (and (prc? oper) ; applying a lambda-expr is like a 'let'
                    (not (prc-rest oper))
                    (= (length (prc-parms oper)) (length args)))
             (let* ((parms (prc-parms oper))
                    (vars (clone-vars parms))
                    (new-substs (append (pair-up parms vars) substs)))
               (new-let ptree
                        oper
                        vars
                        (map (lambda (x) (cp x new-substs)) args)
                        (cp (prc-body oper) new-substs)))
             (new-call (node-source ptree) (node-decl ptree)
               (cp (app-oper ptree) substs)
               (map (lambda (x) (cp x substs)) (app-args ptree))))))

        ((fut? ptree)
         (new-fut (node-source ptree) (node-decl ptree)
           (cp (fut-val ptree) substs)))

        (else
         (compiler-internal-error "cp, unknown parse tree node type"))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Beta-reduction:
; --------------

; (beta-reduce lst) takes a list of pairs of parse-trees and environments
; and returns a list of the parse-trees transformed with the following
; transformations:
;
;  - constant propagation
;  - copy propagation
;  - useless variable elimination
;
; It is assumed that the parse-trees have already been assignment-converted.

(define (beta-reduce lst)

  (define (process lst)
    (if (null? lst)
      '()
      (let ((x (car lst)))
        (let ((ptree (car x)) (env (cdr x)))
          (if (def? ptree)
           (let ((var (def-var ptree))
                 (val (br (def-val ptree) env '() #f)))
             (let ((rest (process (cdr lst))))
               (var-sets-set! var (set-remove (var-sets var) ptree))
               (cons (new-def (node-source ptree) (node-decl ptree)
                       var
                       val)
                     rest)))
           (let ((val (br ptree env '() #f)))
             (let ((rest (process (cdr lst))))
               (cons val rest))))))))

  (process lst))

(define (br ptree env substs expansion-limit)

  (cond ((cst? ptree)
         (new-cst (node-source ptree) (node-decl ptree)
           (cst-val ptree)))

        ((ref? ptree)
         (let ((var (ref-var ptree)))
           (var-refs-set! var (set-remove (var-refs var) ptree))
           (let ((new-var (var-subst var substs)))
             (let ((x (var-to-val new-var substs)))
               (if (and x (or (cst? x) (ref? x)))
                 (clone-ptree x)
                 (new-ref (node-source ptree) (node-decl ptree)
                   new-var))))))

        ((set? ptree) ; variable guaranteed to be a global variable
         (let ((var (set-var ptree)))
           (var-sets-set! var (set-remove (var-sets var) ptree))
           (new-set (node-source ptree) (node-decl ptree)
             var
             (br (set-val ptree) env substs expansion-limit))))

        ((tst? ptree)
         (new-tst (node-source ptree) (node-decl ptree)
           (br (tst-pre ptree) env substs expansion-limit)
           (br (tst-con ptree) env substs expansion-limit)
           (br (tst-alt ptree) env substs expansion-limit)))

        ((conj? ptree)
         (new-conj (node-source ptree) (node-decl ptree)
           (br (conj-pre ptree) env substs expansion-limit)
           (br (conj-alt ptree) env substs expansion-limit)))

        ((disj? ptree)
         (new-disj (node-source ptree) (node-decl ptree)
           (br (disj-pre ptree) env substs expansion-limit)
           (br (disj-alt ptree) env substs expansion-limit)))

        ((prc? ptree)
         (new-prc (node-source ptree) (node-decl ptree)
           (prc-name ptree)
           (prc-min ptree)
           (prc-rest ptree)
           (prc-parms ptree)
           (br (prc-body ptree) env substs expansion-limit)))

        ((app? ptree)
         (let ((oper (app-oper ptree))
               (args (app-args ptree)))
           (if (ref? oper)
             (let ((br-oper (br oper env substs expansion-limit)))
               ; at this point (or (cst? br-oper) (ref? br-oper))
               (or (br-inline-user ptree br-oper args env substs
                                   expansion-limit)
                   (br-app ptree br-oper args env substs expansion-limit)))
             (br-app ptree oper args env substs expansion-limit))))

        ((fut? ptree)
         (new-fut (node-source ptree) (node-decl ptree)
           (br (fut-val ptree) env substs expansion-limit)))

        (else
         (compiler-internal-error "br, unknown parse tree node type"))))

(define (var-subst var substs)
  (if (null? substs)
    var
    (let ((couple (car substs)))
      (if (eq? (car couple) var)
        (if (ref? (cdr couple))
          (var-subst (ref-var (cdr couple)) (cdr substs))
          var)
        (var-subst var (cdr substs))))))

(define (var-to-val var substs)
  (if (global? var)
    (global-single-def var)
    (let ((x (assq var substs)))
      (if x (cdr x) #f))))

(define (br-app ptree oper args env substs expansion-limit)
  (if (and (prc? oper) ; applying a lambda-expr is like a 'let'
           (not (prc-rest oper))
           (= (length (prc-parms oper)) (length args)))
    (br-let ptree oper args env substs expansion-limit)
    (new-call (node-source ptree) (node-decl ptree)
      (br oper env substs expansion-limit)
      (map (lambda (arg) (br arg env substs expansion-limit)) args))))

(define (br-let ptree proc vals env substs expansion-limit)
  (let* ((vars
          (prc-parms proc))
         (new-substs
          (br-extend-substs vars vals substs))
         (br-vals
          (map (lambda (x) (br x env new-substs expansion-limit)) vals))
         (new-body
          (br (prc-body proc) env new-substs expansion-limit)))

    ; remove useless bindings

    (let loop ((l1 vars)
               (l2 br-vals)
               (new-vars '())
               (new-vals '()))
      (if (null? l1)
        (new-let ptree
                 proc
                 (reverse new-vars)
                 (reverse new-vals)
                 new-body)
        (let ((var (car l1))
              (br-val (car l2)))
          (if (and (set-empty? (var-refs var))
                   (or (cst? br-val)
                       (ref? br-val)
                       (prc? br-val)))
            (begin
              (delete-ptree br-val)
              (loop (cdr l1)
                    (cdr l2)
                    new-vars
                    new-vals))
            (loop (cdr l1)
                  (cdr l2)
                  (cons var new-vars)
                  (cons br-val new-vals))))))))

(define (br-extend-substs vars vals substs)
  (let loop ((l1 vars)
             (l2 vals)
             (new-substs substs))
    (if (null? l1)
      new-substs
      (let ((var (car l1))
            (val (car l2)))
        (cond ((or (cst? val)
                   (and (ref? val)
                        (or (bound? (ref-var val))
                            (global-singly-bound? val)))
                   (and (prc? val)
                        (set-every? oper-pos? (var-refs var))))
               (loop (cdr l1)
                     (cdr l2)
                     (cons (cons var val) new-substs)))
              (else
               (loop (cdr l1)
                     (cdr l2)
                     new-substs)))))))

(define (br-inline-user ptree br-oper args env substs expansion-limit)
  (and (ref? br-oper)
       (let* ((var (ref-var br-oper))
              (val (var-to-val var substs)))
         (and val
              (prc? val)
              (inline? (node-decl val))
              (let* ((size-val
                       (ptree-size val))
                     (size-ptree
                       (ptree-size ptree))
                     (new-limit
                       (- (if expansion-limit
                            (car expansion-limit)
                            (quotient (* (inlining-limit (node-decl ptree))
                                         size-ptree)
                                      100))
                          (- size-val 1))))
                (and (>= new-limit 0)
                     (let ((cloned-oper (clone-ptree val)))
                       (delete-ptree br-oper)
                       (br-app ptree cloned-oper args env substs
                               (if expansion-limit
                                 (begin
                                   (set-car! expansion-limit new-limit)
                                   expansion-limit)
                                 (list new-limit))))))))))

(define (ptree-size ptree)
  (let loop ((lst (node-children ptree)) (n 1))
    (if (null? lst)
      n
      (loop (cdr lst) (+ n (ptree-size (car lst)))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Lambda-lifting procedure:
; ------------------------

; (lambda-lift! ptree) modifies the parse-tree 'ptree' so that some
; of its procedures (i.e. lambda-expressions) are replaced with
; weaker ones (i.e. lambda-expressions having fewer or no closed variables).
; It is assumed that 'ptree' has already been assignment-converted.
; Presently, only named procedures are lambda-lifted.

(define (lambda-lift! ptree)
  (ll! ptree (set-empty) '()))

(define (ll! ptree cst-procs env)

  (define (new-env env vars)
    (define (loop i l)
      (if (pair? l)
        (let ((var (car l)))
          (cons (cons var (cons (length (set->list (var-refs var))) i))
                (loop (+ i 1) (cdr l))))
        env))
    (loop (length env) vars))

  (cond ((or (cst? ptree)
             (ref? ptree)
             (set? ptree)
             (def? ptree) ; guaranteed to be a toplevel definition
             (tst? ptree)
             (conj? ptree)
             (disj? ptree)
             (fut? ptree))
         (for-each (lambda (child) (ll! child cst-procs env))
                   (node-children ptree)))

        ((prc? ptree)
         (ll! (prc-body ptree) cst-procs (new-env env (prc-parms ptree))))

        ((app? ptree)
         (let ((oper (app-oper ptree))
               (args (app-args ptree)))
           (if (and (prc? oper) ; applying a lambda-expr is like a 'let'
                    (not (prc-rest oper))
                    (= (length (prc-parms oper)) (length args)))
             (ll!-let ptree cst-procs (new-env env (prc-parms oper)))
             (for-each (lambda (child) (ll! child cst-procs env))
                       (node-children ptree)))))

        (else
         (compiler-internal-error "ll!, unknown parse tree node type"))))

(define (ll!-let ptree cst-procs env)
  (let* ((proc (app-oper ptree))
         (vals (app-args ptree))
         (vars (prc-parms proc))
         (var-val-map (pair-up vars vals)))

    (define (var->val var) (cdr (assq var var-val-map)))

    (define (liftable-proc-vars vars)
      (let loop ((cst-proc-vars
                   (set-keep (lambda (var)
                               (let ((val (var->val var)))
                                 (and (prc? val)
                                      (lambda-lift? (node-decl val))
                                      (set-every? oper-pos? (var-refs var)))))
                             (list->set vars))))
        (let* ((non-cst-proc-vars
                 (set-keep (lambda (var)
                             (let ((val (var->val var)))
                               (and (prc? val)
                                    (not (set-member? var cst-proc-vars)))))
                           (list->set vars)))
               (cst-proc-vars*
                 (set-keep (lambda (var)
                             (let ((val (var->val var)))
                               (set-empty?
                                 (set-intersection (free-variables val)
                                                   non-cst-proc-vars))))
                           cst-proc-vars)))
          (if (set-equal? cst-proc-vars cst-proc-vars*)
            cst-proc-vars
            (loop cst-proc-vars*)))))

    (define (transitively-closed-free-variables vars)
      (let ((tcfv-map
              (map (lambda (var) (cons var (free-variables (var->val var))))
                   vars)))
        (let loop ((changed? #f))
          (for-each (lambda (var-tcfv)
                      (let loop2 ((l (set->list (cdr var-tcfv)))
                                  (fv (cdr var-tcfv)))
                        (if (null? l)
                          (if (not (set-equal? fv (cdr var-tcfv)))
                            (begin
                              (set-cdr! var-tcfv fv)
                              (set! changed? #t)))
                          (let ((x (assq (car l) tcfv-map)))
                            (loop2 (cdr l)
                                   (if x (set-union fv (cdr x)) fv))))))
                    tcfv-map)
          (if changed?
            (loop #f)
            tcfv-map))))

    (let* ((tcfv-map
             (transitively-closed-free-variables (liftable-proc-vars vars)))
           (cst-proc-vars-list
             (map car tcfv-map))
           (cst-procs*
             (set-union (list->set cst-proc-vars-list) cst-procs)))

      (define (var->tcfv var) (cdr (assq var tcfv-map)))

      (define (order-vars vars)
        (map car
             (sort-list (map (lambda (var) (assq var env)) vars)
                        (lambda (x y)
                          (if (= (cadr x) (cadr y))
                            (< (cddr x) (cddr y))
                            (< (cadr x) (cadr y)))))))

      (define (lifted-vars var)
        (order-vars (set->list (set-difference (var->tcfv var) cst-procs*))))

      (define (lift-app! var)
        (let* ((val (var->val var))
               (vars (lifted-vars var)))

          (define (new-ref* var)
            (new-ref (var-source var) (node-decl val) var))

          (if (not (null? vars))
            (for-each (lambda (oper)
                        (let ((node (node-parent oper)))
                          (node-children-set! node
                            (cons (app-oper node)
                                  (append (map new-ref* vars)
                                          (app-args node))))))
                      (set->list (var-refs var))))))

      (define (lift-prc! var)
        (let* ((val (var->val var))
               (vars (lifted-vars var)))
          (if (not (null? vars))
            (let ((cloned-vars (clone-vars vars)))
              (prc-parms-set! val (append cloned-vars (prc-parms val)))
              (for-each (lambda (x) (var-bound-set! x val)) cloned-vars)
              (prc-min-set! val (+ (prc-min val) (length vars)))
              (node-fv-invalidate! val)
              (ll-rename! val (pair-up vars cloned-vars))))))

      (for-each lift-app! cst-proc-vars-list)
      (for-each lift-prc! cst-proc-vars-list)
      (for-each (lambda (node) (ll! node cst-procs* env)) vals)
      (ll! (prc-body proc) cst-procs* env))))

(define (ll-rename! ptree var-map)

  (if (ref? ptree)
    (let* ((var (ref-var ptree))
           (x (assq var var-map)))
      (if x
        (begin
          (var-refs-set! var (set-remove (var-refs var) ptree))
          (var-refs-set! (cdr x) (set-adjoin (var-refs (cdr x)) ptree))
          (ref-var-set! ptree (cdr x))))))

  (node-fv-set! ptree #t)

  (for-each (lambda (child) (ll-rename! child var-map))
            (node-children ptree)))

;------------------------------------------------------------------------------
;
; Debugging stuff:

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; (parse-tree->expression ptree) returns the Scheme expression corresponding to
; the parse tree 'ptree'.

(define (parse-tree->expression ptree)
  (se ptree '() (list 0)))

(define (se ptree env num)

  (cond ((cst? ptree)
         (list QUOTE-sym (cst-val ptree)))

        ((ref? ptree)
         (se-var->id (ref-var ptree) env))

        ((set? ptree)
         (list SET!-sym
               (se-var->id (set-var ptree) env)
               (se (set-val ptree) env num)))

        ((def? ptree)
         (list DEFINE-sym
               (se-var->id (def-var ptree) env)
               (se (def-val ptree) env num)))

        ((tst? ptree)
         (list IF-sym (se (tst-pre ptree) env num)
                      (se (tst-con ptree) env num)
                      (se (tst-alt ptree) env num)))

        ((conj? ptree)
         (list AND-sym (se (conj-pre ptree) env num)
                       (se (conj-alt ptree) env num)))

        ((disj? ptree)
         (list OR-sym (se (disj-pre ptree) env num)
                      (se (disj-alt ptree) env num)))

        ((prc? ptree)
         (let ((new-env (se-rename (prc-parms ptree) env num)))
           (list LAMBDA-sym
             (se-parameters (prc-parms ptree)
                            (prc-rest ptree)
                            (prc-min ptree)
                            new-env)
             (se (prc-body ptree) new-env num))))

        ((app? ptree)
         (let ((oper (app-oper ptree))
               (args (app-args ptree)))
           (if (and (prc? oper) ; applying a lambda-expr is like a 'let'
                    (not (prc-rest oper))
                    (= (length (prc-parms oper)) (length args)))
             (let ((new-env (se-rename (prc-parms oper) env num)))
               (list
                 (if (set-empty?
                       (set-intersection
                         (list->set (prc-parms oper))
                         (apply set-union (map free-variables args))))
                   LET-sym
                   LETREC-sym)
                 (se-bindings (prc-parms oper) args new-env num)
                 (se (prc-body oper) new-env num)))
             (map (lambda (x) (se x env num)) (cons oper args)))))

        ((fut? ptree)
         (list FUTURE-sym (se (fut-val ptree) env num)))

        (else
         (compiler-internal-error "se, unknown parse tree node type"))))

(define (se-var->id var env)
  (let ((x (assq var env)))
    (if x (cdr x) (var-name var))))

(define (se-parameters parms rest min env)

  (define (se-parms parms rest n env)
    (cond ((null? parms)
           '())
          ((and rest (null? (cdr parms)))
           (se-var->id (car parms) env))
          (else
           (let ((parm (se-var->id (car parms) env)))
             (cons (if (> n 0) parm (list parm))
                   (se-parms (cdr parms) rest (- n 1) env))))))

  (se-parms parms rest min env))

(define (se-bindings vars vals env num)
  (if (null? vars)
    '()
    (cons (list (se-var->id (car vars) env) (se (car vals) env num))
          (se-bindings (cdr vars) (cdr vals) env num))))

(define (se-rename vars env num)

  (define (rename vars)
    (if (null? vars)
      env
      (cons (cons (car vars)
                  (string->canonical-symbol
                    (string-append (symbol->string (var-name (car vars)))
                                   "#"
                                   (number->string (car num)))))
            (rename (cdr vars)))))

  (set-car! num (+ (car num) 1))
  (rename vars))

;==============================================================================
