;==============================================================================

; file: "_t-c-1.scm"

; Copyright (C) 1994-1996 by Marc Feeley, All Rights Reserved.

(include "fixnum.scm")

;------------------------------------------------------------------------------
;
; Back end for C language (part 1)
; -----------------------

; Back end parameters:

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; ***** REGISTER USAGE

(define targ-nb-gvm-regs 5) ; Number of registers in the virtual machine.
(define targ-nb-arg-regs 3) ; Number of registers used to pass arguments.

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; ***** SUPPORTED WORD SIZES (in bytes per word)

; It is assumed that the target machine's word size (in bytes)
; is a power of 2.  The minimal supported word size must be known to
; decide if an integer should be represented with a fixnum or bignum.  The
; largest supported word size is needed in order to generate portable macros
; to pack strings and bignums into words.

(define targ-min-word-size 4) ; At least 4 bytes per word
(define targ-max-word-size 8) ; Up to 8 bytes per word

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; ***** DATA REPRESENTATION

; Note: The data representation is mostly machine independent.  This means
; that it is possible to change the format of objects by redefining the
; appropriate macros in "gambit.h".  However, the back end must know
; a few things about the representation in order for the output to be
; portable to different word sizes and to perform a few optimizations.
; In the following definitions, space is expressed in number of words.

; Lower bound on free space in heap at the beginning of basic blocks:

(define targ-heap-fudge 1024)

; Number of tag bits per pointer:

(define targ-tag-bits 2)
(define targ-alignment (expt 2 targ-tag-bits))

; Upper bound on space needed by various objects:

(define (targ-max-words i) ; minimum k such that targ-min-word-size*k >= i
                           ; and targ-min-word-size*k mod targ-alignment = 0

  (define (round-up-to-multiple-of n i)
    (* (quotient (+ i (- n 1)) n) n))

  (quotient (round-up-to-multiple-of
              targ-min-word-size
              (round-up-to-multiple-of targ-alignment i))
            targ-min-word-size))

; Space occupied by various types of objects.

(define targ-pair-space        (targ-max-words (* 3 targ-min-word-size)))
(define targ-cell-space        (targ-max-words (* 2 targ-min-word-size)))
(define targ-will-space        (targ-max-words (* 4 targ-min-word-size)))
(define targ-flonum-space      (targ-max-words 16))
(define targ-placeholder-space (targ-max-words (* 3 targ-min-word-size)))
(define (targ-closure-space n) (targ-max-words (* (+ n 2) targ-min-word-size)))
(define (targ-string-space n)  (targ-max-words (+ (+ n 1) targ-min-word-size)))
(define (targ-vector8-space n) (targ-max-words (+ n targ-min-word-size)))
(define (targ-vector-space n)  (targ-max-words (* (+ n 1) targ-min-word-size)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; ***** PROCEDURE CALLING CONVENTION

(define (targ-label-info min-args nb-parms rest? closed?)

; After a LABEL instruction, the following is true:
;
;  * return address is in R0
;
;  * if nb-parms <= targ-nb-arg-regs,
;
;      then, parameter `n' is in Rn
;
;      else, the first `m' = nb-parms - targ-nb-arg-regs
;            parameters will be on the stack and, assuming `i' = n - m,
;            parameter `n' is in
;
;            Ri, if i > 0
;            or else in slot `frame_size + i' of the stack frame
;
;  * if `CLOSED' is present, register number `targ-nb-arg-regs + 1' contains
;    a pointer to the closure object
;
; for example, if we assume that targ-nb-arg-regs = 3, then after the
; instruction LABEL(1,2,PROC,5):
;
;   R0 = return address
;   S1 = parameter 1
;   S2 = parameter 2
;   R1 = parameter 3
;   R2 = parameter 4
;   R3 = parameter 5

  (let ((nb-stacked (max 0 (- nb-parms targ-nb-arg-regs))))

    (define (location-of-parms i)
      (if (> i nb-parms)
        '()
        (cons (cons i
                    (if (> i nb-stacked)
                      (make-reg (- i nb-stacked))
                      (make-stk i)))
              (location-of-parms (+ i 1)))))

    (let ((x (cons (cons 'return 0) (location-of-parms 1))))
      (make-pcontext nb-stacked
        (if closed?
          (cons (cons 'closure-env (make-reg (+ targ-nb-arg-regs 1))) x)
          x)))))

(define (targ-jump-info nb-args)

; After a JUMP instruction, the following is true:
;
;  * the return address is passed in R0
;
;  * if nb-args <= targ-nb-arg-regs,
;
;      then, argument `n' is in Rn
;
;      else, `m' = nb-args - targ-nb-arg-regs arguments are passed
;            on the stack and, assuming `i' = n - m, argument `n' is in
;
;            Ri, if i > 0
;            or else in slot `frame_size + i' of the stack frame

  (let ((nb-stacked (max 0 (- nb-args targ-nb-arg-regs))))

    (define (location-of-args i)
      (if (> i nb-args)
        '()
        (cons (cons i
                    (if (> i nb-stacked)
                      (make-reg (- i nb-stacked))
                      (make-stk i)))
              (location-of-args (+ i 1)))))

    (make-pcontext nb-stacked
                   (cons (cons 'return (make-reg 0))
                         (location-of-args 1)))))

;------------------------------------------------------------------------------
;
; Initialization/finalization of back-end

(let ((targ (make-target 4 'C)))

  (define (begin! info-port)

    (set! targ-info-port info-port)
    (set! targ-return-reg (make-reg 0))

    (target-end!-set!         targ end!)
    (target-dump-set!         targ targ-dump)
    (target-nb-regs-set!      targ targ-nb-gvm-regs)
    (target-prim-info-set!    targ targ-prim-info)
    (target-label-info-set!   targ targ-label-info)
    (target-jump-info-set!    targ targ-jump-info)
    (target-proc-result-set!  targ (make-reg 1))
    (target-task-return-set!  targ targ-return-reg)

    #f)

  (define (end!)
    #f)

  (target-begin!-set! targ begin!)
  (put-target targ))

(define targ-info-port #f)
(define targ-return-reg #f)

;------------------------------------------------------------------------------
;
; Primitive procedure database

(define targ-prim-proc-table
  (map (lambda (x)
         (cons (string->canonical-symbol (car x))
               (apply make-proc-obj (car x) #t #f (cdr x))))
       prim-procs))

(define (targ-prim-info name)
  (let ((x (assq name targ-prim-proc-table)))
    (if x (cdr x) #f)))

(define (targ-get-prim-info name)
  (let ((proc (targ-prim-info (string->canonical-symbol name))))
    (if proc
      proc
      (compiler-internal-error
        "targ-get-prim-info, unknown primitive:" name))))

;------------------------------------------------------------------------------
;
; Dumping of a compilation module

(define (targ-dump procs filename c-intf options)
  (set! targ-debug-info? (memq 'DEBUG options))
  (targ-heap-begin!)
  (targ-use-prc (car procs) #f)
  (for-each (lambda (proc) (targ-use-prc proc #t)) procs)
  (targ-heap-dump (proc-obj-name (car procs))
                  (string-append filename ".c")
                  c-intf)
  (targ-heap-end!)
  (set! targ-debug-info? #f))

(define targ-debug-info? #f)

;------------------------------------------------------------------------------
;
; Object management

; These routines are used to keep track of all the Scheme objects and
; global variables that the compilation module uses.  These objects and
; variables will have to be created before the module is run (either
; at compile time, link time or run time).

(define targ-glo-vars #f)  ; global variables used
(define targ-prc-objs #f)  ; procedure objects
(define targ-lbl-alloc #f) ; label table allocation pointer
(define targ-cns-objs #f)  ; pair objects
(define targ-cns-alloc #f) ; pair object allocation pointer
(define targ-sym-objs #f)  ; interned symbol objects
(define targ-sym-alloc #f) ; interned symbol object allocation pointer
(define targ-sub-objs #f)  ; objects that have a subtype (vector, ...)
(define targ-sub-alloc #f) ; allocation pointer for objects with a subtype
(define targ-prc-objs* #f) ; procedure objects remaining to scan

(define targ-module-decl   #f) ; declarations from "##c-decl" calls
(define targ-module-rd-res #f) ; set of resources read from
(define targ-module-wr-res #f) ; set of resources written to

(define targ-fp-cache #f) ; floating point number cache
(define targ-fr-cell #f)  ; cell of floating point region start

(define (targ-heap-begin!)
  (let ((nb-res (+ targ-nb-gvm-regs targ-nb-non-reg-res)))
    (set! targ-glo-vars      (queue-empty))
    (set! targ-prc-objs      (queue-empty))
    (set! targ-lbl-alloc     0)
    (set! targ-cns-objs      (queue-empty))
    (set! targ-cns-alloc     0)
    (set! targ-sym-objs      (queue-empty))
    (set! targ-sym-alloc     0)
    (set! targ-sub-objs      (queue-empty))
    (set! targ-sub-alloc     0)
    (set! targ-prc-objs*     (queue-empty))
    (set! targ-module-decl   (queue-empty))
    (set! targ-module-rd-res (make-vector nb-res #f))
    (set! targ-module-wr-res (make-vector nb-res #f))
    (set! targ-fp-cache #f)
    (set! targ-fr-cell #f)
    #f))

(define (targ-heap-end!)
  (set! targ-glo-vars #f)
  (set! targ-prc-objs #f)
  (set! targ-lbl-alloc #f)
  (set! targ-cns-objs #f)
  (set! targ-cns-alloc #f)
  (set! targ-sym-objs #f)
  (set! targ-sym-alloc #f)
  (set! targ-sub-objs #f)
  (set! targ-sub-alloc #f)
  (set! targ-prc-objs* #f)
  (set! targ-module-decl #f)
  (set! targ-module-rd-res #f)
  (set! targ-module-wr-res #f)
  (set! targ-fp-cache #f)
  (set! targ-fr-cell #f)
  #f)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-use-glo glo supply?)
  (let ((x (assq glo (queue->list targ-glo-vars))))
    (if x
      (begin
        (if supply? (set-cdr! (cddr x) #t) (set-car! (cddr x) #t))
        (cadr x))
      (let ((y (targ-make-cell #f)))
        (queue-put! targ-glo-vars
                    (cons glo (cons y (if supply? (cons #f #t) (cons #t #f)))))
        y))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-use-obj obj)
  (let ((type (targ-obj-type obj)))
    (case type
      ((PROCEDURE)
       (targ-use-prc obj #f))
      ((PAIR)
       (let ((x (assoc obj (queue->list targ-cns-objs))))
         (if x
           (cdr x)
           (let ((y (list "CNS" targ-cns-alloc)))
             (queue-put! targ-cns-objs (cons obj y))
             (set! targ-cns-alloc (+ targ-cns-alloc 1))
             (targ-use-obj (car obj))
             (targ-use-obj (cdr obj))
             y))))
      ((SUBTYPED)
       (let ((subtype (targ-obj-subtype obj)))
         (cond ((eq? subtype 'SYMBOL)
                (let ((x (assq obj (queue->list targ-sym-objs))))
                  (if x
                    (cdr x)
                    (let ((y (list "SYM" targ-sym-alloc)))
                      (queue-put! targ-sym-objs (cons obj y))
                      (set! targ-sym-alloc (+ targ-sym-alloc 1))
                      y))))
               (else
                (let ((x (assoc obj (queue->list targ-sub-objs))))
                  (if x
                    (cdr x)
                    (let ((y (list "SUB" targ-sub-alloc)))
                      (queue-put! targ-sub-objs (cons obj y))
                      (set! targ-sub-alloc (+ targ-sub-alloc 1))
                      (case subtype
                        ((VECTOR)
                         (for-each targ-use-obj (vector->lst obj)))
                        ((RATNUM)
                         (targ-use-obj (targ-numerator obj))
                         (targ-use-obj (targ-denominator obj)))
                        ((CPXNUM)
                         (targ-use-obj (targ-real-part obj))
                         (targ-use-obj (targ-imag-part obj))))
                      y)))))))
      ((BOOLEAN)
       (if (eq? obj #t) '("TRU") '("FAL")))
      ((NULL)
       '("NUL"))
      ((UNDEF)
       '("UND"))
      ((FIXNUM)
       (list "FIX" (targ-c-long obj)))
      ((CHAR)
       (list "CHR" (targ-c-char obj)))
      (else
       (compiler-internal-error
         "targ-use-obj, unknown object 'obj'" obj)))))

(define (targ-use-prc proc supply?)
  (let* ((name (proc-obj-name proc))
         (p (if (proc-obj-code proc)
              (let ((x (assq proc (queue->list targ-prc-objs))))
                (if x
                  (cddr x)
                  (let* ((y (list 'PRC (targ-make-cell #f)))
                         (z (cons proc (cons #f y))))
                    (queue-put! targ-prc-objs z)
                    (queue-put! targ-prc-objs* z)
                    y)))
              #f)))
    (if (proc-obj-primitive? proc)
      (let ((i (targ-use-glo (string->symbol name) supply?)))
        (if p
          p
          (list 'PRM i (targ-c-id-glo name))))
      p)))

(define (targ-heap-ref-obj obj)

  (define (err)
    (compiler-internal-error
      "targ-heap-ref-obj, object was not scanned 'obj'" obj))

  (let ((type (targ-obj-type obj)))
    (case type
      ((PAIR)
       (let ((x (assoc obj (queue->list targ-cns-objs))))
         (if x
           (cons "REF_CNS" (cddr x))
           (err))))
      ((SUBTYPED)
       (let ((subtype (targ-obj-subtype obj)))
         (cond ((eq? subtype 'SYMBOL)
                (let ((x (assq obj (queue->list targ-sym-objs))))
                  (if x
                    (cons "REF_SYM" (cddr x))
                    (err))))
               (else
                (let ((x (assoc obj (queue->list targ-sub-objs))))
                  (if x
                    (cons "REF_SUB" (cddr x))
                    (err)))))))
      ((BOOLEAN)
       (if (eq? obj #t) '("REF_TRU") '("REF_FAL")))
      ((NULL)
       '("REF_NUL"))
      ((UNDEF)
       '("REF_UND"))
      ((FIXNUM)
       (list "REF_FIX" obj))
      ((CHAR)
       (list "REF_CHR" (targ-c-char obj)))
      (else
       (compiler-internal-error
         "targ-heap-ref-obj, unknown object 'obj'" obj)))))

;------------------------------------------------------------------------------
;
; Dumping of objects

; These routines write out the C code that represents the Scheme objects
; contained in the compilation module.

(define (targ-heap-dump module-name filename c-intf)

  (let ((c-decls (c-intf-decls c-intf))
        (c-procs (c-intf-procs c-intf))
        (c-inits (c-intf-inits c-intf)))

    (c-intf-decls-set! c-intf '()) ; clear c-intf structure so front-end
    (c-intf-procs-set! c-intf '()) ; will not generate C-interface file.
    (c-intf-inits-set! c-intf '())

    ; scan and convert each procedure

    (if targ-info-port
      (begin
        (display "Dumping:" targ-info-port)
        (newline targ-info-port)))

    (let loop ()
      (if (not (queue-empty? targ-prc-objs*))
        (begin
          (targ-scan-procedure (queue-get! targ-prc-objs*))
          (loop))))

    (if targ-info-port
      (newline targ-info-port))

    ; write out C code

    (let* ((vars (targ-split-supplied (queue->list targ-glo-vars)))
           (supplied (car vars))
           (not-supplied (cdr vars))
           (glo-vars (append supplied not-supplied))
           (sym-list (queue->list targ-sym-objs))
           (cns-list (queue->list targ-cns-objs))
           (sub-list (queue->list targ-sub-objs))
           (prc-list (queue->list targ-prc-objs))
           (prm-list (targ-extract-primitives prc-list)))

      (set! targ-port (open-output-file filename))
      (set! targ-line-size 0)

      (targ-dump-header module-name filename)

      (targ-define-count "GLO_COUNT" (length glo-vars))
      (targ-define-count "SUP_COUNT" (length supplied))
      (targ-define-count "SYM_COUNT" (length sym-list))
      (targ-define-count "CNS_COUNT" (length cns-list))
      (targ-define-count "LBL_COUNT" targ-lbl-alloc)
      (targ-define-count "SUB_COUNT" targ-sub-alloc)

      (targ-display "#include \"gambit.h\"") (targ-line)

      (targ-dump-linking-info glo-vars sym-list)

      (targ-dump-sym sym-list)
      (targ-dump-cns cns-list)
      (targ-dump-sub sub-list)
      (targ-dump-prc prc-list c-decls)
      (targ-dump-ini prm-list c-inits)

      (close-output-port targ-port))))

(define (targ-split-supplied glo-vars)
  (let loop ((lst glo-vars) (supplied '()) (not-supplied '()))
    (if (null? lst)
      (cons (reverse supplied) (reverse not-supplied))
      (let* ((x (car lst))
             (supply? (cdddr x)))
        (if supply?
          (loop (cdr lst) (cons x supplied) not-supplied)
          (loop (cdr lst) supplied (cons x not-supplied)))))))

(define (targ-dump-header module-name filename)

  (targ-display "/* file: \"")
  (targ-display filename)
  (targ-display "\", produced by Gambit-C ")
  (targ-display compiler-version)
  (targ-display " */")
  (targ-line)

  (targ-display "#define ")
  (targ-code '("MODULE_NAME "))
  (targ-display "\"") ; ********* use targ-c-string ?
  (targ-display module-name)
  (targ-display "\"")
  (targ-line)

  (targ-display "#define ")
  (targ-code '("MODULE_LINKER "))
  (targ-code (targ-c-id-linker module-name))
  (targ-line)

  (targ-display "#define ")
  (targ-code '("VERSION "))
  (targ-display header-version)
  (targ-line)

  (targ-display "#define ")
  (targ-code '("MH_PROC "))
  (targ-code (targ-c-id-host module-name))
  (targ-line))

(define (targ-dump-linking-info glo-vars sym-objs)

  (targ-code* '("BEGIN_MODULE"))

  (for-each
    (lambda (x)
      (let ((name (symbol->string (car x)))
            (supply? (cdddr x))
            (demand? (caddr x)))
        (if supply?
          (targ-code* (list "SUPPLY_GLO" (targ-c-id-glo name))))
        (if demand?
          (targ-code* (list "DEMAND_GLO" (targ-c-id-glo name))))))
    glo-vars)

  (for-each
    (lambda (x)
      (let ((name (symbol->string (car x))))
        (targ-code* (list "DEMAND_SYM" (targ-c-id-sym name)))))
    sym-objs)

  (targ-line)

  (targ-dump-section "BEGIN_GLO" "END_GLO" #f glo-vars
    (lambda (i x)
      (targ-cell-set! (cadr x) i)
      (let ((name (symbol->string (car x))))
        (targ-code* (list "DEF_GLO" (targ-c-string name)))))))

(define (targ-define-count name n)
  (if (not (= n 0))
    (begin
      (targ-display "#define ")
      (targ-code (list name))
      (targ-display " ")
      (targ-display n)
      (targ-line))))

(define (targ-define-prefix var)
  (targ-display "#undef ")
  (targ-code var)
  (targ-line)
  (targ-display "#define ")
  (targ-code var))

(define (targ-dump-section begin-name end-name comma? lst dump)
  (if (not (null? lst))
    (begin
      (targ-code* (list begin-name))
      (if comma? (targ-display " "))
      (dump 0 (car lst))
      (let loop ((i 1) (lst (cdr lst)))
        (if (not (null? lst))
          (begin
            (if comma? (targ-display ","))
            (dump i (car lst))
            (loop (+ i 1) (cdr lst)))))
      (targ-code* (list end-name))
      (targ-line))))

(define (targ-dump-sym objs)
  (targ-dump-section "BEGIN_SYM" "END_SYM" #f objs
    (lambda (i x)
      (let ((name (symbol->string (car x))))
        (targ-code* (list "DEF_SYM"
                          (targ-c-id-sym name)
                          (targ-c-string name)))))))

(define (targ-dump-cns objs)
  (targ-dump-section "BEGIN_CNS" "END_CNS" #t objs
    (lambda (i x)
      (let ((obj (car x)) (i (caddr x)))
        (targ-code* (list "DEF_CNS"
                          (targ-heap-ref-obj (car obj))
                          (targ-heap-ref-obj (cdr obj))))))))

(define (targ-dump-sub objs)

  (define (vector-like-object def-name chunk-name id elems convert cycle)
    (let ((len (length elems)))
      (targ-code* (list def-name id len))
      (let loop1 ((i 0) (elems elems))
        (if (<= i len)
          (let ((n (min cycle (- len i))))
            (let loop2 ((j (- n 1)) (elems elems) (lst '()))
              (if (>= j 0)
                (loop2 (- j 1) (cdr elems) (cons (convert (car elems)) lst))
                (begin
                  (targ-display "               ")
                  (targ-code* (cons chunk-name (cons n (reverse lst))))
                  (if (= n cycle) (loop1 (+ i n) elems))))))))))

  (define (dump-VECTOR obj id)
    (vector-like-object "DEF_SUB_VEC" 'VEC id
                        (vector->lst obj)
                        targ-heap-ref-obj
                        1))

  (define (dump-STRING obj id)
    (vector-like-object "DEF_SUB_STR" 'STR id
                        (string->lst obj)
                        targ-c-char
                        targ-max-word-size))

  (define (dump-BIGNUM obj id)
    (vector-like-object "DEF_SUB_BIG" 'BIG id
                        (targ-BIGNUM-digits obj)
                        (lambda (x) x)
                        (quotient targ-max-word-size 2)))

  (define (dump-RATNUM obj id)
    (targ-code* (list "DEF_SUB_RAT" id
                      (targ-heap-ref-obj (targ-numerator obj))
                      (targ-heap-ref-obj (targ-denominator obj)))))

  (define (dump-FLONUM obj id)
    (let ((hi-lo-bits (targ-flonum->hi-lo-bits obj)))
      (targ-code* (list "DEF_SUB_FLO" id
                        (targ-c-hex-ulong (car hi-lo-bits))
                        (targ-c-hex-ulong (cdr hi-lo-bits))))))

  (define (dump-CPXNUM obj id)
    (targ-code* (list "DEF_SUB_CPX" id
                      (targ-heap-ref-obj (targ-real-part obj))
                      (targ-heap-ref-obj (targ-imag-part obj)))))

  (if (not (null? objs))
    (begin
      (for-each (lambda (x)
                  (let ((obj (car x)) (id (targ-sub-name (caddr x))))
                    (let ((subtype (targ-obj-subtype obj)))
                      (case subtype
                        ((VECTOR) (dump-VECTOR obj id))
                        ((STRING) (dump-STRING obj id))
                        ((BIGNUM) (dump-BIGNUM obj id))
                        ((RATNUM) (dump-RATNUM obj id))
                        ((FLONUM) (dump-FLONUM obj id))
                        ((CPXNUM) (dump-CPXNUM obj id))
                        (else
                         (compiler-internal-error
                           "targ-dump, unknown subtype" subtype))))))
                objs)
      (targ-line)
      (targ-dump-section "BEGIN_SUB" "END_SUB" #t objs
        (lambda (i x)
          (let ((id (targ-sub-name (caddr x))))
            (targ-code* (list "DEF_SUB" id))))))))

(define (targ-dump-prc objs c-decls)

  (for-each
    (lambda (x)
      (targ-display x)
      (targ-line))
    c-decls)

  (for-each targ-code* (queue->list targ-module-decl))

  (targ-begin-cod targ-module-rd-res targ-module-wr-res
                  "BEGIN_M_COD" 'MD- 'MR- 'MW-
                  #f #f #f #f)

  (targ-code* '("BEGIN_M_SW"))

  (targ-line)

  (for-each targ-dump-cod objs)

  (targ-code* '("END_M_SW"))

  (targ-end-cod "END_M_COD")

  (targ-line)

  (targ-code* '("BEGIN_LBL"))

  (let loop1 ((i 0) (objs objs))
    (if (pair? objs)
      (let ((obj (car objs)))
        (let* ((proc (car obj))
               (p (cdr obj))
               (name (proc-obj-name proc))
               (info (car p))
               (lbls (vector-ref info 1)))
          (if (= i 0)
            (targ-display " ")
            (targ-display ","))
          (targ-code* (list "DEF_LBL_INFO"
                            (targ-c-id-host name)
                            (targ-heap-ref-obj (vector-ref info 4))))
          (let loop2 ((i (+ i 1)) (lbls lbls))
            (if (pair? lbls)
              (let ((x (car lbls))
                    (host (targ-c-id-host name)))
                (targ-display ",")
                (targ-code*
                  (cons (case (vector-ref x 2)
                          ((ENTRY SUBPROC)   "DEF_LBL_PROC")
                          ((RETURN)          "DEF_LBL_RETN")
                          ((RETURN-TASK)     "DEF_LBL_RETT")
                          ((RETURN-INTERNAL) "DEF_LBL_RETI")
                          (else
                           (compiler-internal-error
                             "targ-dump-prc, unknown label type")))
                        (cons host (vector-ref x 3))))
                (loop2 (+ i 1) (cdr lbls)))
              (loop1 i (cdr objs))))))))

  (targ-code* '("END_LBL"))
  (targ-line))

(define (targ-dump-cod obj)
  (let* ((proc (car obj))
         (p (cdr obj))
         (name (proc-obj-name proc))
         (info (car p))
         (code (vector-ref info 0))
         (val-lbls (vector-ref info 1))
         (rd-res (vector-ref info 2))
         (wr-res (vector-ref info 3)))

    (define (mk-glbl lbl)
      (targ-make-glbl (targ-lbl-num lbl) name))

    (define (def var val)
      (targ-define-prefix (list var))
      (targ-display " ")
      (targ-code val)
      (targ-line))

    (def "PH_PROC" (targ-c-id-host name))
    (def "PH_LBL0" (caddr p))

    (targ-begin-cod rd-res wr-res
                    "BEGIN_P_COD" 'PD- 'PR- 'PW-
                    "BEGIN_P_SLBL" "DEF_P_SLBL" "END_P_SLBL"
                    (map mk-glbl val-lbls))

    (targ-code* '("BEGIN_P_SW"))

    (let loop ((l code))
      (if (pair? l)
        (let ((x (car l)))
          (cond ((and (pair? x) (eq? (car x) 'LABEL))
                 (let* ((lbl (cdr x))
                        (glbl (mk-glbl lbl)))
                   (cond ((targ-lbl-val? lbl)
                          (targ-code*
                            (list "DEF_SLBL" (targ-lbl-num lbl) glbl)))
                         ((targ-lbl-goto? lbl)
                          (targ-code* (list "DEF_GLBL" glbl))))))
                ((and (pair? x) (eq? (car x) 'APPEND))
                 (targ-code-append (cdr x)))
                (else
                 (targ-display "   ")
                 (targ-code* x)))
          (loop (cdr l)))))

    (targ-code* '("END_P_SW"))

    (targ-end-cod "END_P_COD")

    (targ-line)))

(define (targ-begin-cod rd-res wr-res beg decl rd wr
                        beg-slbl def-slbl end-slbl lbls)
  (let ((nb-res (+ targ-nb-gvm-regs targ-nb-non-reg-res)))

    (define (resource-set name1 name2 element?)
      (targ-define-prefix (cons name1 "ALL"))
      (let loop ((i 0))
        (if (< i nb-res)
          (if (element? i)
            (begin
              (targ-display " ")
              (targ-code (cons name2 (targ-res-name i)))
              (loop (+ i 1)))
            (loop (+ i 1)))))
      (targ-line))

    (resource-set
      decl
      'D-
      (lambda (i) (or (vector-ref rd-res i) (vector-ref wr-res i))))

    (resource-set
      rd
      'R-
      (lambda (i) (vector-ref rd-res i)))

    (resource-set
      wr
      'W-
      (lambda (i) (vector-ref wr-res i)))

    (targ-code* (list beg))

    (if lbls
      (begin
        (targ-code* (list beg-slbl))
        (let loop ((lst lbls))
          (if (not (null? lst))
            (let ((x (car lst)))
              (targ-code* (list def-slbl x))
              (loop (cdr lst)))))
        (targ-code* (list end-slbl))))))

(define (targ-end-cod end)
  (targ-code* (list end)))

(define (targ-dump-ini objs c-inits)

  (targ-code* '("BEGIN_INI"))

  (for-each
    (lambda (obj)
      (let* ((proc (car obj))
             (p (cdr obj))
             (name (proc-obj-name proc)))
        (targ-code* (list "DEF_PRM"
                          (targ-use-glo (string->symbol name) #f)
                          (targ-c-id-glo name)
                          (caddr p)))))
    objs)

  (for-each
    (lambda (x)
      (targ-display x)
      (targ-line))
    c-inits)

  (targ-code* '("END_INI")))

(define (targ-extract-primitives objs)
  (let loop ((objs objs) (prms '()))
    (if (pair? objs)
      (let ((obj (car objs)))
        (let ((proc (car obj)))
          (if (proc-obj-primitive? proc)
            (loop (cdr objs) (cons obj prms))
            (loop (cdr objs) prms))))
      (reverse prms))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define targ-linker-tag "")
(define targ-host-tag   "H_")
(define targ-glo-tag    "G_")
(define targ-sym-tag    "S_")

(define targ-port #f)
(define targ-line-size #f)

(define (targ-code* x)
  (targ-code x)
  (targ-line))

(define (targ-code x)

  (define (one-head head args)
    (targ-display-prefixed head)
    (targ-code-args args))

  (define (two-heads head1 head2 args)
    (targ-display-prefixed head1)
    (targ-display head2)
    (targ-code-args args))

  (define (goto-lbl n name)
    (targ-display-prefixed "L")
    (targ-code n)
    (targ-display "_")
    (targ-display (targ-name->c-id name)))

  (if (pair? x)
    (let ((head (car x)) (tail (cdr x)))
      (case head
        ((C-LONG)      (targ-display-c-long      tail))
        ((C-HEX-ULONG) (targ-display-c-hex-ulong tail))
        ((C-CHAR)      (targ-display-c-char      tail))
        ((C-STRING)    (targ-display-c-string    tail))
        ((C-ID-LINKER) (targ-display-c-id        targ-linker-tag tail))
        ((C-ID-HOST)   (targ-display-c-id        targ-host-tag   tail))
        ((C-ID-GLO)    (targ-display-c-id        targ-glo-tag    tail))
        ((C-ID-SYM)    (targ-display-c-id        targ-sym-tag    tail))
        ((SUB-NAME)    (targ-display-prefixed "X") (targ-code tail))
        ((GLBL)        (goto-lbl (car tail) (cadr tail)))
        ((D-)          (two-heads "D_"     tail '()))
        ((R-)          (two-heads "R_"     tail '()))
        ((W-)          (two-heads "W_"     tail '()))
        ((MD-)         (two-heads "MD_"    tail '()))
        ((MR-)         (two-heads "MR_"    tail '()))
        ((MW-)         (two-heads "MW_"    tail '()))
        ((PD-)         (two-heads "PD_"    tail '()))
        ((PR-)         (two-heads "PR_"    tail '()))
        ((PW-)         (two-heads "PW_"    tail '()))
        ((PSR)         (two-heads "PSR"    tail '()))
        ((R)           (two-heads "R"      tail '()))
        ((FR)          (two-heads "FR"     tail '()))
        ((SET-R)       (two-heads "SET_R"  (car tail) (cdr tail)))
        ((VEC)         (two-heads "VEC"    (car tail) (cdr tail)))
        ((STR)         (two-heads "STR"    (car tail) (cdr tail)))
        ((BIG)         (two-heads "BIG"    (car tail) (cdr tail)))
        ((SEQ)         (targ-code-seq tail))
        ((APPEND)      (targ-code-append tail))
        ((CELL)        (targ-code tail))
        ((PRM)         (one-head "PRM" tail))
        ((PRC)         (one-head "PRC" tail))
        (else          (one-head head tail))))
    (if x (targ-display x))))

(define (targ-code-args args)
  (targ-code-list
    args
    (lambda () (targ-display ","))
    (lambda () (targ-display "("))
    (lambda () (targ-display ")"))
    (lambda () (if (> targ-line-size 25) (targ-line)))))

(define (targ-code-seq args)
  (targ-code-list
    args
    (lambda () (targ-display " "))
    (lambda () 0)
    (lambda () 0)
    (lambda () 0)))

(define (targ-code-append args)
  (targ-code-list
    args
    (lambda () 0)
    (lambda () 0)
    (lambda () 0)
    (lambda () 0)))

(define (targ-code-list args sep prefix suffix line-break)
  (if (pair? args)
    (let ((arg (car args)))
      (if arg
        (begin
          (prefix)
          (line-break)
          (targ-code arg)
          (let loop ((args (cdr args)))
            (if (pair? args)
              (let ((arg (car args)))
                (if arg
                  (begin
                    (sep)
                    (line-break)
                    (targ-code arg)))
                (loop (cdr args)))))
          (suffix)
          (line-break))
        (targ-code-list (cdr args) sep prefix suffix line-break)))))

(define (targ-display x)
  (display x targ-port)
  (if (eqv? x #\newline)
    (set! targ-line-size 0)
    (set! targ-line-size (+ targ-line-size 1))))

(define (targ-line)
  (targ-display #\newline))

(define (targ-make-cell x)
  (cons 'CELL x))

(define (targ-cell-set! x y)
  (set-cdr! x y))

(define (targ-c-long n)
  (cons 'C-LONG n))

(define (targ-c-hex-ulong n)
  (cons 'C-HEX-ULONG n))

(define (targ-c-char c)
  (cons 'C-CHAR c))

(define (targ-c-string str)
  (cons 'C-STRING str))

(define (targ-c-id-linker name)
  (cons 'C-ID-LINKER name))

(define (targ-c-id-host name)
  (cons 'C-ID-HOST name))

(define (targ-c-id-glo name)
  (cons 'C-ID-GLO name))

(define (targ-c-id-sym name)
  (cons 'C-ID-SYM name))

(define (targ-sub-name n)
  (cons 'SUB-NAME n))

(define (targ-display-prefixed str)
  (targ-display c-id-prefix) ; c-id-prefix is defined in "_parms.scm"
  (targ-display str))

(define (targ-display-c-long n)
  (targ-display n)
  (targ-display "L"))

(define (targ-display-c-hex-ulong n)
  (targ-display "0x")
  (targ-display (number->string n 16))
  (targ-display "L"))

(define (targ-display-c-char c)
  (targ-display "'")
  (targ-display (targ-char->c-char c))
  (targ-display "'"))

(define (targ-display-c-string str)
  (targ-display "\"")
  (targ-display str) ;********** check for " and \ and \n, etc
  (targ-display "\""))

(define (targ-display-c-id type name)
  (targ-display-prefixed type)
  (targ-display (targ-name->c-id name)))

(define (targ-char->c-char c)

  (define oct-table "01234567")

  (cond ((char=? c #\\)
         "\\\\")
        ((char=? c #\')
         "\\'")
        ((char=? c #\newline)
         "\\n")
        (else
         (let ((n (character-encoding c)))
           (if (or (< n 32) (>= n 127))
             (string #\\
                     (string-ref oct-table (quotient n 64))
                     (string-ref oct-table (modulo (quotient n 8) 8))
                     (string-ref oct-table (modulo n 8)))
             (string c))))))

(define (targ-name->c-id s)
  (scheme-id->c-id s))

;------------------------------------------------------------------------------

(define (targ-scan-procedure obj)
  (let* ((proc (car obj))
         (p (cdr obj))
         (nb-res (+ targ-nb-gvm-regs targ-nb-non-reg-res)))

    (if targ-info-port
      (begin
        (display "  #[" targ-info-port)
        (if (proc-obj-primitive? proc)
          (display "primitive " targ-info-port)
          (display "procedure " targ-info-port))
        (display (proc-obj-name proc) targ-info-port)
        (display "]" targ-info-port)))

    (set! targ-proc-name               (proc-obj-name proc))
    (set! targ-proc-code               (queue-empty))
    (set! targ-proc-rd-res             (make-vector nb-res #f))
    (set! targ-proc-wr-res             (make-vector nb-res #f))
    (set! targ-proc-lbl-tbl            (queue-empty))
    (set! targ-proc-lbl-tbl-ord        (queue-empty))
    (set! targ-var-descr-queue         (queue-empty))
    (set! targ-first-class-label-queue (queue-empty))

    (let ((x (proc-obj-code proc)))
      (if (bbs? x)
        (targ-scan-scheme-procedure x)
        (targ-scan-c-procedure x)))

    (targ-cell-set! (caddr p) (+ targ-lbl-alloc 1))

    ; Assign label numbers sequentially, starting with "value" labels
    ; and then "goto" labels

    (let ((code (queue->list targ-proc-code))
          (ord-lbls (queue->list targ-proc-lbl-tbl-ord)))
      (let loop2 ((l ord-lbls) (i 0) (val-lbls '()))
        (if (pair? l)
          (let ((x (car l)))
            (if (targ-lbl-val? x)
              (begin
                (targ-cell-set! (targ-lbl-num x) i)
                (loop2 (cdr l) (+ i 1) (cons x val-lbls)))
              (loop2 (cdr l) i val-lbls)))
          (let ((info (targ-debug-info)))
            (targ-use-obj info)
            (set! targ-lbl-alloc (+ targ-lbl-alloc (+ i 1)))
            (set-car! p
              (vector code
                      (reverse val-lbls)
                      targ-proc-rd-res
                      targ-proc-wr-res
                      info))
            (let loop3 ((l ord-lbls) (i i))
              (if (pair? l)
                (let ((x (car l)))
                  (if (and (targ-lbl-goto? x)
                           (not (targ-lbl-val? x)))
                    (begin
                      (targ-cell-set! (targ-lbl-num x) i)
                      (loop3 (cdr l) (+ i 1)))
                    (loop3 (cdr l) i)))))))))

    (if targ-info-port
      (newline targ-info-port))

    ))

(define (targ-debug-info)

  (define (number i lst)
    (if (null? lst)
      '()
      (cons (lst->vector (cons i (car lst)))
            (number (+ i 1) (cdr lst)))))

  (if targ-debug-info?
    (vector (lst->vector (number 0 (queue->list targ-first-class-label-queue)))
            (lst->vector (queue->list targ-var-descr-queue)))
    #f))

(define (targ-scan-scheme-procedure bbs)

  (set! targ-proc-entry-lbl   (bbs-entry-lbl-num bbs))
  (set! targ-proc-lbl-counter (bbs-lbl-counter bbs))

  (let loop ((prev-bb #f)
             (prev-gvm-instr #f)
             (l (bbs->code-list bbs)))
    (if (not (null? l))
      (let ((pres-bb (code-bb (car l)))
            (pres-gvm-instr (code-gvm-instr (car l)))
            (pres-slots-needed (code-slots-needed (car l)))
            (next-gvm-instr (if (null? (cdr l))
                              #f
                              (code-gvm-instr (cadr l)))))

        (targ-gen-gvm-instr prev-gvm-instr
                            pres-gvm-instr
                            next-gvm-instr
                            pres-slots-needed)

        (loop pres-bb pres-gvm-instr (cdr l))))))

(define (targ-scan-c-procedure c-proc)

  (define (ps-opnd opnd)
    (cond ((reg? opnd)
           (let ((n (reg-num opnd)))
             (cons 'PSR n)))
          (else ; must be stack slot
           (list "PSSTK" (- (stk-num opnd) targ-proc-fp)))))

  (let* ((arity (c-proc-arity c-proc))
         (pc (targ-label-info arity arity #f #f))
         (pc-map (pcontext-map pc))
         (pc-fs (pcontext-fs pc))
         (ret (cdr (assq 'return pc-map)))
         (final-fs (+ arity 1)))

    (set! targ-proc-entry-lbl 1)

    (targ-start-bb pc-fs)

    (set! targ-proc-entry-frame targ-proc-exit-frame) ; ********** for targ-update-fr but probably not needed since it can't be called from here!
    (targ-begin-fr) ; ************* not needed either

    (targ-emit-label-entry targ-proc-entry-lbl arity #f)
    (targ-ref-lbl-goto targ-proc-entry-lbl)

    (targ-emit (list "IF_NARGS_EQ" arity '("NOTHING")))
    (targ-emit (list "WRONG_NARGS" (targ-ref-lbl-val targ-proc-entry-lbl)))
    (targ-emit (list 'APPEND
                     (list "DEF_GLBL" (targ-make-glbl "" targ-proc-name))
                     #\newline))

    (let loop1 ((i 1))
      (if (and (<= i arity) (<= i targ-nb-arg-regs))
        (begin
          (targ-emit
            (targ-loc (make-stk (+ pc-fs i)) (targ-opnd (make-reg i))))
          (loop1 (+ i 1)))))

    (targ-emit (targ-loc (make-stk final-fs) (targ-opnd ret)))
    (targ-emit (targ-adjust-stack final-fs))

    (targ-emit
      (list 'APPEND
            "#undef " c-id-prefix "NARGS" #\newline
            "#define " c-id-prefix "NARGS " arity #\newline))

    (let loop2 ((i 1))
      (if (<= i arity)
        (let ((opnd (make-stk i)))
          (targ-emit
            (list 'APPEND
                  "#undef " c-id-prefix "ARG" i #\newline
                  "#define " c-id-prefix "ARG" i
                  " " (ps-opnd opnd) #\newline))
          (loop2 (+ i 1)))))

    (targ-emit
      (list 'APPEND
            "#undef " c-id-prefix "RESULT" #\newline
            "#define " c-id-prefix "RESULT " (ps-opnd (make-reg 1)) #\newline
            '("BEGIN_C_COD") #\newline
            (c-proc-body c-proc)
            '("END_C_COD") #\newline))

    (let ((adj-stk (targ-adjust-stack 0)))
      (targ-emit
        (list 'SEQ
              adj-stk
              (list "JUMPPRM"
                    '("NOTHING")
                    (targ-opnd (make-stk final-fs))))))))

;------------------------------------------------------------------------------

; Information attached to a procedure

(define targ-proc-name            #f) ; procedure's name
(define targ-proc-code            #f) ; code of the procedure
(define targ-proc-entry-lbl       #f) ; entry label
(define targ-proc-lbl-counter     #f) ; label counter
(define targ-proc-rd-res          #f) ; set of resources read from
(define targ-proc-wr-res          #f) ; set of resources written to
(define targ-proc-lbl-tbl         #f) ; table of all labels
(define targ-proc-lbl-tbl-ord     #f) ; table of all labels ordered by def time
(define targ-proc-fp              #f) ; frame pointer
(define targ-proc-hp              #f) ; heap pointer

(define targ-proc-instr-source    #f)
(define targ-proc-entry-frame     #f)
(define targ-proc-exit-frame      #f)

; Emit a piece of code

(define (targ-emit code)
  (queue-put! targ-proc-code code))

; Emit a label

(define (targ-emit-label-simp lbl)
  (targ-emit-label lbl #f))

(define (targ-emit-label-entry lbl nb-parms label-descr)
  (targ-emit-label lbl 'ENTRY nb-parms 0))

(define (targ-emit-label-subproc lbl nb-parms nb-closed label-descr)
  (targ-emit-label lbl 'SUBPROC nb-parms nb-closed))

(define (targ-emit-label-return lbl fs link label-descr)
  (targ-emit-label lbl 'RETURN fs link))

(define (targ-emit-label-return-task lbl fs link label-descr)
  (targ-emit-label lbl 'RETURN-TASK fs link))

(define (targ-emit-label-return-internal lbl fs link label-descr)
  (targ-emit-label lbl 'RETURN-INTERNAL fs link))

; Add a label to the code stream

(define (targ-emit-label lbl class . info)
  (let ((x (targ-get-lbl lbl)))
    (targ-emit (cons 'LABEL x))
    (targ-add-label x class info)))

(define (targ-add-label lbl-struct class info)
  (vector-set! lbl-struct 2 class)
  (vector-set! lbl-struct 3 info)
  (queue-put! targ-proc-lbl-tbl-ord lbl-struct))

; Add label "n" to label table and return label object

(define (targ-get-lbl n)
  (let ((x (assq n (queue->list targ-proc-lbl-tbl))))
    (if x
      (cdr x)
      (let ((y (vector (targ-make-cell #f) ; eventual label number (set later)
                       #f                  ; used as a "goto" label?
                       #f                  ; class (not #f if "value" label)
                       #f)))               ; extra info if "value" label
        (queue-put! targ-proc-lbl-tbl (cons n y))
        y))))

(define (targ-lbl-num lbl-struct)
  (vector-ref lbl-struct 0))

; Mark a label as referenced for "value" and return eventual label number

(define (targ-ref-lbl-val n)
  (let ((x (targ-get-lbl n)))
    (targ-lbl-num x)))

(define (targ-lbl-val? lbl-struct)
  (vector-ref lbl-struct 2))

; Mark a label as target for "goto" and return eventual label number

(define (targ-ref-lbl-goto n)
  (let ((x (targ-get-lbl n)))
    (vector-set! x 1 #t)
    (targ-make-glbl (targ-lbl-num x) targ-proc-name)))

(define (targ-lbl-goto? lbl-struct)
  (vector-ref lbl-struct 1))

(define (targ-make-glbl n name)
  (list 'GLBL n name))

; To generate new, unique labels

(define (targ-new-lbl)
  (targ-proc-lbl-counter))

(define (targ-heap-reserve space fs)
  (set! targ-proc-hp (+ targ-proc-hp space))
  (if (> targ-proc-hp targ-heap-fudge)
    (targ-check-heap space fs)))

(define (targ-check-heap space fs)
  (let ((lbl (targ-new-lbl)))
    (targ-need-heap)
    (targ-update-fr)
    (targ-end-fr)
    (targ-emit (list 'SEQ
                     (targ-adjust-stack fs)
                     (list "CHECK_HEAP"
                           (targ-ref-lbl-val lbl)
                           (+ targ-heap-fudge space))))
    (targ-gen-label-return* lbl 'RETURN-INTERNAL fs)
    (set! targ-proc-hp 0)))

(define (targ-poll fs)
  (let ((lbl (targ-new-lbl)))
    (targ-rd-fp)
    (targ-update-fr)
    (targ-end-fr)
    (targ-emit (list 'SEQ
                     (targ-adjust-stack fs)
                     (list "POLL" (targ-ref-lbl-val lbl))))
    (targ-gen-label-return* lbl 'RETURN-INTERNAL fs)))

(define (targ-start-bb fs)
  (set! targ-proc-hp 0)
  (set! targ-proc-fp fs))

(define (targ-begin-fr) ; start of a floating point region
  (targ-fp-cache-init)
  (set! targ-fr-cell (targ-make-cell #f))
  (targ-emit (list 'APPEND targ-fr-cell)))

(define (targ-update-fr)
  (let* ((frame
          targ-proc-entry-frame);*****targ-proc-exit-frame
         (live
          (frame-live frame))
         (any-closed-live?
          (not (set-empty? (set-intersection
                             live
                             (list->set (frame-closed frame)))))))

    (define (live? var)
      (or (set-member? var live)
          (and (eq? var closure-env-var) any-closed-live?)))

    (let loop1 ((i 1) (l (reverse (frame-slots frame))))
      (if (pair? l)
        (begin
          (if (live? (car l))
            (targ-fp-cache-write-if-dirty (make-stk i)))
          (loop1 (+ i 1) (cdr l)))
        (let loop2 ((i 0) (l (frame-regs frame)))
          (if (pair? l)
            (begin
              (if (live? (car l))
                (targ-fp-cache-write-if-dirty (make-reg i)))
              (loop2 (+ i 1) (cdr l)))))))))

(define (targ-end-fr) ; end of a floating point region

  (define (make-decls i)
    (if (>= i 0)
      (cons (list "D_FR" (cons 'FR i)) (make-decls (- i 1)))
      '()))

  (let ((n (targ-fp-cache-size)))
    (if (> n 0)
      (let ((decls (reverse (make-decls (- n 1)))))
        (targ-cell-set!
          targ-fr-cell
          (list 'APPEND (cons 'SEQ (cons '("BEGIN_FR") decls)) #\newline))
        (targ-emit '(APPEND ("END_FR") #\newline))))))

; Management of resources

(define targ-nb-non-reg-res 2)

(define (targ-res-name i)
  (if (< i targ-nb-non-reg-res)
    (vector-ref '#("HEAP" "FP") i)
    (string-append "R" (number->string (- i targ-nb-non-reg-res)))))

(define (targ-rd-res i)
  (vector-set! targ-module-rd-res i #t)
  (vector-set! targ-proc-rd-res i #t))

(define (targ-wr-res i)
  (targ-rd-res i)
  (vector-set! targ-module-wr-res i #t)
  (vector-set! targ-proc-wr-res i #t))

(define (targ-need-heap)
  (targ-wr-res 0))

(define (targ-rd-fp)
  (targ-rd-res 1))

(define (targ-wr-fp)
  (targ-wr-res 1))

(define (targ-rd-reg n)
  (targ-rd-res (+ n targ-nb-non-reg-res)))

(define (targ-wr-reg n)
  (targ-wr-res (+ n targ-nb-non-reg-res)))

(define (targ-use-all-res)
  (let loop ((i (- (+ targ-nb-non-reg-res targ-nb-gvm-regs) 1)))
    (if (>= i 0)
      (begin
        (targ-wr-res i)
        (loop (- i 1))))))

;------------------------------------------------------------------------------

(define (targ-add-var-descr! descr)

  (define (index x l)
    (let loop ((l l) (i 0))
      (cond ((not (pair? l))    #f)
            ((equal? (car l) x) i)
            (else               (loop (cdr l) (+ i 1))))))

  (let ((n (index descr (queue->list targ-var-descr-queue))))
    (if n
      n
      (let ((m (length (queue->list targ-var-descr-queue))))
        (queue-put! targ-var-descr-queue descr)
        m))))

(define (targ-add-first-class-label! source slots frame)
  (let loop ((i 0) (l1 slots) (l2 '()))
    (if (pair? l1)
      (let ((var (car l1)))
        (let ((x (frame-live? var frame)))
          (if (and x (or (pair? x) (not (temp-var? x))))
            (let ((descr-index
                    (targ-add-var-descr!
                     (if (pair? x)
                        (map (lambda (y) (targ-add-var-descr! (var-name y))) x)
                        (var-name x)))))
              (loop (+ i 1) (cdr l1) (cons (+ (* i 16384) descr-index) l2)))
            (loop (+ i 1) (cdr l1) l2))))
      (let ((label-descr (cons source l2)))
        (queue-put! targ-first-class-label-queue label-descr)
        label-descr))))

;------------------------------------------------------------------------------

(define (targ-gen-gvm-instr prev-gvm-instr gvm-instr next-gvm-instr sn)

  (set! targ-proc-instr-source
    (comment-get (gvm-instr-comment gvm-instr) 'SOURCE))
  (set! targ-proc-exit-frame
    (gvm-instr-frame gvm-instr))
  (set! targ-proc-entry-frame
    (and prev-gvm-instr (gvm-instr-frame prev-gvm-instr)))

  (case (gvm-instr-type gvm-instr)

    ((LABEL)
     (set! targ-proc-entry-frame targ-proc-exit-frame)
     (targ-start-bb (frame-size targ-proc-exit-frame))
     (case (LABEL-type gvm-instr)
       ((SIMPLE)
        (targ-gen-LABEL-SIMPLE (LABEL-lbl-num gvm-instr)
                               sn))
       ((ENTRY)
        (targ-gen-LABEL-ENTRY (LABEL-lbl-num gvm-instr)
                              (LABEL-ENTRY-nb-parms gvm-instr)
                              (LABEL-ENTRY-min gvm-instr)
                              (LABEL-ENTRY-rest? gvm-instr)
                              (LABEL-ENTRY-closed? gvm-instr)
                              sn))
       ((RETURN)
        (targ-gen-LABEL-RETURN (LABEL-lbl-num gvm-instr)
                               sn))
       ((TASK-ENTRY)
        (targ-gen-LABEL-TASK-ENTRY (LABEL-lbl-num gvm-instr)
                                   sn))
       ((TASK-RETURN)
        (targ-gen-LABEL-TASK-RETURN (LABEL-lbl-num gvm-instr)
                                    sn))
       (else
        (compiler-internal-error
          "targ-gen-gvm-instr, unknown label type"))))

    ((APPLY)
     (targ-gen-APPLY (APPLY-prim gvm-instr)
                     (APPLY-opnds gvm-instr)
                     (APPLY-loc gvm-instr)
                     sn))

    ((COPY)
     (targ-gen-COPY (COPY-opnd gvm-instr)
                    (COPY-loc gvm-instr)
                    sn))

    ((CLOSE)
     (targ-gen-CLOSE (CLOSE-parms gvm-instr)
                     sn))

    ((IFJUMP)
     (targ-gen-IFJUMP (IFJUMP-test gvm-instr)
                      (IFJUMP-opnds gvm-instr)
                      (IFJUMP-true gvm-instr)
                      (IFJUMP-false gvm-instr)
                      (IFJUMP-poll? gvm-instr)
                      (if (and next-gvm-instr
                               (memq (LABEL-type next-gvm-instr)
                               '(SIMPLE TASK-ENTRY)))
                        (LABEL-lbl-num next-gvm-instr)
                        #f)))

    ((JUMP)
     (targ-gen-JUMP (JUMP-opnd gvm-instr)
                    (JUMP-nb-args gvm-instr)
                    (JUMP-poll? gvm-instr)
                    (JUMP-safe? gvm-instr)
                    (if (and next-gvm-instr
                             (memq (LABEL-type next-gvm-instr)
                                   '(SIMPLE TASK-ENTRY)))
                      (LABEL-lbl-num next-gvm-instr)
                      #f)))

    (else
     (compiler-internal-error
       "targ-gen-gvm-instr, unknown 'gvm-instr'" gvm-instr))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-LABEL-SIMPLE lbl sn)
  (targ-emit-label-simp lbl)
  (targ-begin-fr))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-LABEL-ENTRY lbl nb-parms min rest? closed? sn)

  (let ((label-descr (targ-add-first-class-label!
                       targ-proc-instr-source
                       '()
                       targ-proc-exit-frame)))
    (if (= lbl targ-proc-entry-lbl)
      (begin
        (targ-emit-label-entry lbl nb-parms label-descr)
        (targ-ref-lbl-val lbl)
        (targ-ref-lbl-goto lbl))
      (let ((nb-closed (length (frame-closed targ-proc-exit-frame))))
        (targ-emit-label-subproc lbl nb-parms nb-closed label-descr))))

  (let ((nb-parms* (if rest? (- nb-parms 1) nb-parms))
        (lbl* (targ-ref-lbl-val lbl)))

    (define (check-args i)
      (if (> i nb-parms*)

        (targ-emit
          (if rest?
            (list "BUILD_REST" lbl*)
            (list "WRONG_NARGS" lbl*)))

        (let ((nb-stacked (max 0 (- i targ-nb-arg-regs)))
              (nb-stacked* (max 0 (- nb-parms targ-nb-arg-regs))))

          (define (setup-args j)
            (if (<= j nb-parms)
              (let* ((rest (setup-args (+ j 1)))
                     (src-reg (- j nb-stacked))
                     (src (cond ((<= j i) (cons 'R src-reg))
                                ((and rest? (= j nb-parms)) '("NUL"))
                                (else '("UNA")))))
                (if (<= j nb-stacked*)
                  (begin
                    (if (<= j i) (targ-rd-reg src-reg))
                    (cons (list "PUSH" src) rest))
                  (if (and (<= j i) (= nb-stacked nb-stacked*))
                    rest
                    (let ((dst-reg (- j nb-stacked*)))
                      (if (<= j i) (targ-rd-reg src-reg))
                      (targ-wr-reg dst-reg)
                      (cons (list 'SET-R dst-reg src) rest)))))
              '()))

          (let ((x (setup-args (+ nb-stacked 1))))
            (targ-emit (list "IF_NARGS_EQ"
                             i
                             (if (null? x) '("NOTHING") (cons 'SEQ x)))))

          (check-args (+ i 1)))))

    (check-args min)

    (if (= lbl targ-proc-entry-lbl)
      (targ-emit (list 'APPEND
                       (list "DEF_GLBL" (targ-make-glbl "" targ-proc-name))
                       #\newline)))

    (targ-begin-fr)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-LABEL-RETURN lbl sn)
  (targ-gen-label-return* lbl 'RETURN (frame-size targ-proc-exit-frame)))

(define (targ-gen-label-return* lbl kind fs)

  (define (adjust-slots l n)
    (cond ((= n 0) l)
          ((< n 0) (adjust-slots (cdr l) (+ n 1)))
          (else    (adjust-slots (cons empty-var l) (- n 1)))))

  (let* ((frame targ-proc-entry-frame)
         (slots (reverse (adjust-slots (frame-slots frame)
                                       (- fs (frame-size frame)))))
         (regs (frame-regs frame))
         (vars (if (eq? kind 'RETURN-INTERNAL)
                 (append slots regs)
                 slots)))
    (let ((label-descr
            (targ-add-first-class-label!
              targ-proc-instr-source
              vars
              frame)))
      (let ((link (pos-in-list ret-var vars)))
        (if link
          (let ((n (length slots)))
            (case kind
              ((RETURN)
               (targ-emit-label-return lbl n link label-descr))
              ((RETURN-TASK)
               (targ-emit-label-return-task lbl n link label-descr))
              ((RETURN-INTERNAL)
               (targ-emit-label-return-internal lbl n link label-descr))
              (else
               (compiler-internal-error
                 "targ-gen-label-return*, unknown label kind")))
            (targ-begin-fr))
          (compiler-internal-error
            "targ-gen-label-return*, no return address in frame"))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-LABEL-TASK-ENTRY lbl sn)

  (targ-emit-label-simp lbl)

  (let ((fs (frame-size targ-proc-exit-frame)))
    (if (= fs 0)

      (targ-emit '("PUSH" (r . 0)))

      (begin
       (targ-emit '("ADJFP" 1))
        (let loop ((i 0))
          (if (< i fs)
            (begin
              (targ-emit (list "SET_STK" (- i) (list "STK" (- (+ i 1)))))
              (loop (+ i 1)))
            (targ-emit (list "SET_STK" (- i) '(r . 0)))))))

    (targ-emit (list "TASK_PUSH" fs))

    (targ-begin-fr)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-LABEL-TASK-RETURN lbl sn)
  (let ((lbl2 (targ-new-lbl))
        (fs (+ (frame-size targ-proc-exit-frame) 1)))

    (targ-start-bb fs)

    (targ-gen-label-return* lbl 'RETURN-TASK fs)
    (targ-emit (list "TASK_POP" (targ-ref-lbl-val lbl2)))
    (targ-gen-label-return* lbl2 'RETURN fs)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-APPLY prim opnds loc sn)
  (let ((proc (proc-obj-inlinable prim)))
    (if proc
      (proc opnds loc sn)
      (compiler-internal-error
        "targ-gen-APPLY, unknown 'prim'" prim))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-COPY opnd loc sn)
  (targ-emit
    (targ-loc loc (targ-opnd opnd))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-CLOSE parms sn)

  (define (close parms* sn*)
    (if (pair? parms*)

      (let* ((parm (car parms*))
             (lbl (closure-parms-lbl parm))
             (loc (closure-parms-loc parm))
             (opnds (closure-parms-opnds parm))
             (sn** (targ-sn-opnds opnds sn*)))
        (close (cdr parms*) (targ-sn-loc loc sn**))
        (let ((x (targ-opnd loc)))
          (targ-emit (list "SETUP_CLO"
                           (length opnds)
                           x
                           (targ-ref-lbl-val lbl)
                           (targ-seq-opnds opnds "ADD_CLO_ELEM")))))

      (begin

        (targ-heap-reserve
          (apply +
                 (map (lambda (parm)
                        (targ-closure-space
                          (length (closure-parms-opnds parm))))
                      parms))
          sn*)

        (for-each (lambda (parm)
                    (let ((loc (closure-parms-loc parm))
                          (opnds (closure-parms-opnds parm)))
                      (targ-emit
                        (targ-loc loc (list "ALLOC_CLO" (length opnds))))))
                  parms))))

  (close (reverse parms) sn))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-IFJUMP test opnds true-lbl false-lbl poll? next-lbl)
  (let ((proc (proc-obj-test test)))
    (if proc

      (let* ((fs (frame-size targ-proc-exit-frame))
             (fs* (targ-sn-opnds opnds fs)))

        (if (> targ-proc-hp 0)
          (targ-check-heap 0 fs*))

        (if poll?
          (targ-poll fs*))

        (targ-update-fr)

        (let ((adj-stk (targ-adjust-stack fs)))

          (define (emit adj-stk test lbl other)
            (targ-emit
              (list 'SEQ
                    adj-stk
                    (list "IF_GOTO" test (targ-ref-lbl-goto lbl))
                    other)))

          (cond ((eqv? true-lbl next-lbl)
                 (emit adj-stk (proc #t opnds fs) false-lbl #f))
                ((eqv? false-lbl next-lbl)
                 (emit adj-stk (proc #f opnds fs) true-lbl #f))
                (else
                 (emit adj-stk (proc #f opnds fs) true-lbl
                       (list "GOTO" (targ-ref-lbl-goto false-lbl)))))

          (targ-end-fr)))

      (compiler-internal-error
        "targ-gen-IFJUMP, unknown 'test'" test))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-gen-JUMP opnd nb-args poll? safe? next-lbl)
  (let* ((fs (frame-size targ-proc-exit-frame))
         (fs* (targ-sn-opnd opnd fs)))

    (if (> targ-proc-hp 0)
      (targ-check-heap 0 fs*))

    (if poll?
      (targ-poll fs*))

    (targ-update-fr)

    (let* ((adj-stk (targ-adjust-stack fs))
           (set-nargs (if nb-args (list "SET_NARGS" nb-args) #f)))

      (cond ((lbl? opnd)
             (let ((n (lbl-num opnd)))
               (if (and next-lbl (= next-lbl n)) ; fall through?
                 (if (or adj-stk set-nargs)
                   (targ-emit
                     (list 'SEQ
                           adj-stk
                           set-nargs)))
                 (targ-emit
                   (list 'SEQ
                         adj-stk
                         set-nargs
                         (list "GOTO" (targ-ref-lbl-goto n)))))))
            ((and (obj? opnd)
                  (proc-obj? (obj-val opnd)))
             (let* ((proc (obj-val opnd))
                    (x (targ-use-prc proc #f)))
               (if (eq? (car x) 'PRM)
                 (targ-emit
                   (list 'SEQ
                         adj-stk
                         (list "JUMPPRM"
                               set-nargs
                               x)))
                 (let ((name (proc-obj-name proc)))
                   (if (and nb-args (targ-arg-check-avoidable? proc nb-args))
                     (targ-emit
                       (list 'SEQ
                             adj-stk
                             (list "JUMPLOCAL"
                                   set-nargs
                                   x
                                   (targ-make-glbl "" name))))
                     (targ-emit
                       (list 'SEQ
                             adj-stk
                             set-nargs
                             (list "JUMPLOCAL"
                                   '("NOTHING")
                                   x
                                   (targ-make-glbl 0 name)))))))))
            ((glo? opnd)
             (targ-emit
               (list 'SEQ
                     adj-stk
                     (cons (begin
                             (targ-wr-reg (+ targ-nb-arg-regs 1))
                             (if safe? "JUMPGLOSAFE" "JUMPGLONOTSAFE"))
                           (cons (if nb-args set-nargs '("NOTHING"))
                                 (cdr (targ-opnd opnd)))))))
            (else
             (targ-emit
               (list 'SEQ
                     adj-stk
                     (list (if nb-args
                             (begin
                               (targ-wr-reg (+ targ-nb-arg-regs 1))
                               (if safe? "JUMPGENSAFE" "JUMPGENNOTSAFE"))
                             "JUMPPRM")
                           (if nb-args set-nargs '("NOTHING"))
                           (targ-opnd opnd))))))

      (targ-end-fr))))

(define (targ-arg-check-avoidable? proc nb-args)
  (let ((x (proc-obj-call-pat proc)))
    (if (and (pair? x) (null? (cdr x))) ; proc accepts a fixed nb of args?
      (let ((arg-count (car x)))
        (= arg-count nb-args))          ; nb of arguments = nb of parameters?
      #f)))

;------------------------------------------------------------------------------

(define (targ-opnd opnd) ; fetch GVM operand

  (if (and targ-fp-cache-enabled? (or (reg? opnd) (stk? opnd)))
    (targ-fp-cache-write-if-dirty opnd))

  (cond ((reg? opnd)
         (let ((n (reg-num opnd)))
           (targ-rd-reg n)
           (cons 'R n)))

        ((stk? opnd)
         (targ-rd-fp)
         (list "STK" (- (stk-num opnd) targ-proc-fp)))

        ((glo? opnd)
         (let ((name (glo-name opnd)))
           (list "GLO"
                 (targ-use-glo name #f)
                 (targ-c-id-glo (symbol->string name)))))

        ((clo? opnd)
         (list "CLO"
               (targ-opnd (clo-base opnd))
               (clo-index opnd)))

        ((lbl? opnd)
         (let ((n (lbl-num opnd)))
           (list "LBL" (targ-ref-lbl-val n))))

        ((obj? opnd)
         (targ-use-obj (obj-val opnd)))

        (else
         (compiler-internal-error
           "targ-opnd, unknown 'opnd'" opnd))))

(define (targ-loc loc val) ; store GVM location
  (let ((x (targ-loc-no-invalidate loc val)))

    (if (and targ-fp-cache-enabled? (or (reg? loc) (stk? loc)))
      (targ-fp-cache-invalidate loc))

    x))

(define (targ-loc-no-invalidate loc val) ; store GVM location without
                                         ; invalidating flonum cache
  (cond ((reg? loc)
         (let ((n (reg-num loc)))
           (targ-wr-reg n)
           (list 'SET-R n val)))

        ((stk? loc)
         (targ-rd-fp)
         (list "SET_STK" (- (stk-num loc) targ-proc-fp) val))

        ((glo? loc)
         (let ((name (glo-name loc)))
           (list "SET_GLO"
                 (targ-use-glo name #t)
                 (targ-c-id-glo (symbol->string name))
                 val)))

        ((clo? loc)
         (list "SET_CLO"
               (targ-opnd (clo-base loc))
               (clo-index loc)))

        (else
         (compiler-internal-error
           "targ-loc, unknown 'loc'" loc))))

(define (targ-opnd-flo opnd) ; fetch unboxed flonum GVM operand
  (if (and targ-fp-cache-enabled? (or (reg? opnd) (stk? opnd)))
    (let ((x (targ-fp-cache-probe opnd)))
      (if x
        (cons 'FR x)
        (let ((x (cons 'FR (targ-fp-cache-enter opnd #f))))
          (targ-emit
            (list "SET_FR"
                  x
                  (list "FLOUNBOX" (targ-opnd opnd))))
          x)))
    (list "FLOUNBOX" (targ-opnd opnd))))

(define (targ-loc-flo loc val) ; store unboxed flonum to GVM location
  (if (and targ-fp-cache-enabled? (or (reg? loc) (stk? loc)))
    (begin
      (targ-fp-cache-invalidate loc)
      (let ((x (cons 'FR (targ-fp-cache-enter loc #t))))
        (list "SET_FR" x val)))
    (targ-loc loc (list "FLOBOX" val))))

(define (targ-seq-opnds opnds op)

  (define (seq opnds)
    (if (pair? opnds)
      (cons (list op (targ-opnd (car opnds)))
            (seq (cdr opnds)))
      '()))

  (cons 'SEQ (seq opnds)))

(define (targ-adjust-stack fs)
  (if (= targ-proc-fp fs)
    #f
    (let ((fp targ-proc-fp))
      (set! targ-proc-fp fs)
      (targ-wr-fp)
      (list "ADJFP" (- fs fp)))))

(define (targ-sn-opnd opnd sn)
  (cond ((stk? opnd)
         (max (stk-num opnd) sn))
        ((clo? opnd)
         (targ-sn-opnd (clo-base opnd) sn))
        (else
         sn)))

(define (targ-sn-opnds opnds sn)
  (if (pair? opnds)
    (targ-sn-opnd (car opnds) (targ-sn-opnds (cdr opnds) sn))
    sn))

(define (targ-sn-loc loc sn)
  (if loc
    (targ-sn-opnd loc sn)
    sn))

;------------------------------------------------------------------------------

; Floating point number cache management.

(define targ-fp-cache-enabled? #f)
(set! targ-fp-cache-enabled? #t)

(define (targ-fp-cache-init)
  (set! targ-fp-cache (vector 0 '#())))

(define (targ-fp-cache-size)
  (vector-ref targ-fp-cache 0))

(define (targ-fp-cache-write loc i)
  (targ-emit
    (targ-loc-no-invalidate loc (list "FLOBOX" (cons 'FR i)))))

(define (targ-fp-cache-write-if-dirty loc)
  (let ((v (vector-ref targ-fp-cache 1)))
    (let ((n (vector-length v)))
      (let loop ((i 0))
        (if (< i n)
          (let ((x (vector-ref v i)))
            (if (and x (cdr x) (eqv? (car x) loc))
              (begin
                (targ-fp-cache-write loc i)
                (set-cdr! x #f))
              (loop (+ i 1)))))))))

(define (targ-fp-cache-enter opnd dirty?) ; allocate new entry for opnd
  (let ((v1 (vector-ref targ-fp-cache 1)))
    (let ((n (vector-length v1)))
      (let loop1 ((i 0))
        (if (< i n)
          (if (vector-ref v1 i)
            (loop1 (+ i 1))
            (begin
              (vector-set! v1 i (cons opnd dirty?))
              i))
          (let ((v2 (make-vector (+ n 1))))
            (let loop2 ((i 0))
              (if (< i n)
                (begin
                  (vector-set! v2 i (vector-ref v1 i))
                  (loop2 (+ i 1)))))
            (vector-set! v2 n (cons opnd dirty?))
            (vector-set! targ-fp-cache 0 (+ n 1))
            (vector-set! targ-fp-cache 1 v2)
            n))))))

(define (targ-fp-cache-probe opnd) ; opnd must be a reg or stack slot
  (let ((v (vector-ref targ-fp-cache 1)))
    (let ((n (vector-length v)))
      (let loop ((i 0))
        (if (< i n)
          (let ((x (vector-ref v i)))
            (if (and x (eqv? (car x) opnd))
              i
              (loop (+ i 1))))
          #f)))))

(define (targ-fp-cache-invalidate opnd) ; opnd must be a reg or stack slot
  (let ((v (vector-ref targ-fp-cache 1)))
    (let ((n (vector-length v)))
      (let loop ((i 0))
        (if (< i n)
          (let ((x (vector-ref v i)))
            (if (and x (eqv? (car x) opnd))
              (vector-set! v i #f))
            (loop (+ i 1))))))))

;==============================================================================

; DATABASE OF PRIMITIVES

(set! targ-prim-proc-table
  (append
    (map (lambda (x)
           (cons (string->canonical-symbol (car x))
                 (apply make-proc-obj (car x) #t #f (cdr x))))
'(
("##c-code"  0            #t 0        (#f)   EXTENDED)
))
targ-prim-proc-table))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Procedure specialization

(define (targ-spec name specializer-maker)
  (let ((proc-name (string->canonical-symbol name)))
    (let ((proc (targ-get-prim-info name)))
      (if proc
        (proc-obj-specialize-set! proc (specializer-maker proc proc-name))
        (compiler-internal-error
          "targ-spec, unknown primitive:" name)))))

; Safe specialization

(define (targ-s name)
  (lambda (proc proc-name)
    (let ((spec (targ-get-prim-info name)))
      (lambda (decls args) spec))))

; Unsafe specialization

(define (targ-u name)
  (lambda (proc proc-name)
    (let ((spec (targ-get-prim-info name)))
      (lambda (decls args) (if (not (safe? decls)) spec proc)))))

; Safe arithmetic specialization

(define (targ-s-a fix-name flo-name)
  (targ-arith #t fix-name flo-name))

; Unsafe arithmetic specialization

(define (targ-u-a fix-name flo-name)
  (targ-arith #f fix-name flo-name))

(define (targ-arith fix-safe? fix-name flo-name)
  (lambda (proc proc-name)
    (let ((fix-spec (if fix-name (targ-get-prim-info fix-name) proc))
          (flo-spec (if flo-name (targ-get-prim-info flo-name) proc)))
      (lambda (decls args)
        (let ((arith (arith-implementation proc-name decls)))
          (cond ((eq? arith FIXNUM-sym)
                 (if (or fix-safe? (not (safe? decls))) fix-spec proc))
                ((eq? arith FLONUM-sym)
                 (if (not (safe? decls)) flo-spec proc))
                (else
                 proc)))))))

; Safe specialization for ##eqv?

(define (targ-s-**eqv?)
  (lambda (proc proc-name)
    (let ((spec (targ-get-prim-info "##eq?")))
      (lambda (decls args)
        (if (and (= (length args) 2)
                 (or (targ-eq-testable-object? (car args))
                     (targ-eq-testable-object? (cadr args))))
          spec
          proc)))))

(define (targ-eq-testable-object? obj)
  (and (not (eq? obj undef-object)) ; the undef-object denotes a non-constant
       (or (symbol-object? obj)
           (memq (targ-obj-type obj)
                 '(BOOLEAN NULL FIXNUM CHAR)))))

;------------------------------------------------------------------------------

(define (targ-op name descr)
  (descr (targ-get-prim-info name)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-APPLY-alloc compute-space flo-result? f)
  (targ-setup-inlinable-proc
    #f
    flo-result?
    (lambda (opnds sn)
      (targ-heap-reserve
        (compute-space (length opnds))
        (targ-sn-opnds opnds sn))
      (f opnds sn))))

(define (targ-APPLY-cons)
  (targ-APPLY-alloc
    (lambda (n) targ-pair-space)
    #f
    (targ-APPLY-simp-generator #f "CONS")))

(define (targ-APPLY-list)
  (targ-APPLY-alloc
    (lambda (n) (* n targ-pair-space))
    #f
    (lambda (opnds sn)
      (cond ((null? opnds)
             '("NUL"))
            ((null? (cdr opnds))
             (list "CONS" (targ-opnd (car opnds)) '("NUL")))
            (else
             (let ((rev-opnds (reverse opnds)))
               (targ-emit
                 (list "ALLOC_LIST"
                       (length opnds)
                       (targ-opnd (car rev-opnds))
                       (targ-seq-opnds (cdr rev-opnds) "ADD_LIST_ELEM"))))
             (list "GET_LIST" (length opnds)))))))

(define (targ-APPLY-make-cell)
  (targ-APPLY-alloc
    (lambda (n) targ-cell-space)
    #f
    (targ-APPLY-simp-generator #f "MAKECELL")))

(define (targ-APPLY-make-will)
  (targ-APPLY-alloc
    (lambda (n) targ-will-space)
    #f
    (targ-APPLY-simp-generator #f "MAKEWILL")))

(define (targ-APPLY-make-placeholder)
  (targ-APPLY-alloc
    (lambda (n) targ-placeholder-space)
    #f
    (targ-APPLY-simp-generator #f "MAKEPLACEHOLDER")))

(define (targ-APPLY-vector kind)

  (define (compute-space n)
    (case kind
      ((STRING)   (targ-string-space n))
      ((VECTOR8)  (targ-vector8-space n))
      ((VECTOR16) (targ-vector8-space (* n 2)))
      (else       (targ-vector-space n))))

  (define allocator-name
    (case kind
      ((STRING)   "ALLOC_STRING")
      ((VECTOR8)  "ALLOC_VECTOR8")
      ((VECTOR16) "ALLOC_VECTOR16")
      (else       "ALLOC_VECTOR")))

  (define getter-operation
    (case kind
      ((STRING)   "GET_STRING")
      ((VECTOR8)  "GET_VECTOR8")
      ((VECTOR16) "GET_VECTOR16")
      (else       "GET_VECTOR")))

  (define add-operation
    (case kind
      ((STRING)   "ADD_STRING_ELEM")
      ((VECTOR8)  "ADD_VECTOR8_ELEM")
      ((VECTOR16) "ADD_VECTOR8_ELEM")
      (else       "ADD_VECTOR_ELEM")))

  (targ-APPLY-alloc
    compute-space
    #f
    (lambda (opnds sn)
      (targ-emit
        (list allocator-name
              (length opnds)
              (targ-seq-opnds opnds add-operation)))
      (list getter-operation (length opnds)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-APPLY-force)
  (lambda (prim)
    (proc-obj-inlinable-set!
      prim
      (lambda (opnds loc sn)
        (let ((lbl (targ-new-lbl))
              (opnd (car opnds))
              (sn* (targ-sn-loc loc sn)))

          (targ-update-fr)
          (targ-emit (list "FORCE1"
                           (targ-ref-lbl-val lbl)
                           (targ-opnd opnd)))
          (targ-end-fr)
          (targ-gen-label-return* lbl 'RETURN-INTERNAL sn*)
          (targ-emit (list "FORCE2"))
          (if loc
            (targ-emit (targ-loc loc (list "FORCE3")))))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-APPLY-identity)
  (targ-setup-inlinable-proc*
    (lambda (opnds sn)
      (targ-opnd (car opnds)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-IFJUMP-simp name)
  (targ-setup-test-proc*
    (targ-IFJUMP-simp-generator #f name)))

(define (targ-IFJUMP-simpflo name)
  (targ-setup-test-proc*
    (targ-IFJUMP-simp-generator #t name)))

(define (targ-IFJUMP-fold name)
  (targ-setup-test-proc*
    (targ-IFJUMP-fold-generator #f name)))

(define (targ-IFJUMP-foldflo name)
  (targ-setup-test-proc*
    (targ-IFJUMP-fold-generator #t name)))

(define (targ-IFJUMP-APPLY name)
  (targ-setup-inlinable-proc*
    (targ-APPLY-simp-generator #f name)))

(define (targ-APPLY-simp side-effects? name)
  (targ-setup-inlinable-proc
    side-effects?
    #f
    (targ-APPLY-simp-generator #f name)))

(define (targ-APPLY-fold name0 name1 name2)
  (targ-setup-inlinable-proc
    #f
    #f
    (targ-APPLY-fold-generator name0 name1 name2 #f)))

(define (targ-APPLY-simpflo* name) ; primitive with flonum result
  (targ-APPLY-alloc
    (lambda (n) targ-flonum-space)
    #t
    (targ-APPLY-simp-generator #f name)))

(define (targ-APPLY-simpflo name) ; primitive with flonum args and result
  (targ-APPLY-alloc
    (lambda (n) targ-flonum-space)
    #t
    (targ-APPLY-simp-generator #t name)))

(define (targ-APPLY-foldflo name0 name1 name2) ; primitive with flonum args and
  (targ-APPLY-alloc                            ; result
    (lambda (n) targ-flonum-space)
    #t
    (targ-APPLY-fold-generator name0 name1 name2 #t)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-setup-test-proc* generator)
  (lambda (prim)
    ((targ-setup-test-proc generator)
     prim)
    ((targ-setup-inlinable-proc
       #f
       #f
       (lambda (opnds sn)
         (list "BOOL" (generator opnds sn))))
     prim)))

(define (targ-setup-test-proc generator)
  (lambda (prim)
    (proc-obj-test-set!
      prim
      (lambda (not? opnds fs)
        (let ((test (generator opnds fs)))
          (if not?
            (list "NOT" test)
            test))))))

(define (targ-IFJUMP-simp-generator flo? name)
  (lambda (opnds fs)
    (targ-IFJUMP-simp-gen opnds flo? name)))

(define (targ-IFJUMP-simp-gen opnds flo? name)
  (let loop ((l opnds) (args '()))
    (if (pair? l)
      (let ((opnd (car l)))
        (loop (cdr l)
              (cons (if flo? (targ-opnd-flo opnd) (targ-opnd opnd))
                    args)))
      (cons name (reverse args)))))

(define (targ-IFJUMP-fold-generator flo? name)
  (lambda (opnds fs)
    (targ-IFJUMP-fold-gen opnds flo? name)))

(define (targ-IFJUMP-fold-gen opnds flo? name)

  (define (multi-opnds opnds)
    (let* ((opnd1 (car opnds))
           (opnd2 (cadr opnds))
           (opnd1* (if flo? (targ-opnd-flo opnd1) (targ-opnd opnd1)))
           (opnd2* (if flo? (targ-opnd-flo opnd2) (targ-opnd opnd2)))
           (r (list name opnd1* opnd2*)))
      (if (pair? (cddr opnds))
        (list "AND" r (multi-opnds (cdr opnds)))
        r)))

  (cond ((or (not (pair? opnds))
             (not (pair? (cdr opnds))))
         1)
        (else
         (multi-opnds opnds))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (targ-setup-inlinable-proc* generator)
  (lambda (prim)
    ((targ-setup-test-proc
       (lambda (opnds fs)
         (list "NOT" (list "FALSEP" (generator opnds fs)))))
     prim)
    ((targ-setup-inlinable-proc #f #f generator)
     prim)))

(define (targ-setup-inlinable-proc side-effects? flo-result? generator)
  (lambda (prim)
    (proc-obj-inlinable-set!
      prim
      (lambda (opnds loc sn)
        (if side-effects? ; always generate code for side-effecting operations
          (let ((x (generator opnds sn)))
            (targ-emit ; if a result is needed, copy the first argument
              (if (and loc (not (eqv? (car opnds) loc)))
                (list 'SEQ
                      x
                      (targ-loc loc (targ-opnd (car opnds))))
                x)))
          (if loc ; only generate functional operations if result needed
            (let ((sn* (targ-sn-loc loc sn)))
              (let ((x (generator opnds sn*)))
                (targ-emit
                  (if flo-result?
                    (targ-loc-flo loc x)
                    (targ-loc loc x)))))))))))

(define (targ-APPLY-simp-generator flo? name)
  (lambda (opnds sn)
    (targ-APPLY-simp-gen opnds flo? name)))

(define (targ-APPLY-simp-gen opnds flo? name)
  (let loop ((l opnds) (args '()))
    (if (pair? l)
      (let ((opnd (car l)))
        (loop (cdr l)
              (cons (if flo? (targ-opnd-flo opnd) (targ-opnd opnd))
                    args)))
      (cons name (reverse args)))))

(define (targ-APPLY-fold-generator name0 name1 name2 flo?)
  (lambda (opnds sn)
    (targ-APPLY-fold-gen opnds name0 name1 name2 flo?)))

(define (targ-APPLY-fold-gen opnds name0 name1 name2 flo?)
  (if (not (pair? opnds))
    name0
    (let* ((o (car opnds))
           (r (if flo? (targ-opnd-flo o) (targ-opnd o))))
      (if (not (pair? (cdr opnds)))
        (list name1 r)
        (let loop ((l (cdr opnds)) (r r))
          (if (pair? l)
            (let ((opnd (car l)))
              (loop (cdr l)
                    (list name2
                          r
                          (if flo? (targ-opnd-flo opnd) (targ-opnd opnd)))))
            r))))))

;------------------------------------------------------------------------------

; Table of inlinable operations (for APPLY and IFJUMP gvm instructions)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##TYPE"                (targ-APPLY-simp #f "TYPE"))
(targ-op "##TYPE-CAST"           (targ-APPLY-simp #f "TYPECAST"))
(targ-op "##SUBTYPE"             (targ-APPLY-simp #f "SUBTYPE"))
(targ-op "##SUBTYPE-SET!"        (targ-APPLY-simp #t "SUBTYPESET"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##NOT"                 (targ-IFJUMP-simp "FALSEP"))
(targ-op "##NULL?"               (targ-IFJUMP-simp "NULLP"))
(targ-op "##UNASSIGNED?"         (targ-IFJUMP-simp "UNASSIGNEDP"))
(targ-op "##UNBOUND?"            (targ-IFJUMP-simp "UNBOUNDP"))
(targ-op "##EQ?"                 (targ-IFJUMP-simp "EQP"))
(targ-op "##EOF-OBJECT?"         (targ-IFJUMP-simp "EOFP"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##FIXNUM?"             (targ-IFJUMP-simp "FIXNUMP"))
(targ-op "##FLONUM?"             (targ-IFJUMP-simp "FLONUMP"))
(targ-op "##SPECIAL?"            (targ-IFJUMP-simp "SPECIALP"))
(targ-op "##PAIR?"               (targ-IFJUMP-simp "PAIRP"))
(targ-op "##SUBTYPED?"           (targ-IFJUMP-simp "SUBTYPEDP"))
(targ-op "##PROCEDURE?"          (targ-IFJUMP-simp "PROCEDUREP"))
(targ-op "##PLACEHOLDER?"        (targ-IFJUMP-simp "PLACEHOLDERP"))
(targ-op "##VECTOR?"             (targ-IFJUMP-simp "VECTORP"))
(targ-op "##SYMBOL?"             (targ-IFJUMP-simp "SYMBOLP"))
(targ-op "##PORT?"               (targ-IFJUMP-simp "PORTP"))
(targ-op "##RATNUM?"             (targ-IFJUMP-simp "RATNUMP"))
(targ-op "##CPXNUM?"             (targ-IFJUMP-simp "CPXNUMP"))
(targ-op "##STRUCTURE?"          (targ-IFJUMP-simp "STRUCTUREP"))
(targ-op "##STRING?"             (targ-IFJUMP-simp "STRINGP"))
(targ-op "##BIGNUM?"             (targ-IFJUMP-simp "BIGNUMP"))
(targ-op "##CHAR?"               (targ-IFJUMP-simp "CHARP"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##FIXNUM.+"            (targ-APPLY-fold "FIX0"  "FIXPOS" "FIXADD"))
(targ-op "##FIXNUM.*"            (targ-APPLY-fold "FIX1"  "FIXPOS" "FIXMUL"))
(targ-op "##FIXNUM.-"            (targ-APPLY-fold #f      "FIXNEG" "FIXSUB"))
(targ-op "##FIXNUM.QUOTIENT"     (targ-APPLY-fold #f      "FIXINV" "FIXQUO"))
(targ-op "##FIXNUM.REMAINDER"    (targ-APPLY-simp #f "FIXREM"))
(targ-op "##FIXNUM.MODULO"       (targ-APPLY-simp #f "FIXMOD"))
(targ-op "##FIXNUM.LOGIOR"       (targ-APPLY-fold "FIX0"  "FIXPOS" "FIXIOR"))
(targ-op "##FIXNUM.LOGXOR"       (targ-APPLY-fold "FIX0"  "FIXPOS" "FIXXOR"))
(targ-op "##FIXNUM.LOGAND"       (targ-APPLY-fold "FIXM1" "FIXPOS" "FIXAND"))
(targ-op "##FIXNUM.LOGNOT"       (targ-APPLY-simp #f "FIXNOT"))
(targ-op "##FIXNUM.ASHR"         (targ-APPLY-simp #f "FIXASHR"))
(targ-op "##FIXNUM.LSHR"         (targ-APPLY-simp #f "FIXLSHR"))
(targ-op "##FIXNUM.SHL"          (targ-APPLY-simp #f "FIXSHL"))

(targ-op "##FIXNUM.ZERO?"        (targ-IFJUMP-simp "FIXZEROP"))
(targ-op "##FIXNUM.POSITIVE?"    (targ-IFJUMP-simp "FIXPOSITIVEP"))
(targ-op "##FIXNUM.NEGATIVE?"    (targ-IFJUMP-simp "FIXNEGATIVEP"))
(targ-op "##FIXNUM.ODD?"         (targ-IFJUMP-simp "FIXODDP"))
(targ-op "##FIXNUM.EVEN?"        (targ-IFJUMP-simp "FIXEVENP"))
(targ-op "##FIXNUM.="            (targ-IFJUMP-fold "FIXEQ"))
(targ-op "##FIXNUM.<"            (targ-IFJUMP-fold "FIXLT"))
(targ-op "##FIXNUM.>"            (targ-IFJUMP-fold "FIXGT"))
(targ-op "##FIXNUM.<="           (targ-IFJUMP-fold "FIXLE"))
(targ-op "##FIXNUM.>="           (targ-IFJUMP-fold "FIXGE"))

(targ-op "##FIXNUM.->CHAR"       (targ-APPLY-simp #f "FIXTOCHR"))
(targ-op "##FIXNUM.<-CHAR"       (targ-APPLY-simp #f "FIXFROMCHR"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##FLONUM.->FIXNUM"     (targ-APPLY-simp #f "FLOTOFIX"))
(targ-op "##FLONUM.<-FIXNUM"     (targ-APPLY-simpflo* "FLOFROMFIX"))
(targ-op "##FLONUM.+"            (targ-APPLY-foldflo "FLO0" "FLOPOS" "FLOADD"))
(targ-op "##FLONUM.*"            (targ-APPLY-foldflo "FLO1" "FLOPOS" "FLOMUL"))
(targ-op "##FLONUM.-"            (targ-APPLY-foldflo #f     "FLONEG" "FLOSUB"))
(targ-op "##FLONUM./"            (targ-APPLY-foldflo #f     "FLOINV" "FLODIV"))
(targ-op "##FLONUM.ABS"          (targ-APPLY-simpflo "FLOABS"))
(targ-op "##FLONUM.FLOOR"        (targ-APPLY-simpflo "FLOFLOOR"))
(targ-op "##FLONUM.CEILING"      (targ-APPLY-simpflo "FLOCEILING"))
(targ-op "##FLONUM.ROUND"        (targ-APPLY-simpflo "FLOROUND"))
(targ-op "##FLONUM.EXP"          (targ-APPLY-simpflo "FLOEXP"))
(targ-op "##FLONUM.LOG"          (targ-APPLY-simpflo "FLOLOG"))
(targ-op "##FLONUM.SIN"          (targ-APPLY-simpflo "FLOSIN"))
(targ-op "##FLONUM.COS"          (targ-APPLY-simpflo "FLOCOS"))
(targ-op "##FLONUM.TAN"          (targ-APPLY-simpflo "FLOTAN"))
(targ-op "##FLONUM.ASIN"         (targ-APPLY-simpflo "FLOASIN"))
(targ-op "##FLONUM.ACOS"         (targ-APPLY-simpflo "FLOACOS"))
(targ-op "##FLONUM.ATAN"         (targ-APPLY-simpflo "FLOATAN"))
(targ-op "##FLONUM.SQRT"         (targ-APPLY-simpflo "FLOSQRT"))

(targ-op "##FLONUM.ZERO?"        (targ-IFJUMP-simpflo "FLOZEROP"))
(targ-op "##FLONUM.NEGATIVE?"    (targ-IFJUMP-simpflo "FLONEGATIVEP"))
(targ-op "##FLONUM.POSITIVE?"    (targ-IFJUMP-simpflo "FLOPOSITIVEP"))
(targ-op "##FLONUM.="            (targ-IFJUMP-foldflo "FLOEQ"))
(targ-op "##FLONUM.<"            (targ-IFJUMP-foldflo "FLOLT"))
(targ-op "##FLONUM.>"            (targ-IFJUMP-foldflo "FLOGT"))
(targ-op "##FLONUM.<="           (targ-IFJUMP-foldflo "FLOLE"))
(targ-op "##FLONUM.>="           (targ-IFJUMP-foldflo "FLOGE"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##CHAR=?"              (targ-IFJUMP-fold "CHAREQP"))
(targ-op "##CHAR<?"              (targ-IFJUMP-fold "CHARLTP"))
(targ-op "##CHAR>?"              (targ-IFJUMP-fold "CHARGTP"))
(targ-op "##CHAR<=?"             (targ-IFJUMP-fold "CHARLEP"))
(targ-op "##CHAR>=?"             (targ-IFJUMP-fold "CHARGEP"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##CONS"                (targ-APPLY-cons))
(targ-op "##SET-CAR!"            (targ-APPLY-simp #t "SETCAR"))
(targ-op "##SET-CDR!"            (targ-APPLY-simp #t "SETCDR"))
(targ-op "##CAR"                 (targ-IFJUMP-APPLY "CAR"))
(targ-op "##CDR"                 (targ-IFJUMP-APPLY "CDR"))
(targ-op "##CAAR"                (targ-IFJUMP-APPLY "CAAR"))
(targ-op "##CADR"                (targ-IFJUMP-APPLY "CADR"))
(targ-op "##CDAR"                (targ-IFJUMP-APPLY "CDAR"))
(targ-op "##CDDR"                (targ-IFJUMP-APPLY "CDDR"))
(targ-op "##CAAAR"               (targ-IFJUMP-APPLY "CAAAR"))
(targ-op "##CAADR"               (targ-IFJUMP-APPLY "CAADR"))
(targ-op "##CADAR"               (targ-IFJUMP-APPLY "CADAR"))
(targ-op "##CADDR"               (targ-IFJUMP-APPLY "CADDR"))
(targ-op "##CDAAR"               (targ-IFJUMP-APPLY "CDAAR"))
(targ-op "##CDADR"               (targ-IFJUMP-APPLY "CDADR"))
(targ-op "##CDDAR"               (targ-IFJUMP-APPLY "CDDAR"))
(targ-op "##CDDDR"               (targ-IFJUMP-APPLY "CDDDR"))
(targ-op "##CAAAAR"              (targ-IFJUMP-APPLY "CAAAAR"))
(targ-op "##CAAADR"              (targ-IFJUMP-APPLY "CAAADR"))
(targ-op "##CAADAR"              (targ-IFJUMP-APPLY "CAADAR"))
(targ-op "##CAADDR"              (targ-IFJUMP-APPLY "CAADDR"))
(targ-op "##CADAAR"              (targ-IFJUMP-APPLY "CADAAR"))
(targ-op "##CADADR"              (targ-IFJUMP-APPLY "CADADR"))
(targ-op "##CADDAR"              (targ-IFJUMP-APPLY "CADDAR"))
(targ-op "##CADDDR"              (targ-IFJUMP-APPLY "CADDDR"))
(targ-op "##CDAAAR"              (targ-IFJUMP-APPLY "CDAAAR"))
(targ-op "##CDAADR"              (targ-IFJUMP-APPLY "CDAADR"))
(targ-op "##CDADAR"              (targ-IFJUMP-APPLY "CDADAR"))
(targ-op "##CDADDR"              (targ-IFJUMP-APPLY "CDADDR"))
(targ-op "##CDDAAR"              (targ-IFJUMP-APPLY "CDDAAR"))
(targ-op "##CDDADR"              (targ-IFJUMP-APPLY "CDDADR"))
(targ-op "##CDDDAR"              (targ-IFJUMP-APPLY "CDDDAR"))
(targ-op "##CDDDDR"              (targ-IFJUMP-APPLY "CDDDDR"))

(targ-op "##LIST"                (targ-APPLY-list))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##MAKE-CELL"           (targ-APPLY-make-cell))
(targ-op "##CELL-REF"            (targ-IFJUMP-APPLY "CELLREF"))
(targ-op "##CELL-SET!"           (targ-APPLY-simp #t "CELLSET"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##WILL?"               (targ-IFJUMP-simp "WILLP"))
(targ-op "##MAKE-WILL"           (targ-APPLY-make-will))
(targ-op "##WILL-OWNER"          (targ-IFJUMP-APPLY "WILLOWNER"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##VECTOR"              (targ-APPLY-vector 'VECTOR))
(targ-op "##VECTOR-LENGTH"       (targ-APPLY-simp #f "VECTORLENGTH"))
(targ-op "##VECTOR-REF"          (targ-IFJUMP-APPLY "VECTORREF"))
(targ-op "##VECTOR-SET!"         (targ-APPLY-simp #t "VECTORSET"))
(targ-op "##VECTOR-SHRINK!"      (targ-APPLY-simp #t "VECTORSHRINK"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##STRING"              (targ-APPLY-vector 'STRING))
(targ-op "##STRING-LENGTH"       (targ-APPLY-simp #f "STRINGLENGTH"))
(targ-op "##STRING-REF"          (targ-APPLY-simp #f "STRINGREF"))
(targ-op "##STRING-SET!"         (targ-APPLY-simp #t "STRINGSET"))
(targ-op "##STRING-SHRINK!"      (targ-APPLY-simp #t "STRINGSHRINK"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##VECTOR8"             (targ-APPLY-vector 'VECTOR8))
(targ-op "##VECTOR8-LENGTH"      (targ-APPLY-simp #f "VECTOR8LENGTH"))
(targ-op "##VECTOR8-REF"         (targ-APPLY-simp #f "VECTOR8REF"))
(targ-op "##VECTOR8-SET!"        (targ-APPLY-simp #t "VECTOR8SET"))
(targ-op "##VECTOR8-SHRINK!"     (targ-APPLY-simp #t "VECTOR8SHRINK"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##VECTOR16"            (targ-APPLY-vector 'VECTOR16))
(targ-op "##VECTOR16-LENGTH"     (targ-APPLY-simp #f "VECTOR16LENGTH"))
(targ-op "##VECTOR16-REF"        (targ-APPLY-simp #f "VECTOR16REF"))
(targ-op "##VECTOR16-SET!"       (targ-APPLY-simp #t "VECTOR16SET"))
(targ-op "##VECTOR16-SHRINK!"    (targ-APPLY-simp #t "VECTOR16SHRINK"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##STRUCT-REF"          (targ-IFJUMP-APPLY "STRUCTREF"))
(targ-op "##STRUCT-SET!"         (targ-APPLY-simp #t "STRUCTSET"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##SYMBOL->STRING"      (targ-APPLY-simp #f "SYMBOLTOSTR"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##CLOSURE-CODE"        (targ-APPLY-simp #f "CLOSURECODE"))
(targ-op "##CLOSURE-REF"         (targ-APPLY-simp #f "CLOSUREREF"))
(targ-op "##CLOSURE-SET!"        (targ-APPLY-simp #t "CLOSURESET"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##GLOBAL-VAR-REF"      (targ-APPLY-simp #f "GLOBALVARREF"))
(targ-op "##GLOBAL-VAR-SET!"     (targ-APPLY-simp #t "GLOBALVARSET"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##MAKE-PLACEHOLDER"    (targ-APPLY-make-placeholder))
(targ-op "##FORCE"               (targ-APPLY-force))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op "##IDENTITY"            (targ-APPLY-identity))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-op
  "##C-CODE"
  (lambda (prim)
    (proc-obj-inlinable-set!
      prim
      (lambda (opnds loc sn)
        (targ-use-all-res)
        (let ((n (length opnds)))
          (if (= n 0)
            (compiler-error
              "\"##c-code\" needs at least one argument")
            (let ((code (car opnds))
                  (args (map (lambda (opnd) (targ-opnd opnd))
                             (cdr opnds))))
              (if (and (obj? code)
                       (string? (obj-val code)))
                (let loop ((i (- n 1))
                           (rev-args (reverse args))
                           (lst (list "#undef " c-id-prefix "RESULT"
                                      #\newline
                                      "#define " c-id-prefix "RESULT "
                                                 c-id-prefix "temp"
                                      #\newline
                                      (obj-val code)
                                      #\newline)))
                  (if (null? rev-args)
                    (begin
                      (targ-emit (cons 'APPEND lst))
                      (if loc
                        (targ-emit (targ-loc loc '("RESULT")))))
                    (let ((arg (car rev-args)))
                      (loop (- i 1)
                            (cdr rev-args)
                            (append
                              (list "#undef " c-id-prefix "ARG" i
                                    #\newline
                                    "#define " c-id-prefix "ARG" i " " arg
                                    #\newline)
                              lst)))))
                (compiler-error
                  "Argument 1 of \"##c-code\" must be a string constant")))))))))

;------------------------------------------------------------------------------

; Table of procedure specializations

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-spec "NOT"         (targ-s "##NOT"))
(targ-spec "NULL?"       (targ-s "##NULL?"))
(targ-spec "EQ?"         (targ-s "##EQ?"))
(targ-spec "EOF-OBJECT?" (targ-s "##EOF-OBJECT?"))

(targ-spec "PAIR?"       (targ-s "##PAIR?"))
(targ-spec "PROCEDURE?"  (targ-s "##PROCEDURE?"))
(targ-spec "VECTOR?"     (targ-s "##VECTOR?"))
(targ-spec "SYMBOL?"     (targ-s "##SYMBOL?"))
(targ-spec "STRING?"     (targ-s "##STRING?"))
(targ-spec "CHAR?"       (targ-s "##CHAR?"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-spec "ZERO?"       (targ-s-a "##FIXNUM.ZERO?"     "##FLONUM.ZERO?"))
(targ-spec "POSITIVE?"   (targ-s-a "##FIXNUM.POSITIVE?" "##FLONUM.POSITIVE?"))
(targ-spec "NEGATIVE?"   (targ-s-a "##FIXNUM.NEGATIVE?" "##FLONUM.NEGATIVE?"))
(targ-spec "ODD?"        (targ-s-a "##FIXNUM.ODD?"      #f))
(targ-spec "EVEN?"       (targ-s-a "##FIXNUM.EVEN?"     #f))

(targ-spec "+"           (targ-u-a "##FIXNUM.+"         "##FLONUM.+"))
(targ-spec "*"           (targ-u-a "##FIXNUM.*"         "##FLONUM.*"))
(targ-spec "-"           (targ-u-a "##FIXNUM.-"         "##FLONUM.-"))
(targ-spec "/"           (targ-u-a #f                   "##FLONUM./"))
(targ-spec "QUOTIENT"    (targ-u-a "##FIXNUM.QUOTIENT"  #f))
(targ-spec "REMAINDER"   (targ-u-a "##FIXNUM.REMAINDER" #f))
(targ-spec "MODULO"      (targ-u-a "##FIXNUM.MODULO"    #f))

(targ-spec "="           (targ-s-a "##FIXNUM.="         "##FLONUM.="))
(targ-spec "<"           (targ-s-a "##FIXNUM.<"         "##FLONUM.<"))
(targ-spec ">"           (targ-s-a "##FIXNUM.>"         "##FLONUM.>"))
(targ-spec "<="          (targ-s-a "##FIXNUM.<="        "##FLONUM.<="))
(targ-spec ">="          (targ-s-a "##FIXNUM.>="        "##FLONUM.>="))

(targ-spec "ABS"         (targ-u-a #f                   "##FLONUM.ABS"))
(targ-spec "FLOOR"       (targ-u-a #f                   "##FLONUM.FLOOR"))
(targ-spec "CEILING"     (targ-u-a #f                   "##FLONUM.CEILING"))
(targ-spec "TRUNCATE"    (targ-u-a #f                   "##FLONUM.TRUNCATE"))
(targ-spec "ROUND"       (targ-u-a #f                   "##FLONUM.ROUND"))
(targ-spec "EXP"         (targ-u-a #f                   "##FLONUM.EXP"))
(targ-spec "LOG"         (targ-u-a #f                   "##FLONUM.LOG"))
(targ-spec "SIN"         (targ-u-a #f                   "##FLONUM.SIN"))
(targ-spec "COS"         (targ-u-a #f                   "##FLONUM.COS"))
(targ-spec "TAN"         (targ-u-a #f                   "##FLONUM.TAN"))
(targ-spec "ASIN"        (targ-u-a #f                   "##FLONUM.ASIN"))
(targ-spec "ACOS"        (targ-u-a #f                   "##FLONUM.ACOS"))
(targ-spec "ATAN"        (targ-u-a #f                   "##FLONUM.ATAN"))
(targ-spec "SQRT"        (targ-u-a #f                   "##FLONUM.SQRT"))

(targ-spec "INTEGER->CHAR" (targ-s "##FIXNUM.->CHAR"))
(targ-spec "CHAR->INTEGER" (targ-s "##FIXNUM.<-CHAR"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-spec "CHAR=?"     (targ-s "##CHAR=?"))
(targ-spec "CHAR<?"     (targ-s "##CHAR<?"))
(targ-spec "CHAR>?"     (targ-s "##CHAR>?"))
(targ-spec "CHAR<=?"    (targ-s "##CHAR<=?"))
(targ-spec "CHAR>=?"    (targ-s "##CHAR>=?"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-spec "CONS"       (targ-s "##CONS"))
(targ-spec "SET-CAR!"   (targ-u "##SET-CAR!"))
(targ-spec "SET-CDR!"   (targ-u "##SET-CDR!"))
(targ-spec "CAR"        (targ-u "##CAR"))
(targ-spec "CDR"        (targ-u "##CDR"))
(targ-spec "CAAR"       (targ-u "##CAAR"))
(targ-spec "CADR"       (targ-u "##CADR"))
(targ-spec "CDAR"       (targ-u "##CDAR"))
(targ-spec "CDDR"       (targ-u "##CDDR"))
(targ-spec "CAAAR"      (targ-u "##CAAAR"))
(targ-spec "CAADR"      (targ-u "##CAADR"))
(targ-spec "CADAR"      (targ-u "##CADAR"))
(targ-spec "CADDR"      (targ-u "##CADDR"))
(targ-spec "CDAAR"      (targ-u "##CDAAR"))
(targ-spec "CDADR"      (targ-u "##CDADR"))
(targ-spec "CDDAR"      (targ-u "##CDDAR"))
(targ-spec "CDDDR"      (targ-u "##CDDDR"))
(targ-spec "CAAAAR"     (targ-u "##CAAAAR"))
(targ-spec "CAAADR"     (targ-u "##CAAADR"))
(targ-spec "CAADAR"     (targ-u "##CAADAR"))
(targ-spec "CAADDR"     (targ-u "##CAADDR"))
(targ-spec "CADAAR"     (targ-u "##CADAAR"))
(targ-spec "CADADR"     (targ-u "##CADADR"))
(targ-spec "CADDAR"     (targ-u "##CADDAR"))
(targ-spec "CADDDR"     (targ-u "##CADDDR"))
(targ-spec "CDAAAR"     (targ-u "##CDAAAR"))
(targ-spec "CDAADR"     (targ-u "##CDAADR"))
(targ-spec "CDADAR"     (targ-u "##CDADAR"))
(targ-spec "CDADDR"     (targ-u "##CDADDR"))
(targ-spec "CDDAAR"     (targ-u "##CDDAAR"))
(targ-spec "CDDADR"     (targ-u "##CDDADR"))
(targ-spec "CDDDAR"     (targ-u "##CDDDAR"))
(targ-spec "CDDDDR"     (targ-u "##CDDDDR"))

(targ-spec "LIST"       (targ-s "##LIST"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-spec "VECTOR"         (targ-s "##VECTOR"))
(targ-spec "VECTOR-LENGTH"  (targ-u "##VECTOR-LENGTH"))
(targ-spec "VECTOR-REF"     (targ-u "##VECTOR-REF"))
(targ-spec "VECTOR-SET!"    (targ-u "##VECTOR-SET!"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-spec "STRING"         (targ-s "##STRING"))
(targ-spec "STRING-LENGTH"  (targ-u "##STRING-LENGTH"))
(targ-spec "STRING-REF"     (targ-u "##STRING-REF"))
(targ-spec "STRING-SET!"    (targ-u "##STRING-SET!"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-spec "##STRUCTURE-REF"  (targ-u "##STRUCT-REF"))
(targ-spec "##STRUCTURE-SET!" (targ-u "##STRUCT-SET!"))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(targ-spec "SYMBOL->STRING" (targ-s "##SYMBOL->STRING"))

(targ-spec "TOUCH"          (targ-s "##FORCE"))
(targ-spec "FORCE"          (targ-s "##FORCE"))

(targ-spec "##EQV?"         (targ-s-**eqv?))

(targ-spec "CALL-WITH-CURRENT-CONTINUATION"
                            (targ-s "##CALL-WITH-CURRENT-CONTINUATION"))

;------------------------------------------------------------------------------
