;==============================================================================

; file: "_utils.scm"

; Copyright (C) 1994, 1995, 1996 by Marc Feeley, All Rights Reserved.

(include "fixnum.scm")

;------------------------------------------------------------------------------
;
; Utilities:
; ---------

(define (make-counter next limit limit-error)
  (lambda ()
    (if (< next limit)
      (let ((result next)) (set! next (+ next 1)) result)
      (limit-error))))

(define (pos-in-list x l)
  (let loop ((l l) (i 0))
    (cond ((not (pair? l)) #f)
          ((eq? (car l) x) i)
          (else            (loop (cdr l) (+ i 1))))))

(define (string-pos-in-list x l)
  (let loop ((l l) (i 0))
    (cond ((not (pair? l))      #f)
          ((string=? (car l) x) i)
          (else                 (loop (cdr l) (+ i 1))))))

(define (nth-after l n)
  (let loop ((l l) (n n))
    (if (> n 0)
      (loop (cdr l) (- n 1))
      l)))

(define (pair-up l1 l2)
  (define (pair l1 l2)
    (if (pair? l1)
      (cons (cons (car l1) (car l2)) (pair (cdr l1) (cdr l2)))
      '()))
  (pair l1 l2))

(define (last-pair l)
  (let loop ((l l))
    (if (pair? (cdr l))
      (loop (cdr l))
      l)))

(define (sort-list l <?)

  (define (mergesort l)

    (define (merge l1 l2)
      (cond ((null? l1) l2)
            ((null? l2) l1)
            (else
             (let ((e1 (car l1)) (e2 (car l2)))
               (if (<? e1 e2)
                 (cons e1 (merge (cdr l1) l2))
                 (cons e2 (merge l1 (cdr l2))))))))

    (define (split l)
      (if (or (null? l) (null? (cdr l)))
        l
        (cons (car l) (split (cddr l)))))

    (if (or (null? l) (null? (cdr l)))
      l
      (let* ((l1 (mergesort (split l)))
             (l2 (mergesort (split (cdr l)))))
        (merge l1 l2))))

  (mergesort l))

(define (lst->vector l)
  (let* ((n (length l))
         (v (make-vector n)))
    (let loop ((l l) (i 0))
      (if (pair? l)
        (begin
          (vector-set! v i (car l))
          (loop (cdr l) (+ i 1)))
        v))))

(define (vector->lst v)
  (let loop ((l '()) (i (- (vector-length v) 1)))
    (if (< i 0)
      l
      (loop (cons (vector-ref v i) l) (- i 1)))))

(define (lst->string l)
  (let* ((n (length l))
         (s (make-string n)))
    (let loop ((l l) (i 0))
      (if (pair? l)
        (begin
          (string-set! s i (car l))
          (loop (cdr l) (+ i 1)))
        s))))

(define (string->lst s)
  (let loop ((l '()) (i (- (string-length s) 1)))
    (if (< i 0)
      l
      (loop (cons (string-ref s i) l) (- i 1)))))

;------------------------------------------------------------------------------
;
; Exception processing
; --------------------

(define (with-exception-handling proc)
  (let ((old-exception-handler throw-to-exception-handler))
    (let ((val
            (call-with-current-continuation
              (lambda (cont)
                (set! throw-to-exception-handler cont)
                (proc)))))
    (set! throw-to-exception-handler old-exception-handler)
    val)))

(define (throw-to-exception-handler val)
  (fatal-err "Internal error, no exception handler at this point" val))

;------------------------------------------------------------------------------
;
; Compiler error processing
; -------------------------

(define (compiler-error msg . args)
  (newline)
  (display "*** ERROR -- ")
  (display msg)
  (for-each (lambda (x) (display " ") (write x)) args)
  (newline)
  (compiler-abort))

(define (compiler-user-error loc msg . args)
  (newline)
  (display "*** ERROR -- In ") (locat-show loc) (newline)
  (display "*** ") (display msg)
  (for-each (lambda (x) (display " ") (write x)) args)
  (newline)
  (compiler-abort))

(define (compiler-internal-error msg . args)
  (newline)
  (display "*** ERROR -- Compiler internal error detected") (newline)
  (display "*** in procedure ") (display msg)
  (for-each (lambda (x) (display " ") (write x)) args)
  (newline)
  (compiler-abort))

(define (compiler-limitation-error msg . args)
  (newline)
  (display "*** ERROR -- Compiler limit reached") (newline)
  (display "*** ") (display msg)
  (for-each (lambda (x) (display " ") (write x)) args)
  (newline)
  (compiler-abort))

(define (compiler-abort)
  (throw-to-exception-handler #f))

;------------------------------------------------------------------------------
;
; Transitive closure and topological sorting.

(define (make-gnode label edges) (vector label edges)) ; graph node
(define (gnode-label x) (vector-ref x 0))
(define (gnode-edges x) (vector-ref x 1))

(define (transitive-closure graph)
  (define changed? #f)
  (define (closure edges)
    (list->set (set-union edges
                          (apply set-union
                                 (map (lambda (label)
                                        (gnode-edges (gnode-find label graph)))
                                      (set->list edges))))))
  (let ((new-graph
          (set-map (lambda (x)
                     (let ((new-edges (closure (gnode-edges x))))
                       (if (not (set-equal? new-edges (gnode-edges x)))
                         (set! changed? #t))
                       (make-gnode (gnode-label x) new-edges)))
                   graph)))
    (if changed? (transitive-closure new-graph) new-graph)))

(define (gnode-find label graph)
  (define (find label l)
    (cond ((null? l)                         #f)
          ((eq? (gnode-label (car l)) label) (car l))
          (else                              (find label (cdr l)))))
  (find label (set->list graph)))

(define (topological-sort graph) ; topological sort fixed to handle cycles
  (if (set-empty? graph)
    '()
    (let ((to-remove (or (remove-no-edges graph) (remove-cycle graph))))
      (let ((labels (set-map gnode-label to-remove)))
        (cons labels
              (topological-sort
                (set-map (lambda (x)
                           (make-gnode
                             (gnode-label x)
                             (set-difference (gnode-edges x) labels)))
                         (set-difference graph to-remove))))))))

(define (remove-no-edges graph)
  (let ((nodes-with-no-edges
         (set-keep (lambda (x) (set-empty? (gnode-edges x))) graph)))
    (if (set-empty? nodes-with-no-edges)
      #f
      nodes-with-no-edges)))

(define (remove-cycle graph)
  (define (remove l)
    (let ((edges (gnode-edges (car l))))
      (define (equal-edges? x) (set-equal? (gnode-edges x) edges))
      (define (member-edges? x) (set-member? (gnode-label x) edges))
      (if (set-member? (gnode-label (car l)) edges)
        (let ((edge-graph (set-keep member-edges? graph)))
          (if (set-every? equal-edges? edge-graph)
            edge-graph
            (remove (cdr l))))
        (remove (cdr l)))))
  (remove (set->list graph)))

;------------------------------------------------------------------------------
;
; SET manipulation stuff
; ----------------------

(define (list->set list)    list)         ; convert list to set
(define (set->list set)     set)          ; convert set to list
(define (set-empty)         '())          ; the empty set
(define (set-empty? set)    (null? set))  ; is 'x' the empty set?
(define (set-member? x set) (memq x set)) ; is 'x' a member of the 'set'?
(define (set-singleton x)   (list x))     ; create a set containing only 'x'

(define (set-adjoin set x)                ; add the element 'x' to the 'set'
  (if (memq x set) set (cons x set)))

(define (set-remove set x)                ; remove the element 'x' from 'set'
  (cond ((null? set)       '())
        ((eq? (car set) x) (cdr set))
        (else              (cons (car set) (set-remove (cdr set) x)))))

(define (set-equal? s1 s2)                ; are 's1' and 's2' equal?
  (and (set-subset? s1 s2)
       (set-subset? s2 s1)))

(define (set-subset? s1 s2)               ; is 's1' a subset of 's2'?
  (cond ((null? s1)         #t)
        ((memq (car s1) s2) (set-subset? (cdr s1) s2))
        (else               #f)))

(define (set-difference set . other-sets) ; return difference of sets
  (define (difference s1 s2)
    (cond ((null? s1)         '())
          ((memq (car s1) s2) (difference (cdr s1) s2))
          (else               (cons (car s1) (difference (cdr s1) s2)))))
  (n-ary difference set other-sets))

(define (set-union . sets)                ; return union of sets
  (define (un s1 s2)
    (cond ((null? s1)         s2)
          ((memq (car s1) s2) (un (cdr s1) s2))
          (else               (cons (car s1) (un (cdr s1) s2)))))
  (define (union s1 s2)
    (if (set-smaller? s1 s2)
      (un s1 s2)
      (un s2 s1)))
  (n-ary union '() sets))

(define (set-intersection set . other-sets) ; return intersection of sets
  (define (int s1 s2)
    (cond ((null? s1)         '())
          ((memq (car s1) s2) (cons (car s1) (int (cdr s1) s2)))
          (else               (int (cdr s1) s2))))
  (define (intersection s1 s2)
    (if (set-smaller? s1 s2)
      (int s1 s2)
      (int s2 s1)))
  (n-ary intersection set other-sets))

(define (n-ary function first rest)
  (if (null? rest)
    first
    (n-ary function (function first (car rest)) (cdr rest))))

(define (set-keep keep? set)
  (cond ((null? set)       '())
        ((keep? (car set)) (cons (car set) (set-keep keep? (cdr set))))
        (else              (set-keep keep? (cdr set)))))

(define (set-every? pred? set)
  (or (null? set)
      (and (pred? (car set))
           (set-every? pred? (cdr set)))))

(define (set-map proc set)
  (if (null? set)
    '()
    (cons (proc (car set)) (set-map proc (cdr set)))))

(define (set-smaller? s1 s2)
  (if (null? s1)
    (not (null? s2))
    (if (null? s2)
      #f
      (set-smaller? (cdr s1) (cdr s2)))))

;------------------------------------------------------------------------------
;
; QUEUE manipulation stuff
; ------------------------

(define (list->queue list)    ; convert list to queue
  (cons list (if (pair? list) (last-pair list) '())))

(define (queue->list queue)   ; convert queue to list
  (car queue))

(define (queue-empty)         ; the empty queue
  (cons '() '()))

(define (queue-empty? queue)  ; is the queue empty?
  (null? (car queue)))

(define (queue-get! queue)    ; remove the first element of the queue
  (if (null? (car queue))
    (compiler-internal-error "queue-get!, queue is empty")
    (let ((x (caar queue)))
      (set-car! queue (cdar queue))
      (if (null? (car queue)) (set-cdr! queue '()))
      x)))

(define (queue-put! queue x)  ; add an element to the end of the queue
  (let ((entry (cons x '())))
    (if (null? (car queue))
      (set-car! queue entry)
      (set-cdr! (cdr queue) entry))
    (set-cdr! queue entry)
    x))

;==============================================================================
