/* file: "mem.c" */

/* Copyright (C) 1994-1996 by Marc Feeley, All Rights Reserved. */

#define ___VERSION 9
#include "gambit.h"
___BEGIN_MODULE

___DEMAND_GLO(___G__23_23gc_2dreport)

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "setup.h"
#include "mem.h"

/* Defining the symbol CONSISTENCY_CHECKS will cause the GC to perform */
/* checks that detect when the heap is in an inconsistent state.  This */
/* is useful to detect bugs in the GC and the rest of the system.      */

#define CONSISTENCY_CHECKS_not

/*---------------------------------------------------------------------------*/

/* Constants for allocation of aligned blocks: */

#define ___MALLOC_UP
#define ___MALLOC_ALIGNMENT 1

/*---------------------------------------------------------------------------*/

/* Allocation and reclamation of aligned blocks of memory. */


/* '___malloc_aligned_block (size, multiplier, modulus)' allocates an     */
/* aligned block of memory from the C heap using 'malloc'.  'size' is the */
/* size of the block in bytes and 'multiplier' and 'modulus' indicate its */
/* alignment.  'multiplier' must be a power of two and                    */
/* 0<=modulus<multiplier.  The pointer returned will be equal to          */
/* i*multiplier+modulus for some 'i'.                                     */

void *___malloc_aligned_block (size, multiplier, modulus)
unsigned long size;
unsigned int multiplier, modulus;
{
  void *container;
  unsigned int extra; /* space for alignment to multiplier */
  if (multiplier < sizeof(void*))
    multiplier = sizeof(void*); /* sizeof(void*) assumed to be a power of 2 */
  if (multiplier > ___MALLOC_ALIGNMENT)
    extra = multiplier - ___MALLOC_ALIGNMENT;
  else
    extra = 0;
  if (modulus < sizeof(void*))
    extra += sizeof(void*);
  container = malloc (extra + modulus + size);
  if (container == 0)
    return 0;
#if ___MALLOC_ALIGNMENT != 1
  else if ((long)container & ___MALLOC_ALIGNMENT-1 != 0)
    fatal_error ("Alignment problem with malloc");
#endif
  else
    {
      void *ptr = (void*)(((long)container+extra&-(long)multiplier) + modulus);
      *((void**)((long)ptr-sizeof(void*) & -sizeof(void*))) = container;
      return ptr;
    }
}


/* '___free_aligned_block (ptr)' reclaims the aligned block of memory */
/* 'ptr' that was allocated using '___malloc_aligned_block'.          */

void ___free_aligned_block (ptr)
void *ptr;
{
  free (*((void**)((long)ptr-sizeof(void*) & -sizeof(void*))));
}


/*---------------------------------------------------------------------------*/

/* Arena implementation. */

/* An arena is an area of memory, distibuted in multiple non-contiguous  */
/* sections, that is used to allocate movable Scheme objects.  Each      */
/* section is of the same size and is allocated from the C heap.         */
/* A section consists of two halves which are alternately the from-space */
/* and to-space.  The number of sections can expand and contract to      */
/* accommodate the needs of the program.                                 */

typedef struct asect
  {
    int pos;            /* position in 'sections' array */
    ___WORD *alloc;     /* allocation pointer, grows towards high addresses */
    struct asect *prev; /* previous section in list of sections */
    struct asect *next; /* next section in list of sections */
    ___WORD base[1];    /* space for semispaces */
  } arena_section;

#define sizeof_arena_section(n) (sizeof(arena_section)+((n)-1)*sizeof(___WORD))

typedef struct
  {
    unsigned int max_nb_sections; /* actual size of 'sections' array */
    unsigned int nb_sections;     /* number of sections */
    arena_section *head, *tail;   /* head and tail of list of sections */
    arena_section *sections[1];   /* each section ordered by address */
                                  /* (increasing order if ___MALLOC_UP is */
                                  /* defined otherwise decreasing order) */
  } arena;

#define sizeof_arena(n) (sizeof(arena)+((n)-1)*sizeof(arena_section*))


/* 'find_arena_section (a, ptr)' finds the index of the arena section that   */
/* contains the pointer 'ptr'.  More precisely, if ___MALLOC_UP is defined,  */
/* it returns the integer 'i' (-1<=i<=n-1) such that 'ptr' is between the    */
/* start of section i and section i+1.  -1 is returned if 'ptr' is lower     */
/* than the lowest section and 'n' is returned if 'ptr' is not lower than    */
/* the highest section.  If ___MALLOC_UP is not defined, it returns the      */
/* integer 'i' (0<=i<=n) such that 'ptr' is between the start of section i   */
/* and section i-1.  n is returned if 'ptr' is lower than the lowest section */
/* and 0 is returned if 'ptr' is not lower than the highest section.         */

static int find_arena_section (a, ptr)
arena *a;
void *ptr;
{
  unsigned int ns = a->nb_sections;
  arena_section **sections = a->sections;
#ifdef ___MALLOC_UP
  if ((ns == 0) || (ptr < (void*)sections[0]))
    return -1;
#else
  if ((ns == 0) || (ptr < (void*)sections[ns-1]))
    return ns;
#endif
  else
    {
      /* binary search */
      unsigned int lo = 0, hi = ns-1;

      /* loop invariant: lo <= find_arena_section (a, ptr) <= hi */
      while (lo < hi)
        {
          unsigned int mid = (lo+hi)/2; /* lo <= mid < hi */
#ifdef ___MALLOC_UP
          if (ptr < (void*)sections[mid+1]) hi = mid; else lo = mid+1;
#else
          if (ptr < (void*)sections[mid]) lo = mid+1; else hi = mid;
#endif
        }
      return lo;
    }
}


/* 'adjust_arena (ap, n)' contracts or expands the arena 'ap' so that it    */
/* contains only 'n' sections.  When the arena is contracted, the sections  */
/* at the top of the 'sections' array will be reclaimed (these are the      */
/* highest sections if ___MALLOC_UP is defined, otherwise they are the      */
/* lowest sections).  When expanding the arena there may not be enough      */
/* memory in the C heap to allocate new sections so the operation may fail. */
/* However 'adjust_arena' will always leave the arena in a consistent state */
/* and there will be at least as many sections as when the expansion was    */
/* started.  Failure can be detected by checking the 'nb_sections' field.   */

static void adjust_arena (ap, n)
arena **ap;
unsigned int n;
{
  unsigned int mns, ns;
  arena *a = *ap;
  arena_section *hd, *tl;

  if (a == 0)
    {
      mns = 0;
      ns = 0;
      hd = 0;
      tl = 0;
    }
  else
    {
      mns = a->max_nb_sections;
      ns = a->nb_sections;
      hd = a->head;
      tl = a->tail;
    }

  if ((a == 0) || (n > mns))
    {
      /* must allocate a new arena structure */

      arena *new_a;
      int i;

      while (n > mns) /* grow max_nb_sections until big enough */
        mns = 2*mns + 1;

      new_a = ___malloc_aligned_block (sizeof_arena(mns), ___WS, 0);

      if (new_a == 0)
        return;

      new_a->max_nb_sections = mns;
      new_a->nb_sections = ns;
      new_a->head = hd;
      new_a->tail = tl;
      for (i=ns-1; i>=0; i--)
        new_a->sections[i] = a->sections[i];

      if (a != 0)
        ___free_aligned_block (a);

      a = new_a;

      *ap = a;
    }

  if (n < ns)
    {
      /* contraction of the arena */

      unsigned int i, j;

      while (ns > n)
        {
          arena_section *s = tl;
          tl = tl->prev;
          if (tl == 0)
            hd = 0;
          else
            tl->next = 0;
          for (j=s->pos; j<ns-1; j++)
            {
              a->sections[j] = a->sections[j+1];
              a->sections[j]->pos = j;
            }
          ___free_aligned_block (s);
          ns--;
        }

      a->nb_sections = ns;
      a->head = hd;
      a->tail = tl;

      /* contraction of the arena structure is not implemented */
    }
  else
    {
      /* expansion of the arena */

      int i, j;

      while (ns < n)
        {
          arena_section *s =
            (arena_section*)___malloc_aligned_block
                              (sizeof_arena_section(___SECTION_HALFSIZE<<1),
                               ___WS,
                               0);
          if (s == 0)
            return;
          i = find_arena_section (a, (void*)s);
#ifdef ___MALLOC_UP
          i++;
#endif
          for (j=ns; j>i; j--)
            {
              a->sections[j] = a->sections[j-1];
              a->sections[j]->pos = j;
            }
          a->sections[i] = s;
          if (tl == 0)
            hd = s;
          else
            tl->next = s;
          s->pos = i;
          s->prev = tl;
          s->next = 0;
          tl = s;
          a->nb_sections = ++ns;
          a->head = hd;
          a->tail = tl;
        }
    }
}


/*---------------------------------------------------------------------------*/

/* Constants related to representation of permanent and still objects: */

#ifdef ___USE_HANDLES
#define ___PERM_HAND_OFS 0
#define ___PERM_BODY_OFS 2
#else
#define ___PERM_HAND_OFS ___PERM_BODY_OFS
#define ___PERM_BODY_OFS 1
#endif

#define ___STILL_LINK_OFS 0
#define ___STILL_REFCOUNT_OFS 1
#define ___STILL_MARK_OFS 2
#ifdef ___USE_HANDLES
#define ___STILL_HAND_OFS 3
#define ___STILL_BODY_OFS 5
#else
#define ___STILL_HAND_OFS ___STILL_BODY_OFS
#define ___STILL_BODY_OFS 4
#endif


long ___semispace_fudge;  /* size of  semispace fudge area in words */
long ___overflow_reserve; /* words reserved for heap overflows      */
long ___stack_size;       /* size of stack in words                 */
long ___stack_fudge;      /* size of stack fudge area in words      */

___processor pstate; /* processor state (e.g. regs, allocation ptrs) */

static long max_heap_size; /* upper bound on size of heap in words */

static long words_non_movable; /* words allocated for non-movable objs */

static long words_prev_sections; /* words allocated for movable objs */
                                 /* not including current section    */

/* total words allocated including current section */
#define words_movable ((words_prev_sections+(alloc_ptr-start_ptr))*2)

/* total words allocated including current section */
#define words_allocated (words_non_movable+words_movable)

/* maximum number of words that can be allocated */
#define words_available \
(the_arena->nb_sections * 2 * (___SECTION_HALFSIZE-___semispace_fudge+1) \
- ___overflow_reserve * 2)

static ___WORD still_objs;         /* list of still objects */
static ___WORD still_objs_to_scan; /* still objects remaining to scan */

static arena *the_arena;           /* the allocation arena */

static int tospace_at_top;         /* location of tospace in each section */

static arena_section *alloc_section; /* section currently being allocated in */
static int nb_prev_sections;       /* nb of sections filled before current */
static ___WORD *start_ptr;         /* start of allocation in current section */
static ___WORD *alloc_ptr;         /* allocation pointer in current section */
static ___WORD *alloc_limit;       /* allocation limit in current section */

static arena_section *scan_section; /* section currently being scanned */
static ___WORD *scan_ptr;          /* scan pointer in section being scanned */


/* '___alloc_scmobj (subtype, bytes, kind)' allocates a permanent or still */
/* Scheme object (depending on 'kind') of subtype 'subtype' with a body    */
/* containing 'bytes' bytes, and returns it as an encoded Scheme object.   */
/* A permanent object is allocated when 'kind' = ___PERM and a still       */
/* object is allocated when 'kind' = ___STILL.  The initialization of      */
/* the object's body must be done by the caller.  In the case of still     */
/* objects this initialization must be done before the next allocation     */
/* is requested.  The 'refcount' field of still objects is initially 1.    */
/* The value ___FAL is returned if the object can not be allocated.        */

___WORD ___alloc_scmobj (subtype, bytes, kind)
unsigned int subtype;
unsigned long bytes;
unsigned int kind;
{
  void *ptr;
  unsigned long words = (kind==___PERM ? ___PERM_BODY_OFS : ___STILL_BODY_OFS)
                        + ___WORDS(bytes);

  alloc_ptr = pstate->hp;
  words_non_movable += words;
  if (words_allocated > words_available)
    if (___gc () ||
        words_allocated > words_available)
      {
        words_non_movable -= words;
        ___overflow_reserve = ___OVERFLOW_RESERVE;
        return ___FAL;
      }

#if ___WS == 4
  if (subtype == ___sFLONUM)
    ptr = ___malloc_aligned_block (words<<___LWS,
                                   ___FLONUM_SIZE<<___LWS,
                                   kind == ___PERM
                                   ? (-___PERM_BODY_OFS&___FLONUM_SIZE-1)
                                     << ___LWS
                                   : (-___STILL_BODY_OFS&___FLONUM_SIZE-1)
                                     << ___LWS);
  else
#endif
    ptr = ___malloc_aligned_block (words<<___LWS, ___WS, 0);

  if (ptr == 0)
    {
      words_non_movable -= words;
      return ___FAL;
    }
  else if (kind == ___PERM)
    {
      ___WORD *base = (___WORD*)ptr;

#ifdef ___USE_HANDLES
      base[___PERM_HAND_OFS] = (___WORD)(base+___PERM_BODY_OFS-___BODY_OFS);
#endif
      base[___PERM_BODY_OFS-1] = ___MAKE_HD(bytes, subtype, ___RCN, ___PERM);

      return ___TAG((base + ___PERM_HAND_OFS - ___BODY_OFS),
                    (subtype == ___sPAIR ? ___tPAIR : ___tSUBTYPED));
    }
  else
    {
      ___WORD *base = (___WORD*)ptr;

      base[___STILL_LINK_OFS] = still_objs;
      still_objs = (___WORD)base;
      base[___STILL_REFCOUNT_OFS] = 1;
#ifdef ___USE_HANDLES
      base[___STILL_HAND_OFS] = (___WORD)(base+___STILL_BODY_OFS-___BODY_OFS);
#endif
      base[___STILL_BODY_OFS-1] = ___MAKE_HD(bytes, subtype, ___RCN, ___STILL);

      return ___TAG((base + ___STILL_HAND_OFS - ___BODY_OFS),
                    (subtype == ___sPAIR ? ___tPAIR : ___tSUBTYPED));
    }
}


/* '___make_pair (car, cdr, kind)' creates a Scheme pair having the values   */
/* 'car' and 'cdr' in its CAR and CDR fields.  A permanent or still object   */
/* is allocated, depending on 'kind' (___PERM for permanent object, ___STILL */
/* for still object).  The value ___FAL is returned if the object can not be */
/* allocated.                                                                */

___WORD ___make_pair (car, cdr, kind)
___WORD car, cdr;
unsigned int kind;
{
  ___WORD obj = ___alloc_scmobj (___sPAIR, ___PAIR_SIZE<<___LWS, kind);
  if (obj == ___FAL)
    return ___FAL;
  else
    {
      ___PAIR_CAR(obj) = car;
      ___PAIR_CDR(obj) = cdr;
      return obj;
    }
}


/* '___make_vector (length, init, kind)' creates a Scheme vector of length */
/* 'length' and initialized with the value 'init'.  A permanent or still   */
/* object is allocated, depending on 'kind' (___PERM for permanent object, */
/* ___STILL for still object).  The value ___FAL is returned if the object */
/* can not be allocated.                                                   */

___WORD ___make_vector (length, init, kind)
unsigned long length;
___WORD init;
unsigned int kind;
{
  if (length >= 1<<32-___LF-___LWS)
    return ___FAL;
  else
    {
      ___WORD obj = ___alloc_scmobj (___sVECTOR, length<<___LWS, kind);
      if (obj == ___FAL)
        return ___FAL;
      else
        {
          unsigned int i;
          for (i=0; i<length; i++)
            ___FIELD(obj, i) = init;
          return obj;
        }
    }
}


/* '___make_string (length, init, kind)' creates a Scheme string of length   */
/* 'length' and initialized with the character 'init'.  A permanent or still */
/* object is allocated, depending on 'kind' (___PERM for permanent object,   */
/* ___STILL for still object).  The value ___FAL is returned if the object   */
/* can not be allocated.                                                     */

___WORD ___make_string (length, init, kind)
unsigned long length;
char init;
unsigned int kind;
{
  if (length >= (1<<32-___LF)-1)
    return ___FAL;
  else
    {
      ___WORD obj = ___alloc_scmobj (___sSTRING, length+1, kind);
      if (obj == ___FAL)
        return ___FAL;
      else
        {
          unsigned int i;
          ((char*)___BODY_AS(obj,___tSUBTYPED))[length] = '\0';
          for (i=0; i<length; i++)
            ((char*)___BODY_AS(obj,___tSUBTYPED))[i] = init;
          return obj;
        }
    }
}


/* '___string_to_scheme (str, kind)' converts the null-terminated C string */
/* 'str' to a Scheme string.  A permanent or still object is allocated,    */
/* depending on 'kind' (___PERM for permanent object, ___STILL for still   */
/* object).  The value ___FAL is returned if the object can not be         */
/* allocated.                                                              */

___WORD ___string_to_scheme (str, kind)
char *str;
unsigned int kind;
{
  unsigned long length = strlen (str);
  ___WORD obj = ___alloc_scmobj (___sSTRING, length+1, kind);
  if (obj == ___FAL)
    return ___FAL;
  else
    {
      memcpy (___BODY_AS(obj,___tSUBTYPED), str, length+1);
      return obj;
    }
}


/* '___double_to_scheme (val, kind)' converts a C double floating-point  */
/* number to a Scheme flonum.  A permanent or still object is allocated, */
/* depending on 'kind' (___PERM for permanent object, ___STILL for still */
/* object).  The value ___FAL is returned if the object can not be       */
/* allocated.                                                            */

___WORD ___double_to_scheme (val, kind)
double val;
unsigned int kind;
{
  ___WORD obj = ___alloc_scmobj (___sFLONUM, ___FLONUM_SIZE<<___LWS, kind);
  if (obj == ___FAL)
    return ___FAL;
  else
    {
      ___FLONUM_VAL(obj) = val;
      return obj;
    }
}


/* '___still_obj_refcount_inc (obj)' increments the reference count of */
/* the still object 'obj'.                                             */

void ___still_obj_refcount_inc (obj)
___WORD obj;
{
  ___UNTAG(obj)[___BODY_OFS - ___STILL_BODY_OFS + ___STILL_REFCOUNT_OFS]++;
}


/* '___still_obj_refcount_dec (obj)' decrements the reference count of */
/* the still object 'obj'.                                             */

void ___still_obj_refcount_dec (obj)
___WORD obj;
{
  ___UNTAG(obj)[___BODY_OFS - ___STILL_BODY_OFS + ___STILL_REFCOUNT_OFS]--;
}


/*---------------------------------------------------------------------------*/

/*

Object representation.

Memory allocated Scheme objects can be allocated using one of three
allocation strategies:

   Permanently allocated:
     These objects, called 'permanent objects' for short, are never
     moved or reclaimed, and all pointers to memory allocated
     objects they contain must point to permanent objects.  As a
     consequence, the GC does not have to scan permanent objects.
     Permanent objects can be allocated on the C heap using 'malloc',
     but they are typically allocated in C global variables and
     structures that are set up when the program starts up or when a
     module is dynamically loaded.

   Still dynamically allocated:
     These objects, called 'still objects' for short, are allocated on
     the C heap using 'malloc'.  Still objects are never moved but they
     can be reclaimed by the GC.  A mark-and-sweep GC is used to
     garbage-collect still objects.

   Movable dynamically allocated:
     These objects, called 'movable objects' for short, are allocated
     in the arena which is managed by a compacting GC.  The GC can
     move and reclaim movable objects.

Scheme objects are encoded using integers of type ___WORD (defined as
a 'long' integer on most systems).  A ___WORD either encodes an
immediate value or encodes a pointer when the object is memory
allocated.  The two lower bits of a ___WORD contain a primary type tag
for the object and the other bits contain the immediate value or the
pointer.  Because all memory allocated objects are aligned on ___WORD
boundaries (and a ___WORD is either 4 or 8 bytes), the two lower bits
of pointers are zero and can be used to store the tag without reducing
the address space.  The four tags are:

 immediate:
   ___tFIXNUM    object is a small integer (fixnum)
   ___tSPECIAL   object is a boolean, character, or other immediate

 memory allocated:
   ___tPAIR      object is a pair
   ___tSUBTYPED  object is memory allocated but not a pair

A special type of object exists to support object finalization: 'will'
objects.  Each will contains an object (the will's owner) and a
procedure (the will's action procedure).  The GC counts the number of
references to each will's owner.  When there is only a single
reference to the owner (i.e. the only reference is from the will) then
the action procedure will be invoked.  Some complexities in the object
representation are necessary to support this finalization mechanism
efficiently.

All memory allocated objects, including pairs, are composed of at
least a head and a body.  The head is a single ___WORD that contains
two "head" tag bits (the two lower bits), followed by a one-bit
reference count, a subtype tag (the next 5 bits), and the length of
the object in bytes (the remaining bits).  The head immediately
precedes the body of the object, which contains the rest of the
information associated with the object.  Depending on the subtype, the
body can contain raw binary data (such as when the object is a string)
and Scheme objects (such as when the object is a vector).  Memory
allocated objects have the following layout:

     _head_   _____body______
    /      \ /               \
   +--------+--------+--------+
   |llllssrt|        |        |
   +--------+--------+--------+
     ^  ^ ^^
     |  | ||
length  | | head tag
  subtype ref_count

The four head tags are:

   ___PERM     (P) the object is a permanent object
   ___STILL    (S) the object is a still object
   ___MOVABLE  (M) the object is a movable object
   ___FORW     (F) the object has been moved by the GC

The one-bit reference count is used to implement wills.  The GC sets
this bit to 0 when the only references to the object are from the
owner slot of wills (1 thus means that there is at least one other
kind of reference).

Permanent objects have the following layout:

     _head_   _____body______
    /      \ /               \
   +--------+--------+--------+
   |       P|        |        |
   +--------+--------+--------+

Still objects have the following layout:

     _link_   _ref__   _mark_   _head_   _____body______
    /      \ / count\ /      \ /      \ /               \
   +--------+--------+--------+--------+--------+--------+
   |        |        |        |       S|        |        |
   +--------+--------+--------+--------+--------+--------+

All still objects are linked in a list using the 'link' field.  The
'refcount' field contains a reference count, which counts the number
of pointers to this object that are hidden from the GC (typically
these hidden pointers are in C data structures).  When 'refcount' is
zero, the object will survive a GC only if it is pointed to by a GC
root or a live Scheme object.  The 'mark' field is used by the GC to
indicate that the object has been marked (at the start of a GC it is
set to -1).  The 'mark' field links all objects that have been marked
but have not yet been scanned.  It contains a pointer to the next
still object that needs to be scanned.

Movable objects have the following layout:

     _head_   _____body______
    /      \ /               \
   +--------+--------+--------+
   |       M|        |        |
   +--------+--------+--------+

When a movable object is moved by the GC, the head is replaced with a
pointer to the copy, tagged with ___FORW.


Layout of body.

     _head_   __________body__________
    /      \ /                        \
   +--------+--------+--------+--------+
   |        | field_0| field_1|  etc.  |
   +--------+--------+--------+--------+

Some object types have bodies that only contain pointers to other
Scheme objects.  For example, pairs have two fields (car and cdr) and
vectors have one field per element.  Other object types have bodies
that only contain raw binary data (such as strings and bignums).  The
remaining object types have bodies that contain both pointers to
Scheme objects and raw binary data.  Their layout is summarized below.

Symbols:
    subtype = ___sSYMBOL
    field_0 = name (a Scheme string)
    field_1 = hash code (fixnum)
    field_2 = C pointer to global variable (0 if none allocated)

Procedures:

  non-closures (toplevel procedures and return addresses):
    subtype = ___sPROCEDURE
    field_0 = C pointer to field_0 - ___BODY_OFS
    field_1 = C pointer to host C procedure
    field_2 = return frame descriptor

  closures:
    subtype = ___sPROCEDURE
    field_0 = C pointer to field_0 of entry procedure - ___BODY_OFS
    field_1 = free variable 1
    field_2 = free variable 2
    ...

Wills:
    subtype = ___sWILL
    field_0 = owner object
    field_1 = action procedure
    field_2 = next will in will list

Continuations:
    subtype = ___sCONTINUATION
    field_0 = first frame (C pointer to stack at first and then Scheme obj)
    field_1 = return address
    field_2 = dynamic-environment

Frame:
    subtype = ___sFRAME
    field_0 = next frame (C pointer if unscanned, otherwise Scheme obj)
    field_1 = frame slot 1
    field_2 = frame slot 2
    ...

*/

/*---------------------------------------------------------------------------*/

#ifdef CONSISTENCY_CHECKS

static int reference_location; /* where is offending reference located */

#define IN_OBJECT     0
#define IN_REGISTER   1
#define IN_GLOBAL_VAR 2
#define IN_WILL_LIST  3
#define IN_STACK      4

static ___WORD *container_body; /* pointer to body of object      */
                                /* containing offending reference */

static void bug (obj, msg)
___WORD obj;
char *msg;
{
  fprintf (stderr, ">>> The GC has detected the following inconsistency:\n");
  fprintf (stderr, ">>> The object 0x%08lx %s\n", obj, msg);

  switch (reference_location)
    {
    case IN_OBJECT:
      {
        ___WORD head = container_body[-1];
        unsigned long words = ___HD_WORDS(head);
        unsigned int subtype = ___HD_SUBTYPE(head);
        unsigned int i;

        fprintf (stderr, ">>> The reference was found while scanning ");
        if (___TYP(head) == ___PERM)
          fprintf (stderr, "___PERM ");
        else if (___TYP(head) == ___STILL)
          fprintf (stderr, "___STILL ");
        else if (___TYP(head) == ___MOVABLE)
          fprintf (stderr, "___MOVABLE ");
        else
          fprintf (stderr, "___FORW ");
        fprintf (stderr, "object:\n");
        fprintf (stderr, ">>>  subtype = %d\n", subtype);
        fprintf (stderr, ">>>  length  = %d words\n", words);
        for (i=0; i<words; i++)
          fprintf (stderr, ">>>  body[%2d] = 0x%08lx\n", i, container_body[i]);
        break;
      }

    case IN_REGISTER:
      fprintf (stderr, ">>> The reference was found in a register\n");
      break;

    case IN_GLOBAL_VAR:
      fprintf (stderr, ">>> The reference was found in a global variable\n");
      break;

    case IN_WILL_LIST:
      fprintf (stderr, ">>> The reference was found in a will list\n");
      break;

    case IN_STACK:
      fprintf (stderr, ">>> The reference was found in the stack\n");
      break;
    }

  fflush (stderr);
  fatal_error ("GC inconsistency detected");
}

static void validate_old_obj (obj)
___WORD obj;
{
  ___WORD *hd_ptr = ___BODY(obj)-1;
  ___WORD head;
  int i = find_arena_section (the_arena, hd_ptr);
  if (i >= 0 && i < the_arena->nb_sections)
    {
      long pos = hd_ptr - the_arena->sections[i]->base;
      if (pos >= 0 && pos < ___SECTION_HALFSIZE<<1)
        {
          head = *hd_ptr;
          if (___TYP(head) == ___FORW)
            {
              ___WORD *hd_ptr2 = ___UNTAG_AS(head,___FORW)+___BODY_OFS-1;
              int i2 = find_arena_section (the_arena, hd_ptr2);
              if (i2 >= 0 && i2 < the_arena->nb_sections)
                {
                  long pos2 = hd_ptr2 - the_arena->sections[i2]->base;
                  if (tospace_at_top
                      ? (pos2 < ___SECTION_HALFSIZE ||
                         pos2 >= ___SECTION_HALFSIZE<<1)
                      : (pos2 < 0 ||
                         pos2 >= ___SECTION_HALFSIZE))
                    bug (obj, "was copied outside of tospace");
                  else if (___TYP((*hd_ptr2)) != ___MOVABLE)
                    bug (obj, "was copied and copy is not ___MOVABLE");
                }
              else
                bug (obj, "was copied outside of tospace");
            }
          else if (___TYP(head) != ___MOVABLE)
            bug (obj, "should be ___MOVABLE");
          else if (tospace_at_top
                   ? (pos >= ___SECTION_HALFSIZE &&
                      pos < ___SECTION_HALFSIZE<<1)
                   : (pos >= 0 &&
                      pos < ___SECTION_HALFSIZE))
            bug (obj, "is in tospace");
          return;
        }
    }
  head = *hd_ptr; /* this dereference will likely bomb if there is a bug */
  if (___TYP(head) != ___PERM && ___TYP(head) != ___STILL)
    bug (obj, "is not ___PERM or ___STILL");
}

#endif


static ___WORD *start_of_fromspace (s)
arena_section *s;
{
  if (tospace_at_top)
    return s->base;
  else
    return s->base + ___SECTION_HALFSIZE;
}


static ___WORD *start_of_tospace (s)
arena_section *s;
{
  if (tospace_at_top)
    return s->base + ___SECTION_HALFSIZE;
  else
    return s->base;
}


static void next_alloc_section ()
{
  if (alloc_section == 0)
    {
      nb_prev_sections = 0;
      words_prev_sections = 0;
      alloc_section = the_arena->head;
    }
  else
    {
      nb_prev_sections++;
      words_prev_sections += alloc_ptr - start_ptr;
      alloc_section->alloc = alloc_ptr;
      alloc_section = alloc_section->next;
    }
  if (alloc_section == 0)
    fatal_error ("Heap overflow");
  start_ptr = start_of_tospace (alloc_section);
  alloc_ptr = start_ptr;
  alloc_limit = start_ptr + ___SECTION_HALFSIZE;
}


static void mark_array (start, n, refcount)
___WORD *start;
___WORD n;
___WORD refcount;
{
  ___WORD *alloc = alloc_ptr;
  ___WORD *limit = alloc_limit;
  while (n > 0)
    {
      ___WORD obj = *start;
      if (___MEM_ALLOCATED(obj))
        {
          ___WORD *body;
          ___WORD head;
          unsigned int head_typ;

#ifdef CONSISTENCY_CHECKS
          validate_old_obj (obj);
#endif

          body = ___UNTAG(obj) + ___BODY_OFS;
          head = body[-1];
          head_typ = ___TYP(head);

          if (head_typ == ___MOVABLE)
            {
              unsigned long words = ___HD_WORDS(head);
#if ___WS == 4
              unsigned int subtype = ___HD_SUBTYPE(head);
              int pad = 0;
              while (alloc + words + (subtype == ___sFLONUM ? 2 : 1) > limit)
#else
              while (alloc + words + 1 > limit)
#endif
                {
                  alloc_ptr = alloc;
                  next_alloc_section ();
                  alloc = alloc_ptr;
                  limit = alloc_limit;
                }
#if ___WS == 4
              if (subtype == ___sFLONUM)
                if (((___WORD)alloc&7) == 0)
                  *alloc++ = ___MAKE_HD_WORDS(0, ___sVECTOR);
                else
                  pad = 1;
#endif
              *alloc++ = ___ADD_RC(___RESET_RC(head), refcount);
              *start = ___TAG((alloc - ___BODY_OFS), ___TYP(obj));
              body[-1] = ___TAG((alloc - ___BODY_OFS), ___FORW);
              while (words > 0)
                {
                  *alloc++ = *body++;
                  words--;
                }
#if ___WS == 4
              if (pad)
                *alloc++ = ___MAKE_HD_WORDS(0, ___sVECTOR);
#endif
            }
          else if (head_typ == ___STILL)
            {
              if (body[___STILL_MARK_OFS - ___STILL_BODY_OFS] == -1)
                {
                  body[-1] = ___ADD_RC(___RESET_RC(body[-1]), refcount);
                  body[___STILL_MARK_OFS - ___STILL_BODY_OFS]
                    = (___WORD)still_objs_to_scan;
                  still_objs_to_scan = (___WORD)(body - ___STILL_BODY_OFS);
                }
              else
                body[-1] = ___ADD_RC(body[-1], refcount);
            }
          else if (head_typ == ___FORW)
            {
              ___WORD *copy_body = ___UNTAG_AS(head, ___FORW) + ___BODY_OFS;
              copy_body[-1] = ___ADD_RC(copy_body[-1], refcount);
              *start = ___TAG((copy_body - ___BODY_OFS), ___TYP(obj));
            }
        }
      start++;
      n--;
    }
  alloc_ptr = alloc;
}


extern ___WORD ___internal_return;

static void mark_continuation (orig_ptr, ra)
___WORD *orig_ptr;
___WORD ra;
{
  ___WORD frame, *ptr = orig_ptr;

next_frame:

  frame = *ptr;

  if (frame != 0 && ___TYP(frame) != ___tSUBTYPED)
    {
      /* continuation's frame is in the stack */

      ___WORD *fp, frame_ra;
      int fs, link, i;

      if (ra == ___internal_return)
        {
          ___WORD fs_link = ((___lbl_struct*)(((___WORD*)frame)[0]-___tSUBTYPED))->flags;
          fs = (fs_link>>3)&((1<<14)-1);
          link = (fs_link>>17)&((1<<14)-1);
        }
      else
        {
          ___WORD fs_link = ((___lbl_struct*)(ra-___tSUBTYPED))->flags;
          fs = (fs_link>>3)&((1<<14)-1);
          link = (fs_link>>17)&((1<<14)-1);
        }
      fp = ((___WORD*)frame)+fs;
      frame_ra = fp[-link-1];

      if (___TYP(frame_ra) == ___tFIXNUM)
        *ptr = frame_ra;
      else
        {
          ___WORD *copy_body;
          long words = fs+1;

          while (alloc_ptr + words + 1 > alloc_limit)
            next_alloc_section ();
          *alloc_ptr++ = ___MAKE_HD_WORDS(words, ___sFRAME);
          copy_body = alloc_ptr;
          alloc_ptr += words;

          if (frame_ra == pstate->handler_break)
            {
              /* first frame of that section */

              alloc_ptr[-fs-1] = fp[1];
              ra               = fp[0];
              for (i=-fs; i<0; i++)
                alloc_ptr[i] = fp[i];

              alloc_ptr[-link-1] = ra;
            }
          else
            {
              /* not the first frame of that section */

              alloc_ptr[-fs-1] = (___WORD)fp;
              ra               = frame_ra;

              for (i=-fs; i<0; i++)
                alloc_ptr[i] = fp[i];
            }

          /* leave a forwarding pointer */
          fp[-link-1] = ___TAG((copy_body - ___BODY_OFS), ___tFIXNUM);

          *ptr = ___TAG((copy_body - ___BODY_OFS), ___tFIXNUM);

          ptr = &alloc_ptr[-fs-1];

          goto next_frame;
        }
    }

  if (*orig_ptr != 0 && ___TYP((*orig_ptr)) == ___tFIXNUM)
    *orig_ptr = ___TAG(___UNTAG_AS(*orig_ptr, ___tFIXNUM), ___tSUBTYPED);
  else
    mark_array (orig_ptr, 1, ___RCN);
}


static void move_stack ()
{
  ___WORD *fp = pstate->fp;
  ___WORD *p1 = pstate->sb+2;
  ___WORD *p2 = pstate->stack_base;

  if (p1 != p2)
    {
      pstate->sb = p2-2;
      while (p1 != fp)
        *--p2 = *--p1;
      pstate->fp = p2;
    }
}


static unsigned long scan (body)
___WORD *body;
{
  ___WORD head = body[-1];
  unsigned long words = ___HD_WORDS(head);
  unsigned int subtype = ___HD_SUBTYPE(head);

#ifdef CONSISTENCY_CHECKS
  reference_location = IN_OBJECT;
  container_body = body;
#endif

  switch (subtype)
    {
    case ___sSTRING:
    case ___sBIGNUM:
    case ___sFLONUM:
      break;

    case ___sWILL:
      mark_array (body, 1, ___RC0); /* pointers to next will and owner are weak */
      mark_array (body+1, words-1, ___RCN);
      break;

    case ___sSYMBOL:
      mark_array (body, 1, ___RCN); /* only scan symbol's name */
      break;

    case ___sCONTINUATION:
      mark_continuation (body, body[1]);
      mark_array (body+2, 1, ___RCN); /* only scan dynamic environment */
      break;

    case ___sFRAME:
      if (body[0] != 0 && ___TYP((body[0])) == ___tFIXNUM)
        body[0] = ___TAG(___UNTAG_AS(body[0], ___tFIXNUM), ___tSUBTYPED);
      else
        mark_array (body, 1, ___RCN);
      mark_array (body+1, words-1, ___RCN);
      break;

    case ___sPROCEDURE:
      if (___TYP(head) != ___PERM) /* only scan closures */
        mark_array (body+1, words-1, ___RCN); /* only scan free variables */
      break;

    default:
      mark_array (body, words, ___RCN);
      break;
    }
  return words;
}


static void init_still_objs_to_scan ()
{
  ___WORD *base = (___WORD*)still_objs;
  ___WORD *to_scan = 0;
  while (base != 0)
    {
      if (base[___STILL_REFCOUNT_OFS] == 0)
        base[___STILL_MARK_OFS] = -1;
      else
        {
          base[___STILL_BODY_OFS-1] = ___ADD_RC(base[___STILL_BODY_OFS-1], ___RCN);
          base[___STILL_MARK_OFS] = (___WORD)to_scan;
          to_scan = base;
        }
      base = (___WORD*)base[___STILL_LINK_OFS];
    }
  still_objs_to_scan = (___WORD)to_scan;
}


static void scan_still_objs_to_scan ()
{
  ___WORD *base;
  while ((base = (___WORD*)still_objs_to_scan) != 0)
    {
      still_objs_to_scan = base[___STILL_MARK_OFS];
      scan (base + ___STILL_BODY_OFS);
    };
}


static void scan_movable_objs_to_scan ()
{
  while (scan_section != alloc_section || scan_ptr < alloc_ptr)
    {
      ___WORD *body = scan_ptr + 1;
      unsigned long words = scan (body);
      scan_ptr = body + words;
      if (scan_section != alloc_section &&
          scan_ptr >= scan_section->alloc)
        {
          scan_section = scan_section->next;
          scan_ptr = start_of_tospace (scan_section);
        }
    };
}


static void free_unmarked_still_objs ()
{
  ___WORD *last = &still_objs;
  ___WORD *base = (___WORD*)*last;
  while (base != 0)
    {
      ___WORD link = base[___STILL_LINK_OFS];
      if (base[___STILL_MARK_OFS] == -1)
        {
          ___WORD head = base[___STILL_BODY_OFS-1];
          words_non_movable -= ___HD_WORDS(head) + ___STILL_BODY_OFS;
          ___free_aligned_block (base);
        }
      else
        {
          *last = (___WORD)base;
          last = base + ___STILL_LINK_OFS;
        }
      base = (___WORD*)link;
    }
  *last = 0;
}


static long (*adjust_hook)();


static long default_adjust_hook (avail, live)
long avail, live;
{
  long target = live * 2;
  if (max_heap_size == 0 || target < max_heap_size)
    return target;
  else
    return max_heap_size;
}


static void setup_pstate ()
{
  long unused = words_available - words_allocated;

  if (unused > 0)
    if (alloc_ptr + unused/2 > alloc_limit - ___semispace_fudge)
      pstate->hl = alloc_limit - ___semispace_fudge;
    else
      pstate->hl = alloc_ptr + unused/2;
  else
    pstate->hl = alloc_ptr;

  pstate->hp = alloc_ptr;
}


void ___setup_mem (max_heap_len, gc_hook)
long max_heap_len;
long (*gc_hook)();
{
  /* allocate processor state structure */

  pstate = (___processor)___malloc_aligned_block (sizeof(___processor_struct),
                                                  ___WS, 0);

  /* allocate stack */

  ___stack_size = ___STACK_SIZE;
  ___stack_fudge = ___STACK_FUDGE;

  pstate->stack =
    (___WORD*)___malloc_aligned_block (___stack_size<<___LWS, ___WS, 0);

  pstate->stack_base = pstate->stack + ___stack_size;
  pstate->fp = pstate->stack_base;

  /* allocate heap */

  max_heap_size = max_heap_len >> ___LWS;
  ___semispace_fudge = ___SEMISPACE_FUDGE;
  ___overflow_reserve = ___OVERFLOW_RESERVE;

  if (gc_hook == 0)
    adjust_hook = default_adjust_hook;
  else
    adjust_hook = gc_hook;

  the_arena = 0;
  adjust_arena (&the_arena, ___MIN_NB_SECTIONS);
  if (the_arena == 0 || the_arena->nb_sections < ___MIN_NB_SECTIONS)
    fatal_error ("Can't allocate initial heap");

  still_objs = 0;

  tospace_at_top = 0;

  words_non_movable = 0;
  words_prev_sections = 0;

  alloc_section = 0;
  next_alloc_section ();

  setup_pstate ();
}


int ___gc ()
{
  long i;
  long unused;
  int target_nb_sections;
  int overflow = 0;

  if ((___G__23_23gc_2dreport.val != ___FAL) || debug_level)
  {
    fprintf (stderr, "*** GC...");
    fflush (stderr);
  }

  tospace_at_top = !tospace_at_top;

  alloc_section = 0;
  next_alloc_section ();
  scan_section = alloc_section;
  scan_ptr = alloc_ptr;

  /* transfer all wills to the executable will list */

  {
    ___WORD wills = pstate->non_executable_wills;
    if (wills != ___FAL)
      {
        ___WORD p1 = wills;
        ___WORD p2 = ___BODY_AS(p1, ___tSUBTYPED)[2];
        while (p2 != ___FAL)
          {
            p1 = p2;
            p2 = ___BODY_AS(p1, ___tSUBTYPED)[2];
          }
        ___BODY_AS(p1, ___tSUBTYPED)[2] = pstate->executable_wills;
        pstate->executable_wills = wills;
        pstate->non_executable_wills = ___FAL;
      }
  }

  /* trace externally referenced still objects */

  init_still_objs_to_scan ();

  /* trace registers */

#ifdef CONSISTENCY_CHECKS
    reference_location = IN_REGISTER;
#endif

  mark_array (pstate->r, ___MAX_NB_GVM_REGS, ___RCN);

  /* trace will lists */

#ifdef CONSISTENCY_CHECKS
    reference_location = IN_WILL_LIST;
#endif

  mark_array (&pstate->executable_wills, 1, ___RCN);
  mark_array (&pstate->non_executable_wills, 1, ___RCN);

  /* trace global variables */

#ifdef CONSISTENCY_CHECKS
    reference_location = IN_GLOBAL_VAR;
#endif

  {
    ___glo_struct *p = pstate->glo_list;

    while (p != 0)
    {
      mark_array (&p->val, 1, ___RCN);
      p = p->next;
    }
  }

  /* trace stack */

#ifdef CONSISTENCY_CHECKS
    reference_location = IN_STACK;
#endif

  mark_continuation (&pstate->sb[1], pstate->sb[0]);

  mark_array (pstate->fp, pstate->sb-pstate->fp, ___RCN);

again:

  if ((___WORD*)still_objs_to_scan != 0)
    scan_still_objs_to_scan ();

  if (scan_section != alloc_section || scan_ptr < alloc_ptr)
    {
      scan_movable_objs_to_scan ();
      goto again;
    }

  move_stack ();

#ifdef CONSISTENCY_CHECKS
  {
    int i, j;
    ___WORD *p;
    p = pstate->fp;
    while (p > pstate->stack)
      *--p = -1;
    for (i=0; i<the_arena->nb_sections; i++)
      {
        p = start_of_fromspace (the_arena->sections[i]);
        for (j=___SECTION_HALFSIZE-1; j>=0; j--)
          *p++ = -1;
      }
  }
#endif

  free_unmarked_still_objs ();

  target_nb_sections = (adjust_hook (words_available, words_allocated) +
                        ___overflow_reserve*2 +
                        2*(___SECTION_HALFSIZE-___semispace_fudge+1) - 1) /
                       (2*(___SECTION_HALFSIZE-___semispace_fudge+1));

  if (target_nb_sections < nb_prev_sections+1)
    target_nb_sections = nb_prev_sections+1;

  if (target_nb_sections < ___MIN_NB_SECTIONS)
    target_nb_sections = ___MIN_NB_SECTIONS;

  adjust_arena (&the_arena, target_nb_sections);

  if (alloc_ptr > alloc_limit - ___semispace_fudge)
    next_alloc_section ();

  unused = words_available - words_allocated;

  if (___overflow_reserve == 0)
    {
      if (unused/2 >= ___OVERFLOW_RESERVE)
        ___overflow_reserve = ___OVERFLOW_RESERVE;
    }
  else if (unused <= 0)
    {
      ___overflow_reserve = 0;
      overflow = 1;
    }

  setup_pstate ();

  if ((___G__23_23gc_2dreport.val != ___FAL) || debug_level)
    {
      fprintf (stderr,
               " %ld%% live, %ld KB heap, movable=%ld, non-movable=%ld\n",
               100*words_allocated/words_available,
               words_available>>10-___LWS,
               words_movable<<___LWS,
               words_non_movable<<___LWS);
      fflush (stderr);
    }

  raise_interrupt(2); /* gc interrupt */

  return overflow;
}


int ___heap_limit ()
{
  long unused;

  alloc_ptr = pstate->hp;

  unused = words_available - words_allocated;

  if (unused > 0)
    {
      next_alloc_section ();
      if (alloc_ptr + unused/2 > alloc_limit - ___semispace_fudge)
        pstate->hl = alloc_limit - ___semispace_fudge;
      else
        pstate->hl = alloc_ptr + unused/2;
      pstate->hp = alloc_ptr;
      return 0;

    }

  return ___gc ();
}


/*---------------------------------------------------------------------------*/
