/* file: "os.c" */

/* Copyright (C) 1994-1996 by Marc Feeley, All Rights Reserved. */

/*
This module implements the operating system specific routines including:

  - filename expansion
  - dynamic loading
  - user and timer interrupt handling
  - shell command
  - elapsed time
*/

#define ___VERSION 9
#include "gambit.h"
___BEGIN_MODULE

#include "os.h"

/*---------------------------------------------------------------------------*/

/* Filename expansion */

#include <stdlib.h>
#include <string.h>

#ifdef __WINDOWS__
#include <direct.h>
#else
#include <pwd.h>
#include <sys/types.h>
#include <unistd.h>
#endif


#ifdef __WINDOWS__

#define DIR_SEPARATOR1 '\\'
#define DIR_SEPARATOR2 '/'
#define DRIVE_SEPARATOR ':'

#define DIR_SEPARATOR(c)((c) == DIR_SEPARATOR1 || (c) == DIR_SEPARATOR2)
#define SEPARATOR(c)(DIR_SEPARATOR(c) || (c) == DRIVE_SEPARATOR)

static int normalize_dir_path (dir_path, norm_dir_path, max_size)
char *dir_path, *norm_dir_path;
long max_size;
{
  char cwd1[PATH_MAX_SIZE+1];
  int result = 0;
  if (getcwd (cwd1, PATH_MAX_SIZE+1) != 0)
    {
      long i = 0;
      while (dir_path[i] != '\0' && !SEPARATOR(dir_path[i]))
        i++;
      if (dir_path[i] == DRIVE_SEPARATOR && dir_path[i+1] != '\0')
        {
          char cwd2[PATH_MAX_SIZE+1];
          char tmp = dir_path[i+1];
          dir_path[i+1] = '\0';
          if (chdir (dir_path) == 0)
            {
              if (getcwd (cwd2, PATH_MAX_SIZE+1) != 0)
                {
                  dir_path[i+1] = tmp;
                  if (chdir (dir_path) == 0)
                    {
                      result = getcwd (norm_dir_path, max_size) != 0;
                      chdir (cwd2);
                    }
                }
              chdir (cwd1);
            }
        }
      else if (chdir (dir_path) == 0)
        {
          result = getcwd (norm_dir_path, max_size) != 0;
          chdir (cwd1);
        }
    }
  return result;
}

#else

#define DIR_SEPARATOR1 '/'

#define DIR_SEPARATOR(c)((c) == DIR_SEPARATOR1)
#define SEPARATOR(c)DIR_SEPARATOR(c)

static int normalize_dir_path (dir_path, norm_dir_path, max_size)
char *dir_path, *norm_dir_path;
long max_size;
{
  char cwd[PATH_MAX_SIZE+1];
  int result = 0;
  if (getcwd (cwd, PATH_MAX_SIZE+1) != 0 && chdir (dir_path) == 0)
    {
      result = getcwd (norm_dir_path, max_size) != 0;
      chdir (cwd);
    }
  return result;
}

#endif

static char root_path[] = { DIR_SEPARATOR1, '\0' };

int os_absolute_path (path, abs_path, max_size)
char *path, *abs_path;
long max_size;
{
  char tmp[PATH_MAX_SIZE+1];
  long i = 0;
  long j = 0;
  long k;

  if (path[0] == '\0')
    if (getcwd (abs_path, max_size) == 0)
      return 0;
    else
      goto terminal_dir_separator;

  if (path[0] == '~')
    {
      char *home;

      i++;
      while (path[i]!='\0' && !DIR_SEPARATOR(path[i]) && i <= PATH_MAX_SIZE+1)
        {
          tmp[i-1] = path[i];
          i++;
        }

      if (i > PATH_MAX_SIZE+1)
        return 0;

      if (i == 1)
        {
          home = getenv ("HOME");
          if (home == 0)
            home = root_path;
        }
      else
#ifdef __WINDOWS__
        return 0;
#else
        {
          struct passwd *p;
          tmp[i-1] = '\0';
          p = getpwnam (tmp);
          if (p == 0)
            return 0;
          home = p->pw_dir;
          if (home == 0)
            return 0;
        }
#endif

      j = 0;
      while (*home != '\0' && j < PATH_MAX_SIZE+1)
        tmp[j++] = *home++;
    }

  k = i;

  while (path[i] != '\0' && j < PATH_MAX_SIZE+1)
    tmp[j++] = path[i++];

  if (j >= PATH_MAX_SIZE+1)
    return 0;

  tmp[j] = '\0';

  if (normalize_dir_path (tmp, abs_path, max_size))
    {
      terminal_dir_separator:
      j = strlen (abs_path);
      if (j > 0 && !DIR_SEPARATOR(abs_path[j-1]))
        {
          abs_path[j++] = DIR_SEPARATOR1;
          if (j >= max_size)
            return 0;
          abs_path[j] = '\0';
        }
      return 1;
    }

  i = strlen (path);
  while (i > k && !SEPARATOR(path[i]))
    {
      i--;
      j--;
    }

  if (SEPARATOR(path[i]))
    {
      i++;
      j++;
      if (j >= PATH_MAX_SIZE+1)
        return 0;
      tmp[j] = '\0';
      if (!normalize_dir_path (tmp, abs_path, max_size))
        return 0;
    }
  else
    {
      if (getcwd (abs_path, max_size) == 0)
        return 0;
    }

  j = strlen (abs_path);
  if (!DIR_SEPARATOR(abs_path[j-1]))
    abs_path[j++] = DIR_SEPARATOR1;

  while (path[i] != '\0' && j < max_size)
    abs_path[j++] = path[i++];

  if (j >= max_size)
    return 0;

  abs_path[j] = '\0';

  return 1;
}

/*---------------------------------------------------------------------------*/

/* Dynamic loading. */

#ifdef sparc
#define ___USE_DLOPEN
#endif

#ifdef sgi
#define ___USE_DLOPEN
#endif

#ifdef __alpha
#define ___USE_DLOPEN
#endif

#ifdef linux
#ifdef i386
#define ___USE_DLOPEN
#endif
#endif

#ifdef ___USE_DLOPEN

#include <dlfcn.h>

#else

#ifdef hpux
#include <dl.h>
#endif

#ifdef __WINDOWS__
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

#endif

void *os_dyn_load (path, errmsg)
char *path;
char **errmsg;
{
  void *result = 0;
  char *module_linker = "___init_module";

#ifdef ___USE_DLOPEN

#ifdef RTLD_NOW
  void *descr = dlopen (path, RTLD_NOW);
#else
  void *descr = dlopen (path, 1);
#endif
  if (descr != 0)
    result = dlsym (descr, module_linker);
  if (result == 0)
    *errmsg = (char *)dlerror();

#else

#ifdef hpux
  shl_t descr = shl_load (path, BIND_IMMEDIATE, 0);
  if (descr != 0)
    if (shl_findsym (&descr, module_linker, TYPE_PROCEDURE, &result))
      result = 0;
  if (result == 0)
    *errmsg = "unknown error";
#endif

#ifdef __WINDOWS__
  HANDLE descr = LoadLibrary (path);
  if (descr != 0)
    result = GetProcAddress (descr, module_linker);
  if (result == 0)
    *errmsg = "unknown error";
#endif

#endif

  return result;
}

/*---------------------------------------------------------------------------*/

/* User and timer interrupt handling. */

#include "setup.h"

#include <time.h>

#ifndef __WINDOWS__
#ifndef __MWERKS__
#include <sys/time.h>
#endif
#endif

/* . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . */

#ifdef __MWERKS__

void set_timer_interval( interval )
int interval;
{
}

static void install_interrupt_handlers()
{
}

#endif

/* . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . */

#ifdef GO32

#include <signal.h>

static void user_interrupt_handler()
{
  raise_interrupt(0);
}

void set_timer_interval( interval )
int interval;
{
}

static void install_interrupt_handlers()
{
  signal (SIGINT, user_interrupt_handler);
}

#endif

/* . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . */

#ifdef MACHTEN

#include <signal.h>

static void user_interrupt_handler(sig, code, scp)
int sig, code;
struct sigcontext *scp;
{
  raise_interrupt(0);
}

static void timer_interrupt_handler(sig, code, scp)
int sig, code;
struct sigcontext *scp;
{
  raise_interrupt(1);
}

void set_timer_interval( interval )
int interval;
{
  struct itimerval virtual_tv;
  int secs = interval/1000;
  int usecs = (interval%1000)*1000;
  virtual_tv.it_interval.tv_sec  = secs;
  virtual_tv.it_interval.tv_usec = usecs;
  virtual_tv.it_value.tv_sec     = secs;
  virtual_tv.it_value.tv_usec    = usecs;
  setitimer( ITIMER_VIRTUAL, &virtual_tv, 0 );
}

static void install_interrupt_handlers()
{
  struct sigvec vec;

  vec.sv_handler = user_interrupt_handler;
  vec.sv_flags = SV_INTERRUPT;
  vec.sv_mask = -1;
  sigvec(SIGINT, &vec, 0);

  vec.sv_handler = timer_interrupt_handler;
  vec.sv_flags = SV_INTERRUPT;
  vec.sv_mask = -1;
  sigvec(SIGVTALRM, &vec, 0);

  set_timer_interval( 100 );
}

#endif

/* . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . */

#ifdef __WINDOWS__

#include <signal.h>

static void user_interrupt_handler (n)
int n;
{
  signal (SIGINT, user_interrupt_handler);
  raise_interrupt(0);
}

static void install_interrupt_handlers()
{
  signal (SIGINT, user_interrupt_handler);
}

void set_timer_interval( interval )
int interval;
{
}

#endif

/* . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . */

#ifndef __MWERKS__
#ifndef GO32
#ifndef MACHTEN
#ifndef __WINDOWS__

#include <signal.h>

static void user_interrupt_handler()
{
  raise_interrupt(0);
}

static void timer_interrupt_handler()
{
  raise_interrupt(1);
}

void set_timer_interval( interval )
int interval;
{
  struct itimerval virtual_tv;
  int secs = interval/1000;
  int usecs = (interval%1000)*1000;
  virtual_tv.it_interval.tv_sec  = secs;
  virtual_tv.it_interval.tv_usec = usecs;
  virtual_tv.it_value.tv_sec     = secs;
  virtual_tv.it_value.tv_usec    = usecs;
  setitimer( ITIMER_VIRTUAL, &virtual_tv, 0 );
}

static void install_interrupt_handlers()
{
  struct sigaction act;

  act.sa_handler = user_interrupt_handler;
  act.sa_flags = 0;
#ifdef SA_INTERRUPT
  act.sa_flags |= SA_INTERRUPT;
#endif
  sigemptyset(&act.sa_mask);
  sigaction(SIGINT, &act, 0);

  act.sa_handler = timer_interrupt_handler;
  act.sa_flags = 0;
#ifdef SA_INTERRUPT
  act.sa_flags |= SA_INTERRUPT;
#endif
  sigemptyset(&act.sa_mask);
  sigaction(SIGVTALRM, &act, 0);

  set_timer_interval( 100 );
}

#endif
#endif
#endif
#endif

/*---------------------------------------------------------------------------*/

/* Shell command. */

#ifdef __MWERKS__

int os_shell_command( cmd )
char *cmd;
{
}

#else

int os_shell_command( cmd )
char *cmd;
{
  int result;
  set_timer_interval( 0 );
  result = system( cmd );
  set_timer_interval( 100 );
  return result;
}

#endif

/*---------------------------------------------------------------------------*/

/* Elapsed time. */

#ifdef __MWERKS__

static unsigned long initial_tickcount;

unsigned long os_clock()
{
  return ((unsigned long)TickCount()-initial_tickcount) * 50 / 3;
}

void init_elapsed_time()
{
  initial_tickcount = (unsigned long)TickCount();
}

#else

#ifdef __WINDOWS__
#define ___TIME_WITH_TIME
#endif

#ifndef ___TIME_WITH_CLOCK
#ifndef ___TIME_WITH_TIME
#ifndef ___TIME_WITH_GETRUSAGE
#ifndef ___TIME_WITH_TIMES
#define ___TIME_WITH_TIMES
#endif
#endif
#endif

#ifdef ___TIME_WITH_TIMES

#include <sys/times.h>
#include <time.h>

#ifdef CLK_TCK
#define CLOCK_TICKS_PER_SEC CLK_TCK
#else
#ifdef _SC_CLOCKS_PER_SEC
#define CLOCK_TICKS_PER_SEC sysconf(_SC_CLOCKS_PER_SEC)
#else
#ifdef _SC_CLK_TCK
#define CLOCK_TICKS_PER_SEC sysconf(_SC_CLK_TCK)
#else
@error Cannot find a definition for CLOCK_TICKS_PER_SEC
#endif
#endif
#endif

unsigned long os_clock()
{
  struct tms t;
  times( &t );
  return t.tms_utime * 1000 / CLOCK_TICKS_PER_SEC;
}

#else

#ifdef ___TIME_WITH_GETRUSAGE

#include <sys/time.h>
#include <sys/resource.h>

unsigned long os_clock()
{
  struct rusage ru;
  getrusage( RUSAGE_SELF, &ru );
  return ru.ru_utime.tv_sec * 1000 + ru.ru_utime.tv_usec / 1000;
}

#else

#ifdef ___TIME_WITH_TIME

#include <time.h>

unsigned long os_clock()
{
  time_t t;
  time(&t);
  return t*1000;
}

#else

#ifdef ___TIME_WITH_CLOCK

unsigned long os_clock()
{
  return clock()*1000/CLOCKS_PER_SEC;
}

#else

@error Cannot find a timing function

#endif
#endif
#endif
#endif
#endif

void init_elapsed_time()
{
}

#endif

/*---------------------------------------------------------------------------*/

void os_init()
{
  install_interrupt_handlers();
  init_elapsed_time();
}
